/* check-tabl.c --- Report changed bases in a tabl-format file
   Jim Blandy <jimb@cyclic.com> --- Nov 1994 */

#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <limits.h>
#include <memory.h>
#include <assert.h>

#include "xmalloc.h"
#include "lenstring.h"
#include "hash.h"

struct file_and_name {
  char *file;
  char *name;
};

void read_file (char *filename, lenstring *contents);
struct string_hash *read_and_hash_alignment (struct file_and_name *file_names,
					     lenstring *contents);

void check_file (struct file_and_name *file_names);

void check_syscall (int call, char *filename, char *op);
FILE *careful_open (char *name, char *mode, FILE *defalt);
void give_usage (void);

char *progname;

struct file_and_name reference_file_names;
lenstring reference_contents;
struct string_hash *reference_hash;

void init_gap_table (void);
char gap_table[UCHAR_MAX + 1];
#define is_gap(c) (gap_table[(unsigned char) c])
#define set_gap(c) (gap_table[(unsigned char) (c)] = 1)

#define LOOKUP_SEQUENCE(table, string, len) \
  ((lenstring **) lookup_hash_table ((table), (string), (len)))
#define LOOKUP_SEQUENCE_SOFT(table, string, len) \
  ((lenstring **) lookup_hash_table_soft ((table), (string), (len)))

int
main (int argc, char **argv)
{
  int i;
  int checked_any = 0;

  progname = argv[0];
  init_gap_table ();

  reference_file_names.file = 0;
  for (i = 1; i < argc; i++)
    if (argv[i][0] == '-')
      give_usage ();
    else if (! reference_file_names.file)
      {
	reference_file_names.file = argv[i];
	reference_file_names.name = argv[i];
	if (i + 2 < argc
	    && argv[i+1][0] == '-'
	    && argv[i+1][1] == 'n')
	  reference_file_names.name = argv[i += 2];

	reference_hash = read_and_hash_alignment (&reference_file_names,
						  &reference_contents);
      }
    else 
      {
	struct file_and_name input_file_names;
       
	input_file_names.file = argv[i];
	input_file_names.name = argv[i];
	if (i + 2 < argc
	    && argv[i+1][0] == '-'
	    && argv[i+1][1] == 'n')
	  input_file_names.name = argv[i += 2];

	checked_any = 1;
	check_file (&input_file_names);
      }
	
  if (! reference_file_names.name)
    {
      fprintf (stderr, "%s: no filenames given\n", progname);
      give_usage ();
    }

  if (! checked_any)
    {
      fprintf (stderr, "%s: compare %s to what?\n",
	       progname, reference_file_names.name);
      give_usage ();
    }

  return 0;
}


void
init_gap_table ()
{
  memset (gap_table, 0, UCHAR_MAX + 1);

  set_gap ('-');
  set_gap ('~');
  set_gap ('|');
  set_gap ('.');
  set_gap (' ');
}

struct string_hash *
read_and_hash_alignment (struct file_and_name *file_names,
			 lenstring *contents)
{
  struct string_hash *table = new_hash_table ();

  read_file (file_names->file, contents);

  {
    char *p = contents->text;
    char *end = p + contents->len;

    while (p < end)
      {
	char *name_start, *name_end;
	char *seq_start, *seq_end;

	/* Find the extent of the name.  */
	name_start = p;
	while (p < end && *p != '\n')
	  p++;
	name_end = p;

	/* Skip the newline.  */
	if (p < end) p++;
	
	/* Find the extent of the sequence, and remove all gaps.  */
	for (seq_start = seq_end = p; p < end && *p != '\n'; p++)
	  if (! is_gap(*p))
	    *seq_end++ = *p;

	/* Skip the newline.  */
	if (p < end) p++;
	
	/* Hash the name and its sequence.  */
	{
	  lenstring *l, **cell;

	  cell = LOOKUP_SEQUENCE (table, name_start, name_end - name_start);

	  if (*cell)
	    {
	      fprintf (stderr, "%s: warning: the sequence  ", progname);
	      fwrite (name_start,
		      sizeof (*name_start), name_end - name_start,
		      stderr);
	      fprintf (stderr, "  appears twice in  %s\n", file_names->name);
	    }
	  else
	    {
	      *cell = l = (lenstring *) xmalloc (sizeof (**cell));
	      l->text = seq_start;
	      l->len  = seq_end - seq_start;
	    }
	}
      }
  }

  return table;
}


void
read_file (char *filename, lenstring *contents)
{
  struct stat buf;
  size_t len;
  FILE *f;

  check_syscall (stat (filename, &buf), filename, "while statting");
  contents->len = buf.st_size;
  contents->text = (char *) xmalloc (contents->len
				      * sizeof (*contents->text));
  f = careful_open (filename, "r", 0);
  /* Well, if it's possible, we certainly won't be I/O bound:  */
  len = fread (contents->text, sizeof (*contents->text), contents->len, f);
  if (ferror (f))
    check_syscall (-1, filename, "while reading");
  contents->len = len;
  fclose (f);
}


struct compare_closure {
  struct string_hash *other_table;
  int warned;
  char *this_filename;
  char *other_filename;
};

/* Given the name of a sequence from one table, make sure it exists in 
   the other table.
   STRING[0 .. STRING_LEN - 1] is the sequence name in this table;
   OTHER_TABLE is a pointer to a struct compare_closure describing 
   the other table we should check in.  */
void
check_pairing (char *string, size_t string_len,
	       void **cell, void *other_table_ptr)
{
  struct compare_closure *c =
    (struct compare_closure *) other_table_ptr;

  if (! LOOKUP_SEQUENCE_SOFT (c->other_table, string, string_len))
    {
      if (! c->warned)
	{
	  printf ("sequences appearing in  `%s'  but not  `%s':\n",
		  c->this_filename, c->other_filename);
	  c->warned = 1;
	}
      fputs ("    ", stdout);
      fwrite (string, sizeof (*string), string_len, stdout);
      putchar ('\n');
    }
}


size_t
min_size (size_t a, size_t b)
{
  return ((a < b) ? a : b);
}

#define SECTION_LEN (10)

void
print_section (lenstring *seq, size_t start)
{
  size_t len = SECTION_LEN;

  if (len > seq->len - start)
    len = seq->len - start;

  fwrite (seq->text + start, sizeof (*seq->text), len, stdout);

  if (len < SECTION_LEN)
    {
      printf ("(end of sequence)");
      while (len++ < SECTION_LEN)
	putchar (' ');
    }
  else
    fputs ("(continued)      ", stdout);
}

/* Compare the sequence pointed to by **CELL with its corresponding
   sequence in ((struct compare_closure *) other_table_ptr)->other_table.  */
void
compare_sequences (char *string, size_t string_len,
		   void **cell, void *other_table_ptr)
{
  struct compare_closure *c = (struct compare_closure *) other_table_ptr;
  lenstring **other_seq_ptr
    = LOOKUP_SEQUENCE_SOFT (c->other_table, string, string_len);

  if (other_seq_ptr)
    {
      lenstring *this_seq = * (lenstring **) cell;
      lenstring *other_seq = *other_seq_ptr;

      /* If there's an entry in the hash table, it had better have
         some sequence attached to it.  */
      assert (this_seq);
      assert (other_seq);

      if (other_seq->len != this_seq->len
	  || memcmp (other_seq->text, this_seq->text, this_seq->len))
	{
	  if (! c->warned)
	    {
	      printf ("alignments  `%s'  and  `%s'  differ:\n",
		      c->other_filename, c->this_filename);
	      c->warned = 1;
	    }

	  fputs ("    ", stdout);
	  fwrite (string, sizeof (*string), string_len, stdout);
	  fputs (", base ", stdout);

	  /* Find the column # of the first difference, and display
	     some of the text.  */
	  {
	    int i;
	    int len = min_size (other_seq->len, this_seq->len);

	    for (i = 0; i < len; i++)
	      if (this_seq->text[i] != other_seq->text[i])
		break;
	    assert (i < this_seq->len || i < other_seq->len);
	    printf ("%d:\n        ", i);
	    print_section (other_seq, i);
	    fputs ("  vs.  ", stdout);
	    print_section (this_seq,  i);
	    putchar ('\n');
	  }
	}
    }
}

void
check_file (struct file_and_name *input_file_names)
{
  lenstring input_contents;
  struct string_hash *input_hash;

  input_hash = read_and_hash_alignment (input_file_names, &input_contents);

  /* Report mispairings.  */
  {
    struct compare_closure c;

    /* Look for sequences in reference but not input. */
    c.other_table = input_hash;
    c.warned = 0;
    c.this_filename = reference_file_names.name;
    c.other_filename = input_file_names->name;
    map_hash_table (reference_hash, check_pairing, &c);
    if (c.warned)
      putchar ('\n');

    /* Look for sequences in input but not reference. */
    c.other_table = reference_hash;
    c.warned = 0;
    c.this_filename = input_file_names->name;
    c.other_filename = reference_file_names.name;
    map_hash_table (input_hash, check_pairing, &c);
    if (c.warned)
      putchar ('\n');
  }

  /* Compare the sequences.  */
  {
    struct compare_closure c;

    c.other_table = reference_hash;
    c.warned = 0;
    c.this_filename = input_file_names->name;
    c.other_filename = reference_file_names.name;
    map_hash_table (input_hash, compare_sequences, &c);
    if (c.warned)
      putchar ('\n');
  }

  free_hash_table (input_hash);
  free (input_contents.text);
}


/* If NAME is null, return DEFALT.
   If NAME is non-null, try to open it with MODE; print an error message
   and exit if this doesn't succeed.  */
FILE *
careful_open (char *name, char *mode, FILE *defalt)
{
  if (name)
    {
      FILE *result = fopen (name, mode);

      if (! result)
	{
	  fprintf (stderr, "%s: ", progname);
	  perror (name);
	  exit (2);
	}

      return result;
    }
  else
    return defalt;
}


void
give_usage ()
{
  static const char usage[] = "\
Usage: %s REFERENCE FILE ...\n\
Compare alignments; report changes to bases, and ignore changes to gaps.\n\
The tabl-format file REFERENCE is compared against each of the named FILEs.\n\
All files should be alternating lines containing organism names and\n\
organism sequences.  This is the same as interleaved tabl format, with the\n\
annotations omitted.\n\
\n\
Filenames may be followed by options of the form `-n NAME'; if present,\n\
NAME will be used instead of the filename in reports.  This is helpful\n\
when the actual input files are temp files generated by a format\n\
conversion program.\n\
";

  fprintf (stderr, usage, progname);
  exit (1);
}

void
check_syscall (int call, char *filename, char *op)
{
  if (call != -1)
    return;

  fprintf (stderr, "%s: %s%s",
	   progname, 
	   filename ? filename : "",
	   filename ? ": " : "");
  perror (op);

  exit (2);
}
