/* count-tabl.c --- count bases in sequences
   Jim Blandy <jimb@cyclic.com> --- Nov 1994 */

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "xmalloc.h"
#include "lenstring.h"
#include "getopt.h"

typedef struct
{
  lenstring *strings;
  int len;
} stringlist;

#define NUM_CHARS 256
typedef int count_table[NUM_CHARS];

#define CHX(t, c) ((t)[(unsigned) (c)])

void init_gap_table (void);
count_table gap_table;
#define is_gap(c) (CHX (gap_table, (c)))
#define set_gap(c) (is_gap (c) = 1)

void alignment_count (char *ids_filename, char *seqs_filename);
void sequence_count (lenstring *sequence, lenstring *weights, count_table);
void sequence_count_3p (lenstring *sequence, lenstring *weights,
			count_table, count_table, count_table);
void display_counts (count_table counts, char *indentation);
void display_counts_combination (count_table counts, char *indentation);
void display_counts_short (count_table counts);

void display_count (int count, int total);
void clear_counts (count_table);
void add_counts (count_table, count_table);
void thymine_to_uracil (count_table);
void distribute_ambiguities (count_table);
void lower_to_upper (count_table, double w);
int find_mask (stringlist *ids, stringlist *seqs, lenstring *mask);
void mask_to_weights (lenstring *);
void find_lines (lenstring *string, stringlist *lines);
void read_file (char *filename, lenstring *contents);
int check_syscall (int call, char *filename, char *op);
FILE *careful_open (char *name, char *mode, FILE *defalt);
void give_usage (void);

char *progname;

/* Give per-position counts.  */
int option_position;

/* Display percentages of all combinations of bases.  */
int option_combination;

int
main (int argc, char **argv)
{
  int option;

  progname = argv[0];

  while ((option = getopt (argc, argv, "pc")) != -1)
    switch (option)
      {
      case 'p': option_position = 1; break;
      case 'c': option_combination = 1; break;
      case '?': give_usage (); break;
      }

  if (optind + 2 != argc)
    give_usage ();

  init_gap_table ();

  alignment_count (argv[optind], argv[optind + 1]);

  return 0;
}


/* analyzers */

/* Count base frequencies in SEQS_FILENAME and print a report.  */
void
alignment_count (char *ids_filename,
		 char *seqs_filename)
{
  lenstring ids_text, seqs_text;
  stringlist ids, seqs;
  lenstring mask;
  int mask_row;

  read_file (ids_filename,  &ids_text);
  find_lines (&ids_text, &ids);

  read_file (seqs_filename, &seqs_text);
  find_lines (&seqs_text, &seqs);

  if (ids.len != seqs.len)
    {
      fprintf (stderr, "%s: too %s names in %s for %s\n",
	       progname,
	       (ids.len < seqs.len ? "few" : "many"),
	       ids_filename, seqs_filename);
      exit (2);
    }

  mask_row = find_mask (&ids, &seqs, &mask);
  mask_to_weights (&mask);

  {
    count_table seq_counts;
    count_table seq_counts_p2, seq_counts_p3;
    int row;

    if (! option_combination)
      printf ("%17s%s%5sA%14sC%14sG%14sU\n",
	      "", 
	      (option_position ? "pos" : "   "),
	      "", "", "", "");

    for (row = 0; row < ids.len; row++)
      /* Don't count the mask.  */
      if (row != mask_row)
	{
	  if (option_position)
	    {
	      clear_counts (seq_counts);
	      clear_counts (seq_counts_p2);
	      clear_counts (seq_counts_p3);

	      sequence_count_3p (&seqs.strings[row], &mask,
				 seq_counts,
				 seq_counts_p2,
				 seq_counts_p3);

	      lower_to_upper (seq_counts, 1);
	      distribute_ambiguities (seq_counts);
	      thymine_to_uracil (seq_counts);

	      lower_to_upper (seq_counts_p2, 1);
	      distribute_ambiguities (seq_counts_p2);
	      thymine_to_uracil (seq_counts_p2);

	      lower_to_upper (seq_counts_p3, 1);
	      distribute_ambiguities (seq_counts_p3);
	      thymine_to_uracil (seq_counts_p3);

	      display_lenstring (&ids.strings[row], 17, stdout);
	      fputs ("p 1 ", stdout);
	      display_counts (seq_counts, "                     ");
	      fputs ("                 p 2 ", stdout);
	      display_counts (seq_counts_p2, "                     ");
	      fputs ("                 p 3 ", stdout);
	      display_counts (seq_counts_p3, "                     ");
	      add_counts (seq_counts, seq_counts_p2);
	      add_counts (seq_counts, seq_counts_p3);
	      fputs ("                 all ", stdout);
	      display_counts (seq_counts, "                     ");
	    }
	  else
	    {
	      clear_counts (seq_counts);
	      sequence_count (&seqs.strings[row], &mask, seq_counts);
	      lower_to_upper (seq_counts, 1);
	      distribute_ambiguities (seq_counts);
	      thymine_to_uracil (seq_counts);

	      display_lenstring (&ids.strings[row], 21, stdout);
	      display_counts (seq_counts, "                     ");
	    }
	}
  }
}


/* Display COUNTS in the format requested by the user.
   Preceed all lines but the first by INDENTATION.  */
void
display_counts (count_table counts, char *indentation)
{
  if (option_combination)
    display_counts_combination (counts, indentation);
  else
    display_counts_short (counts);
}


/* Display COUNTS in the combination format.
   Preceed all lines but the first by INDENTATION.  */
void
display_counts_combination (count_table counts, char *indentation)
{
  int total = 0;

  total = (CHX (counts, 'A')
	   + CHX (counts, 'C')
	   + CHX (counts, 'G')
	   + CHX (counts, 'U'));

  /* See the file named "formats" in the analyses source directory for what
     this is supposed to produce.  */
  fputs ("A: ", stdout);       display_count (CHX (counts, 'A'), total);
  fputs ("  A+C: ", stdout);   display_count ((CHX (counts, 'A')
					       + CHX (counts, 'C')),
					      total);
  fputs ("  A+C+G: ", stdout); display_count ((CHX (counts, 'A')
					       + CHX (counts, 'C')
					       + CHX (counts, 'G')),
					      total);

  printf ("\n%sC: ", indentation); display_count (CHX (counts, 'C'), total);
  fputs ("  A+G: ", stdout);   display_count ((CHX (counts, 'A')
					       + CHX (counts, 'G')),
					      total);
  fputs ("  A+C+U: ", stdout); display_count ((CHX (counts, 'A')
					       + CHX (counts, 'C')
					       + CHX (counts, 'U')),
					      total);
  
  printf ("\n%sG: ", indentation); display_count (CHX (counts, 'G'), total);
  fputs ("  A+U: ", stdout);   display_count ((CHX (counts, 'A')
					       + CHX (counts, 'U')),
					      total);
  fputs ("  A+G+U: ", stdout); display_count ((CHX (counts, 'A')
					       + CHX (counts, 'G')
					       + CHX (counts, 'U')),
					      total);
  
  printf ("\n%sU: ", indentation); display_count (CHX (counts, 'U'), total);
  fputs ("  C+G: ", stdout);   display_count ((CHX (counts, 'C')
					       + CHX (counts, 'G')),
					      total);
  fputs ("  C+G+U: ", stdout); display_count ((CHX (counts, 'C')
					       + CHX (counts, 'G')
					       + CHX (counts, 'U')),
					      total);

  printf ("\n%s   %5d total    C+U: ", indentation, total);
  display_count ((CHX (counts, 'C') + CHX (counts, 'U')), total);
  printf ("\n%s                  G+U: ", indentation);
  display_count ((CHX (counts, 'G') + CHX (counts, 'U')), total);
  putchar ('\n');
  putchar ('\n');
}

void
display_counts_short (count_table counts)
{
  int total = 0;

  total = (CHX (counts, 'A')
	   + CHX (counts, 'C')
	   + CHX (counts, 'G')
	   + CHX (counts, 'U'));

  display_count (CHX (counts, 'A'), total); fputs ("  ", stdout);
  display_count (CHX (counts, 'C'), total); fputs ("  ", stdout);
  display_count (CHX (counts, 'G'), total); fputs ("  ", stdout);
  display_count (CHX (counts, 'U'), total); putchar ('\n');
}


/* Count the base frequencies in SEQUENCE, using WEIGHTS as weights.
   Add the counts into TABLE.  */
void
sequence_count (lenstring *sequence, lenstring *weights, count_table table)
{
  unsigned char *seq_text = (unsigned char *) sequence->text;

  if (weights && weights->len > 0)
    /* Count the weighted portion of the sequence.  */
    {
      unsigned char *weights_text = (unsigned char *) weights->text;
      unsigned char *weights_end  = weights_text + weights->len;

      /* The weights might be longer than the sequence.  */
      if (weights_end > weights_text + sequence->len)
	weights_end = weights_text + sequence->len;

      while (weights_text < weights_end)
	table[*seq_text++] += *weights_text++;
    }
  else
    /* Count across the entire sequence, giving everything a weight of
       one.  */
    {
      unsigned char *seq_end = seq_text + sequence->len;

      while (seq_text < seq_end)
	table[*seq_text++]++;
    }
}


/* Count the base frequencies in SEQUENCE, using separate count tables
   for the three positions.  Use WEIGHTS as weights.  */
void
sequence_count_3p (lenstring *sequence,
		   lenstring *weights,
		   count_table table_1,
		   count_table table_2,
		   count_table table_3)
{
  unsigned char *seq_text = (unsigned char *) sequence->text;
  int *tables[3];
  int position = 0;

  tables[0] = table_1;
  tables[1] = table_2;
  tables[2] = table_3;

  if (weights && weights->len > 0)
    /* Count the weighted portion of the sequence.  */
    {
      unsigned char *weights_text = (unsigned char *) weights->text;
      unsigned char *weights_end  = weights_text + weights->len;

      /* The weights might be longer than the sequence.  */
      if (weights_end > weights_text + sequence->len)
	weights_end = weights_text + sequence->len;

      while (weights_text < weights_end)
	{
	  if (! is_gap (*seq_text))
	    {
	      tables[position++][*seq_text] += *weights_text;
	      if (position >= 3)
		position = 0;
	    }
	  seq_text++;
	  weights_text++;
	}
    }
  else
    /* Count across the entire sequence, giving everything a weight of
       one.  */
    {
      unsigned char *seq_end = seq_text + sequence->len;

      while (seq_text < seq_end)
	{
	  if (! is_gap (*seq_text))
	    {
	      tables[position++][*seq_text]++;
	      if (position >= 3)
		position = 0;
	    }
	  seq_text++;
	}
    }
}



/* utilities */

void
init_gap_table ()
{
  clear_counts (gap_table);

  set_gap ('-');
  set_gap ('~');
  set_gap ('|');
  set_gap ('.');
  set_gap (' ');
}

/* Display a count, in the following format:
   "NNNNN (PP.P%)"
   where the N's are COUNT, right-justified, and the P's are the
   percentage that COUNT is of TOTAL.  If COUNT is 100% of total,
   display:
   "NNNNN(100.0%)"  */
void
display_count (int count, int total)
{
  /* The .9995 is the proper cut-off point for when printf would display
     "100%", assuming it rounds properly.  */
  if (count == 0)
    printf ("    0 ( 0.0%%)");
  else if ((double) count >= .9995 * total)
    printf ("%5d(100.0%%)", count);
  else
    printf ("%5d (%4.1f%%)", count,  100 * ((double) count / total));
}

/* Clear a count table.  */
void
clear_counts (count_table t)
{
  memset (t, 0, sizeof (count_table));
}

/* Add the counts in T2 to the counts already in T1.  */
void
add_counts (count_table t1, count_table t2)
{
  int i;

  for (i = 0; i < NUM_CHARS; i++)
    t1[i] += t2[i];
}

/* Add the thymine count to the uracil count.  */
void
thymine_to_uracil (count_table t)
{
  t[(unsigned) 'u'] += t[(unsigned) 't'];
  t[(unsigned) 'U'] += t[(unsigned) 'T'];
}

/* Distribute the counts for ambiguity codes among their possibilities.  */
void
distribute_ambiguities (count_table t)
{
#define AMB2(amb, real1, real2) \
  (t[(unsigned) real1] += t[(unsigned) amb] / 2, \
   t[(unsigned) real2] += (t[(unsigned) amb] - t[(unsigned) amb] / 2))

  AMB2 ('m', 'a', 'c');
  AMB2 ('r', 'a', 'g');
  AMB2 ('w', 'a', 'u');
  AMB2 ('s', 'c', 'g');
  AMB2 ('y', 'c', 'u');
  AMB2 ('k', 'g', 'u');

#define AMB3(amb, real1, real2, real3) \
  (t[(unsigned) real1] += t[(unsigned) amb] / 3, \
   t[(unsigned) real2] += t[(unsigned) amb] / 3, \
   t[(unsigned) real3] += (t[(unsigned) amb] - 2 * t[(unsigned) amb] / 3))

  AMB3 ('b', 'c', 'g', 'u');
  AMB3 ('d', 'a', 'g', 'u');
  AMB3 ('h', 'a', 'c', 'u');
  AMB3 ('v', 'a', 'c', 'g');

#define AMB4(amb, real1, real2, real3, real4) \
  (t[(unsigned) real1] += t[(unsigned) amb] / 4, \
   t[(unsigned) real2] += t[(unsigned) amb] / 4, \
   t[(unsigned) real3] += t[(unsigned) amb] / 4, \
   t[(unsigned) real4] += (t[(unsigned) amb] - 3 * t[(unsigned) amb] / 4))

  AMB4 ('n', 'a', 'c', 'g', 'u');

  /* Upper-case versions of the above.  */
  AMB2 ('M', 'A', 'C');
  AMB2 ('R', 'A', 'G');
  AMB2 ('W', 'A', 'U');
  AMB2 ('S', 'C', 'G');
  AMB2 ('Y', 'C', 'U');
  AMB2 ('K', 'G', 'U');
  AMB3 ('B', 'C', 'G', 'U');
  AMB3 ('D', 'A', 'G', 'U');
  AMB3 ('H', 'A', 'C', 'U');
  AMB3 ('V', 'A', 'C', 'G');
  AMB4 ('N', 'A', 'C', 'G', 'U');
}

/* Add counts for lower-case bases to their upper-case counterparts, with
   weight W.  */
void
lower_to_upper (count_table t, double w)
{
  int c;

  for (c = 'a'; c <= 'z'; c++)
    {
      int u = (unsigned) toupper (c);
      t[u] += t[c] * w;
    }
}


/* Search IDS for a name starting with "mask," and set MASK to the
   corresponding element of SEQS.  If we find one, return its row
   number; else, return -1.  */
int
find_mask (stringlist *ids, stringlist *seqs, lenstring *mask)
{
  int i;

  for (i = 0; i < ids->len; i++)
    {
      char *p = ids->strings[i].text;
      char *end = p + ids->strings[i].len;

      while (p < end && isspace (*p))
	p++;

      /* ick */
      if (p + 4 <= end
	  && tolower (p[0]) == 'm'
	  && tolower (p[1]) == 'a'
	  && tolower (p[2]) == 's'
	  && tolower (p[3]) == 'k')
	{
	  mask->text = seqs->strings[i].text;
	  mask->len  = seqs->strings[i].len;
	  return i;
	}
    }

  mask->text = 0;
  mask->len = 0;
  return -1;
}


/* Destructively convert MASK to weight values.  Characters which are
   digits are replaced with their values (that is, '5' turns into 5).
   All non-digit characters are replaced with '\0'.  */
void
mask_to_weights (lenstring *mask)
{
  unsigned char *p, *end;

  p = (unsigned char *) mask->text;
  end = p + mask->len;

  while (p < end)
    {
      if (isdigit (*p))
	*p = *p - '0';
      else
	*p = 0;
      p++;
    }
}


/* Set LINES to point to the lines in STRING.  */
void
find_lines (lenstring *string, stringlist *lines)
{
  char *start = string->text;
  char *end   = start + string->len;
  char *p = start;
  int i = 0;

  lines->len = 10;
  lines->strings = (lenstring *) xmalloc (lines->len
					  * sizeof (*lines->strings));

  while (p < end)
    {
      if (i >= lines->len)
	{
	  lines->len *= 2;
	  lines->strings
	    = (lenstring *) xrealloc (lines->strings,
				      lines->len * sizeof (*lines->strings));
	}

      lines->strings[i].text = p;
      p = memchr (p, '\n', end - p);
      if (! p) p = end;
      lines->strings[i].len = p - lines->strings[i].text;
      i++;
      p++;
    }

  lines->len = i;
}

/* Read all of FILENAME, and set *CONTENTS to its contents.  */
void
read_file (char *filename, lenstring *contents)
{
  struct stat buf;
  size_t len;
  FILE *f;

  check_syscall (stat (filename, &buf), filename, "while statting");
  contents->len = buf.st_size;
  contents->text = (char *) xmalloc (contents->len
				     * sizeof (*contents->text));
  f = careful_open (filename, "r", 0);
  /* Well, if it's possible, we certainly won't be I/O bound:  */
  len = fread (contents->text, sizeof (*contents->text), contents->len, f);
  if (ferror (f))
    check_syscall (-1, filename, "while reading");
  contents->len = len;
  fclose (f);
}


/* Check the result of a system call. Print an error message and exit
   if appropriate.
   If CALL is not -1, return CALL.
   If CALL is -1, print an error message of the form

       PROGNAME: FILENAME: OP: ERRORMSG

   where PROGNAME is the value of the global variable `progname',
         the FILENAME: section is omitted if FILENAME is zero,
     and ERRORMSG is the error message provided by perror.
   Exit with a value of 2.  */

int
check_syscall (int call, char *filename, char *op)
{
  if (call != -1)
    return call;

  fprintf (stderr, "%s: %s%s",
	   progname,
	   filename ? filename : "",
	   filename ? ": " : "");
  perror (op);

  exit (2);
}


/* If NAME is null, return DEFALT.
   If NAME is non-null, try to open it with MODE; print an error message
   and exit if this doesn't succeed.  */
FILE *
careful_open (char *name, char *mode, FILE *defalt)
{
  if (name)
    {
      FILE *result = fopen (name, mode);

      if (! result)
	{
	  fprintf (stderr, "%s: ", progname);
	  perror (name);
	  exit (2);
	}

      return result;
    }
  else
    return defalt;
}

void
give_usage ()
{
  fprintf (stderr,
	   "\
Usage: %s IDS SEQUENCES\n\
Count base frequencies in sequence data.\n\
SEQUENCES should be the name of a file containing sequences, separated\n\
by newlines.  IDS should be the name of a file contaning the names of\n\
the corresponding lines from SEQUENCES, also separated by newlines.\n\
",
	   progname);

  exit (1);
}
