/* lenstring.h --- strings with explicit lengths
   Jim Blandy <jimb@gnu.ai.mit.edu> --- July 1994 */

#include <stdio.h>
#include <ctype.h>
#include <string.h>

#include "xmalloc.h"
#include "lenstring.h"

#ifdef __GNUC__
#define INLINE inline
#else
#define INLINE
#endif

/* Copy a block of memory, and initialize a lenstring structure to
   point to it.  */
void
make_lenstring (struct lenstring *lenstring,
                char *text, size_t len)
{
  lenstring->len = 0;
  lenstring->text = "\0<nobody home>";

  append_to_lenstring (lenstring, text, len);
}


/* Copy the LEN bytes at TEXT to the end of LENSTRING; realloc
   LENSTRING if necessary.  */
void
append_to_lenstring (struct lenstring *lenstring,
		     char *text, size_t len)
{
  if (len > 0)
    {
      if (lenstring->len == 0)
	lenstring->text = (char *) xmalloc (len + 1);
      else
	lenstring->text = (char *) xrealloc (lenstring->text,
					     lenstring->len + len + 1);
      memcpy (lenstring->text + lenstring->len, text, len);
      lenstring->len += len;
      lenstring->text[lenstring->len] = '\0';
    }
}


/* Read strings (delimited as described in protocol.texi) from
   SOURCE_FILE, until we find an unescaped newline.

   If we hit EOF before reading anything interesting, return EOF.

   If we read some strings, set *STRINGS_OUT to a malloc'ed array of
   lenstrings containing the strings, and return the number of strings
   we read.  Note that the text for all the lenstrings is actually
   kept in one big malloc'ed block; *TEXT_OUT is set to the address of
   this block.  The caller is responsible for freeing *STRINGS_OUT and
   *TEXT_OUT.  */
int
getstrings (FILE *source_file,
	    struct lenstring **strings_out,
	    char **text_out)
{
  /* This holds the text of the strings.  */
  size_t text_len = 0;
  size_t text_size = 10;
  char *text = (char *) xmalloc (text_size * sizeof (*text));

  /* This holds the strings themselves --- effectively, it divides
     text into a bunch of substrings.  We don't set strings[...].text
     to point into text as we read; that would require us to fix up
     all the pointers in strings every time we realloc text.  Rather,
     we just store the lengths of the strings, and compute all the
     pointers when we're done.  */
  size_t strings_len = 0;
  size_t strings_size = 10;
  struct lenstring *strings 
    = (struct lenstring *) xmalloc (strings_size * sizeof (*strings));

  /* Our lookahead character.  */
  int c = getc (source_file);

  for (;;)
    {
      /* Are we done?  */
      if (c == '\n' || c == EOF)
	{
	  if (strings_len > 0 || c == '\n')
	    {
	      /* Compute the text pointers now, when there's no danger of
		 relocation.  Remember the null characters.  */
	      {
		int i;
		char *p = text;

		for (i = 0; i < strings_len; i++)
		  {
		    strings[i].text = p;
		    p += strings[i].len + 1;
		  }
	      }

	      *strings_out = strings;
	      *text_out    = text;

	      return strings_len;
	    }
	  else
	    {
	      free (strings);
	      free (text);

	      return EOF;
	    }
	}

      /* Have we started a quoted string?  */
      else if (c == '"')
	{
	  int start_pos = text_len;

	  if (strings_len >= strings_size)
	    {
	      strings_size *= 2;
	      strings
		= (struct lenstring *)
		  xrealloc (strings, strings_size * sizeof (*strings));
	    }

	  while ((c = getc (source_file)) != EOF && c != '"')
	    {
	      if (text_len >= text_size)
		{
		  text_size *= 2;
		  text = (char *) xrealloc (text, text_size);
		}

	      if (c == '\\')
		{
		  c = getc (source_file);

		  switch (c)
		    {
		    case 'a': text[text_len++] = 7; break;
		    case 'b': text[text_len++] = 8; break;
		    case 't': text[text_len++] = '\t'; break;
		    case 'n': text[text_len++] = '\n'; break;
		    case 'v': text[text_len++] = '\v'; break;
		    case 'f': text[text_len++] = '\f'; break;
		    case 'r': text[text_len++] = '\r'; break;
		    case 'e': text[text_len++] = '\033'; break;
		    case '\n': break;
		    default: text[text_len++] = c; break;
		    }
		}
	      else
		text[text_len++] = c;
	    }

	  strings[strings_len++].len = text_len - start_pos;

	  /* Push a null onto the end, not included in the length.  */
	  if (text_len >= text_size)
	    {
	      text_size *= 2;
	      text = (char *) xrealloc (text, text_size);
	    }
	  text[text_len++] = '\0';

	  c = getc (source_file);
	}

      /* Have we found a space?  */
      else if (isascii (c) && isspace (c))
	c = getc (source_file);

      /* Or have we just discovered an ordinary token?  */
      else
	{
	  int start_pos = text_len;

	  if (strings_len >= strings_size)
	    {
	      strings_size *= 2;
	      strings
		= (struct lenstring *)
		  xrealloc (strings, strings_size * sizeof (*strings));
	    }

	  do
	    {
	      if (text_len >= text_size)
		{
		  text_size *= 2;
		  text = (char *) xrealloc (text, text_size);
		}

	      text[text_len++] = c;
	    
	      c = getc (source_file);
	    }
	  while (!isascii (c) || !isspace (c));

	  strings[strings_len++].len = text_len - start_pos;

	  /* Push a null onto the end, not included in the length.  */
	  if (text_len >= text_size)
	    {
	      text_size *= 2;
	      text = (char *) xrealloc (text, text_size);
	    }
	  text[text_len++] = '\0';
	}
    }
}


/* Write the LEN bytes at TEXT to OUTPUT_FILE, using the quoting
   syntax described in protocol.texi.  This doesn't really belong
   here, as lenstrings aren't involved, but if you think about this
   module as dealing with fully-transparent strings, it's okay.  */
void
write_quoted (FILE *output_file, const char *text, size_t len)
{
  const char *text_end = text + len;

  putc ('"', output_file);

  while (text < text_end)
    {
      switch (*text)
	{
	case '"': 
	case '\\':
	  putc ('\\', output_file); putc (*text, output_file);
	  break;

	case '\n':
          fputs ("\\n", output_file);
	  break;

	default:
	  putc (*text, output_file);
	  break;
	}

      text++;
    }

  putc ('"', output_file);
}

/* Like write_quoted, but for lenstrings.  */
void
write_quoted_lenstring (FILE *output_file,
			struct lenstring *lenstring)
{
  write_quoted (output_file, lenstring->text, lenstring->len);
}
