/* rdp_format.h --- reading and writing RDP format.
   Jim Blandy <jimb@gnu.ai.mit.edu> --- July 1994 */

#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include "xmalloc.h"
#include "ale_getline.h"
#include "node.h"
#include "rdp_format.h"


/* Dummy nodes for returning error codes.  */

/* Return a pointer to this when a system call returns an error.
   Leave the error code in errno.  */
struct node rdp_syserror;

/* Return a pointer to this if we find a syntax error in the file.  */
struct node rdp_parseerror;


/* reading */

/* Skip the informative text at the start of the file.  We'll assume
   that the first line starting with a digit is a data line.  */
static void
skip_header (FILE *file)
{
  int bol = 1;
  int c;

  for (c = getc (file); c != EOF; c = getc (file))
    {
      if (bol && isascii (c) && isdigit (c))
        {
          ungetc (c, file);
          break;
        }
      else if (c == '\n')
        bol = 1;
      else
        bol = 0;
    }
}


/* Build an organism from the text LINE[0..LINE_LEN].  */
static struct node *
extract_organism (char *line, size_t line_len)
{
  char *end = line + line_len;
  char *short_id;
  size_t short_id_len;
  char *description;
  size_t description_len;

  /* We may have an organism number in [brackets] before the short ID.
     Skip any clump of non-whitespace starting with a [bracket.  */
  if (*line == '[')
    {
      while (line < end
	     && ! (isascii (*line)
		   && isspace (*line)))
	line++;
      while (line < end && isascii (*line) && isspace (*line))
	line++;
    }

  /* The next block of non-white-space is the short ID.  */
  short_id = line;

  while (line < end && (!isascii (*line) || !isspace (*line)))
    line++;
  short_id_len = line - short_id;

  /* Skip spaces.  */
  while (line < end && isascii (*line) && isspace (*line))
    line++;
    
  /* Everything from here to the end of the line is the description.  */
  description = line;
  description_len = end - description;

  return new_node (short_id, short_id_len, description, description_len);
}


/* Build a taxon node from the text LINE[0..LINE_LEN].
   Set its depth from the numbers at the start of the line.  */
static struct node *
extract_taxon (char *line, size_t line_len)
{
  char *line_end = line + line_len;
  char *num_start, *num_end;
  char *description_start, *description_end;
  int depth;

  /* We use the numbers as the "name" of the node, and the text as the
     description, since the text isn't unique (and a lot of them have
     no names at all).  */
  num_start = line;
  depth = 1;
  while (line < line_end
	 && ((isascii (*line) && isdigit (*line))
	     || *line == '.'))
    {
      if (*line == '.')
	depth++;
      line++;
    }
  num_end = line;

  /* Skip the space separating the numbers from the text.  */
  while (line < line_end && isascii (*line) && isspace (*line))
    line++;

  /* The entire rest of the line is the text.  */
  description_start = line;
  description_end = line_end;

  {
    struct node *new
      = new_node (num_start, num_end - num_start,
		  description_start, description_end - description_start);

    new->depth = depth;

    /* They tend to use underscores instead of spaces in taxon names,
       although I don't know why.  Spaces are more readable.  */
    {
      char *name = new->name.text;
      char *name_end = new->name.text + new->name.len;

      while (name < name_end)
	{
	  if (*name == '_')
	    *name = ' ';
	  name++;
	}
    }

    return new;
  }
}

/* Read a tree from FILENAME, which is assumed to be in RDP tree
   format.  */
struct node *
read_rdp (char *filename)
{
  FILE *file = fopen (filename, "r");
  struct node *root;

  if (! file)
    return &rdp_syserror;

  root = new_node (filename, strlen (filename), "", 0);

  {
    /* The node whose children we're currently reading.  */
    struct node *parent = root;
    /* The level of parent.  */
    int parent_level = 0;

    /* The line of the file we're looking at.  */
    char *line;
    int line_len;
    
    skip_header (file);

    while ((line_len = ale_getline (file, &line)) != EOF)
      {
	char *line_end = line + line_len;
	char *p = line;

	/* Skip any leading whitespace.  */
	while (p < line_end && isascii (*p) && isspace (*p))
	  p++;

	/* Is the line blank?  */
	if (p >= line_end)
	  ;

	/* Is this a taxon line?  */
	else if (isascii (*p) && isdigit (*p))
          {
            struct node *new_taxon = extract_taxon (p, line_end - p);
	    int level; 

            if (new_taxon == &rdp_parseerror)
	      {
		free (line);
		fclose (file);
		return new_taxon;
	      }

	    level = new_taxon->depth;

            /* Use the new node's level to find the appropriate parent
               for it.  This loop will not execute its body at all
               when the new node is a child of the current parent.  */
            while (parent_level > level - 1)
              {
                parent = parent->parent;
                parent_level--;
              }
              
            /* You can't jump down more than one level at a time.  For
               example, you can't have a taxon numbered 1.3.1.1 as a
               direct child of a taxon numbered 1.3; you have to write
               a 1.3.1 first.  */
            if (parent_level < level - 1)
	      {
		free (line);
		fclose (file);
		return &rdp_parseerror;
	      }
            
            add_child (parent, new_taxon);

            /* Now the new taxon is the default parent for subsequent
               lines.  */
            parent = new_taxon;
            parent_level ++;
          }

	/* It must be an organism line.  */
	else
	  {
	    struct node *new_organism = extract_organism (p, line_end - p);

            if (new_organism == &rdp_parseerror)
	      {
		free (line);
		fclose (file);
		return new_organism;
	      }

            add_child (parent, new_organism);
	  }

	free (line);
      }
  }

  fclose (file);
  return root;
}


/* Write TREE to FILENAME in RDP tree format.  */
void write_rdp (char *FILENAME, struct node *TREE);
