/* gdbm.c --- Ale GDBM manipulation program
   Jim Blandy <jimb@gnu.ai.mit.edu> --- October 1994 */

#include <assert.h>
#include <stdio.h>
#include <string.h>
#include <gdbm.h>

#include "ring.h"
#include "getopt.h"
#include "xmalloc.h"
#include "lenstring.h"
#include "hash.h"
#include "gdbmgenes.h"
#include "careful.h"

void give_usage (void);
void give_help (void);
void give_version (void);
void print_and_die (const char *message);
void gdbm_print_and_die (void);
void put (char *keys_filename,
	  char *annotations_filename,
	  char *sequences_filename);
void get (char *keys_filename,
	  char *annotations_filename,
	  char *sequences_filename);
void choose (char *choose_filename,
	     char *keys_filename,
	     char *annotations_filename,
	     char *sequences_filename);
void list (char *keys_filename);
void verify_index (void);

void read_index_from_gdbm (void);
void write_index_to_gdbm (void);
void write_index_to_stream (FILE *);
void make_index_hash (void);
void index_input_key (lenstring *);
void rebuild_index_string (void);

/* The name of the program, for use in error messages.  */
char *progname;

/* The name of the GDBM file we're operating on, for error messages.  */
char *gdbm_filename = 0;

/* The GDBM file we're operating on.  */
GDBM_FILE g;

enum op {
  op_none = 1000,
  op_put, op_get, op_choose, op_list, op_verify_index
};

/* An array of structures describing the arguments accepted by the
   program.  For use by getopt_long.  */
struct option longopts[] = {
  { "put",          no_argument, 0, 'p' },
  { "get",          no_argument, 0, 'g' },
  { "choose",       optional_argument, 0, 'c' },
  { "list",         no_argument, 0, 'l' },
  { "verify-index", no_argument, 0, op_verify_index },

  { "keys",         required_argument, 0, 'k' },
  { "annotations",  required_argument, 0, 'a' },
  { "sequences",    required_argument, 0, 's' },

  { "help",         no_argument, 0, 'h' },
  { "version",      no_argument, 0, 'v' },

  { 0, 0, 0, 0 }
};

int
main (int argc, char **argv)
{
  int arg, longopts_ix;
  enum op op = op_none;
  char *choose_filename = 0;
  char *keys_filename = 0;
  char *annotations_filename = 0;
  char *sequences_filename = 0;

  progname = careful_prog_name (argv[0]);

  while ((arg = getopt_long (argc, argv,
			     "pgc::lk:a:s:hv", longopts, &longopts_ix))
	 != -1)
    switch (arg)
      {
      case 'p': op = op_put;  break;
      case 'g': op = op_get;  break;
      case 'c': op = op_choose; choose_filename = optarg; break;
      case 'l': op = op_list; break;
      case op_verify_index: op = op_verify_index; break;
      case 'k': keys_filename = optarg; break;
      case 'a': annotations_filename = optarg; break;
      case 's': sequences_filename = optarg; break;

      case 'h': give_help (); break;
      case 'v': give_version (); break;
      default: give_usage (); break;
      }

  if (optind + 1 == argc)
    gdbm_filename = argv[optind];

  if (gdbm_filename == 0)
    {
      fprintf (stderr, "%s: no GDBM file specified\n", progname);
      give_usage ();
    }
  if (op == op_none)
    {
      fprintf (stderr, "%s: no operation specified\n", progname);
      give_usage ();
    }

  assert (op == op_put
	  || op == op_get
	  || op == op_choose
	  || op == op_list
	  || op == op_verify_index);

  /* Open the GDBM file.  */
  g = gdbm_open (gdbm_filename, 0,
		 (op == op_put) ? GDBM_WRCREAT 
		 : (op == op_verify_index) ? GDBM_WRCREAT
		 : GDBM_READER,
		 0666, print_and_die);
  if (! g)
    gdbm_print_and_die ();    

  read_index_from_gdbm ();

  if (op == op_put)
    put (keys_filename, annotations_filename, sequences_filename);
  else if (op == op_get)
    get (keys_filename, annotations_filename, sequences_filename);
  else if (op == op_choose)
    choose (choose_filename,
	    keys_filename, annotations_filename, sequences_filename);
  else if (op == op_list)
    list (keys_filename);
  else if (op == op_verify_index)
    verify_index ();

  if (op == op_put)
    write_index_to_gdbm ();

  gdbm_close (g);

  return 0;
}



/* The alignment key index  */

/* The alignment key index has two jobs: to make it fast to list all
   the keys bound in the GDBM file, and to preserve the order of
   entries in the alignment.

   The index appears in the GDBM file under the key 'ialignment' as a
   string of key names, each terminated by a newline.  This means that
   we can implement the `--list' option simply by retrieving the index
   from the database, and spitting it out on a stdio stream.

   In memory, we keep the stringy form of the index around, but we
   also make a hash table mapping keys onto struct key's.  The struct
   key's are elements of one or both of two doubly-linked rings ---
   one for keys already in the database ("extant" keys), and one for
   keys in the input ("input" keys).  If we're not putting sequences
   in the database, the input ring could be empty.  If we're creating
   a new database, the extant ring could be empty.  Every struct key
   is a member of one ring or the other.

   When doing the `--put' operation, the set of input keys has no
   definite relationship with the database's set of extant keys.  If
   we're adding new sequences to an existing database, the sets of
   extant and input keys might have no keys in common.  If we're
   writing back an entire alignment after adding some new sequences,
   the extant keys could be a subset of the input keys.  If we're
   writing back a subalignment, the input keys could be a subset of
   the extant keys.  Etcetera.

   Whatever their relationship is, the final alignment, containing all
   the keys, input and extant, should always preserve the ordering of
   the input keys.  That's what the user last saw, so it's reasonable
   to construe it as a specification.  It should also preserve as much
   of the extant keys' ordering as possible.  (Well, as much as
   possible without too much work; I don't want to think about what
   constitutes a "minimum disruption of an ordering"...)


   Here's what we'll do:

   Let I be the list of input keys.  
   Let E be the list of extant keys.
   Let C be the intersection of I and E (C stands for "common"); its
      elements appear in the order they have in I.

   Divide E into sublists, breaking before each element of C.  If E
      starts with an element of C, produce an empty list at the head.
   Notice that each sublist (except the first) starts with an element
      of C, and the cdr contains only elements from E-C.

   Let (follow  k C E) be the cdr of the sublist that starts with k.
   Let (initial k C E) be the first (possibly empty) sublist.

   You can guess what (follow k C I) and (initial k C I) are; they're
   sections of I, broken up in the same way.

   The order of the final unioned list U is:
     (apply append (initial k C I) (initial k C E)
	    (mapcar (lambda (k) (append (list k)
					(follow k C I)
					(follow k C E)))
		    C))
   
   Note that if j precedes k in I, j preceeds k in U as well, i.e. the
   input ordering is preserved.  */
 
/* The index, as read from the database: a string of keys separated by
   newlines.  */
lenstring index_string;

struct key {
  /* The name of this key, shared with the hash table.  We leave this
     uninitialized until before we rebuild index_string from the hash
     table and rings.  */
  lenstring name;

  /* Some keys will be in both the extant ring and the input ring.
     All keys will be in one or the other.  If a key is not in a ring,
     the key's links for that ring form a singleton.  */

  /* Links for the extant keys.  */
  struct key *extant_prev, *extant_next;

  /* Links for the input keys.  */
  struct key *input_prev, *input_next;
};

/* The head of the input and extant rings.  */
struct key index_ring;

/* A hash table mapping keys onto struct key's.  */
struct string_hash *index_hash;

#define LOOKUP_KEY(table, string, string_len) \
  ((struct key **) lookup_hash_table ((table), (string), (string_len)))
#define LOOKUP_KEY_SOFT(table, string, string_len) \
  ((struct key **) lookup_hash_table_soft ((table), (string), (string_len)))


/* Allocate and initialize a fresh struct key.  */
struct key *
make_key ()
{
  struct key *k = (struct key *) xmalloc (sizeof (*k));

  k->name.len = 0;
  RING_SINGLETON (k, extant);
  RING_SINGLETON (k, input);

  return k;
}


/* The key under which the index appears in the database.  */
#define INDEX_KEY "ialignment"
datum index_key = { INDEX_KEY, sizeof (INDEX_KEY) - 1 };


/* Set up index_string, by reading the index from the database.  */
void
read_index_from_gdbm ()
{
  datum index;

  index = gdbm_fetch (g, index_key);

  index_string.text = index.dptr;
  index_string.len  = index.dptr ? index.dsize : 0;
}


/* Write index_string back to the GDBM file.  */
void
write_index_to_gdbm ()
{
  datum index;

  index.dptr = index_string.text;
  index.dsize = index_string.len;

  if (gdbm_store (g, index_key, index, GDBM_REPLACE))
    gdbm_print_and_die ();
}

void
write_index_to_stream (FILE *file)
{
  write_lenstring (&index_string, file);
}


/* Construct index_hash and index_ring from index_string.  */
void
make_index_hash ()
{
  if (index_hash)
    return;

  index_hash = new_hash_table ();
  
  /* index_ring serves as the head for the extant and input rings.  */
  RING_SINGLETON (&index_ring, extant);
  RING_SINGLETON (&index_ring, input);

  {
    char *p = index_string.text;
    char *end = index_string.text + index_string.len;

    while (p < end)
      {
	char *key_end;

	/* Find the newline terminating this key.  */
	for (key_end = p; key_end < end && *key_end != '\n'; key_end++)
	  ;

	if (p < key_end)
	  {
	    struct key **cell_ptr = LOOKUP_KEY (index_hash, p, key_end - p);

	    if (! *cell_ptr)
	      {
		*cell_ptr = make_key ();
		RING_INSERT_BEFORE (*cell_ptr, &index_ring, extant);
	      }
	  }

	p = key_end + 1;
      }
  }
}


/* Add KEY to the list of input keys.
   This function updates index_hash and index_ring, but not
   index_string.  */
void
index_input_key (lenstring *key)
{
  struct key **cell_ptr = LOOKUP_KEY (index_hash, key->text, key->len);

  /* Make sure we have a struct key allocated for this one.  */
  if (! *cell_ptr)
    *cell_ptr = make_key ();

  /* If *cell_ptr is a new key, it'll be a singleton.  */
  if (RING_IS_SINGLETON (*cell_ptr, input))
    RING_INSERT_BEFORE (*cell_ptr, &index_ring, input);
}


static size_t total_length;

/* Function for internal use by rebuild_index_string.  */
static void
hashmap_lengths_and_names (char *string, size_t string_len, void **cell,
			   void *dummy)
{
  struct key *key = *(struct key **) cell;

  total_length += string_len + 1;
  key->name.text = string;
  key->name.len  = string_len;
}


/* Rebuild index_string from index_hash and index_ring.  */
void
rebuild_index_string ()
{
  /* Compute the total length of the list, and set the names of the keys.  */
  total_length = 0;
  map_hash_table (index_hash, hashmap_lengths_and_names, 0);

  /* Rebuild index_string.  */
  index_string.text = xrealloc (index_string.text, total_length);
  index_string.len  = 0;
  {
    struct key *ke, *ki;

    /* First we copy the input keys that come before any extant keys.  */
    for (ki = index_ring.input_next;
	 ki != &index_ring && RING_IS_SINGLETON (ki, extant);
	 ki = ki->input_next)
      {
	append_lenstring (&index_string, &ki->name);
	APPEND_CHAR (&index_string, '\n');
      }

    /* Then the extant keys that come before any input keys.  */
    for (ke = index_ring.extant_next;
	 ke != &index_ring && RING_IS_SINGLETON (ke, input);
	 ke = ke->extant_next)
      {
	append_lenstring (&index_string, &ke->name);
	APPEND_CHAR (&index_string, '\n');
      }

    /* Now both ki points to a key which is both extant and input, or
       to the end of the ring.

       If it points to the end of the ring, then there are no keys
       which are both extant and input; the two loops above copied all
       the keys in both rings to index_string, so there's no work to
       be done.

       Otherwise, it points to a key which is both extant and input.
       Copy that key, then copy all the extant-only keys after it in
       the extant ring, and then all the input-only keys after it in
       the input ring.  This will leave us in the same state we were
       at the top of the loop --- either pointing at a extant and
       input key, or the end of the input ring.  */
    while (ki != &index_ring)
      {
	/* Copy this key.  */
	append_lenstring (&index_string, &ki->name);
	APPEND_CHAR (&index_string, '\n');

	ke = ki->extant_next;

	/* Copy subsequent input-only keys.  */
	for (ki = ki->input_next;
	     ki != &index_ring && RING_IS_SINGLETON (ki, extant);
	     ki = ki->input_next)
	  {
	    append_lenstring (&index_string, &ki->name);
	    APPEND_CHAR (&index_string, '\n');
	  }

	/* Copy the extant-only keys after the common key.  */
	for (;
	     ke != &index_ring && RING_IS_SINGLETON (ke, input);
	     ke = ke->extant_next)
	  {
	    append_lenstring (&index_string, &ke->name);
	    APPEND_CHAR (&index_string, '\n');
	  }
      }

    /* Now all the keys should have been copied.  */
    assert (index_string.len == total_length);
  }
}



/* Putting entries in a GDBM file.  */

void
put (char *keys_filename,
     char *annotations_filename,
     char *sequences_filename)
{
  FILE *keys_file = careful_open (keys_filename, "r", stdin);
  FILE *annotations_file = careful_open (annotations_filename, "r", stdin);
  FILE *sequences_file = careful_open (sequences_filename, "r", stdin);

  lenstring key, annotation, sequence;

  datum gdbm_key;
  size_t gdbm_key_allocated;
  datum gdbm_value;
  size_t gdbm_value_allocated;

  gdbm_key_allocated = 0;
  gdbm_key.dptr = 0;

  gdbm_value_allocated = 0;
  gdbm_value.dptr = 0;

  make_index_hash ();

  for (;;)
    {
      {
	int key_eof =
	  (read_delimited_lenstring (&key,        "\n", keys_file)
	   == EOF);
	int ann_eof =
	  (read_delimited_lenstring (&annotation, "\f", annotations_file)
	   == EOF);
	int seq_eof =
	  (read_delimited_lenstring (&sequence,   "\n", sequences_file)
	   == EOF);

	if (key_eof || ann_eof || seq_eof)
	  break;
      }

      /* Construct a key string, with the appropriate prefix character.  */
      if (key.len + 1 > gdbm_key_allocated)
	{
	  gdbm_key_allocated = key.len + 40;
	  gdbm_key.dptr
	    = (char *) xrealloc (gdbm_key.dptr, gdbm_key_allocated);
	}
      gdbm_key.dsize = key.len + 1;
      gdbm_key.dptr[0] = GDBM_TABL_ENTRY;
      memcpy (gdbm_key.dptr + 1, key.text, key.len);

      /* Construct the value.  */
      {
	size_t size = (sizeof (struct tabl_entry)
		       + annotation.len
		       + sequence.len);
	
	if (size > gdbm_value_allocated)
	  {
	    gdbm_value_allocated = size + 100;
	    gdbm_value.dptr
	      = (char *) xrealloc (gdbm_value.dptr, gdbm_value_allocated);
	  }
	gdbm_value.dsize = size;
	
	{
	  struct tabl_entry *e = (struct tabl_entry *) gdbm_value.dptr;

	  e->ann_len = annotation.len;
	  e->seq_len = sequence.len;
	  memcpy (e->data, annotation.text, annotation.len);
	  memcpy (e->data + e->ann_len, sequence.text, sequence.len);
	}
      }

      /* Plop the value into the database.  */
      if (gdbm_store (g, gdbm_key, gdbm_value, GDBM_REPLACE))
	gdbm_print_and_die ();

      /* Add an entry to the index.  */
      index_input_key (&key);

      free (key.text);
      free (annotation.text);
      free (sequence.text);
    }

  if (!feof (keys_file) || !feof (annotations_file) || !feof (sequences_file))
    print_and_die ("Hallelujah!  You don't have the same number of keys,\n"
		   "    annotations, and sequences.\n");

  if (gdbm_key.dptr) free (gdbm_key.dptr);
  if (gdbm_value.dptr) free (gdbm_value.dptr);

  careful_close (keys_file, keys_filename);
  careful_close (annotations_file, annotations_filename);
  careful_close (sequences_file, sequences_filename);

  rebuild_index_string ();
}


/* Utilities for retrieving and writing entries.  */

/* Return the datum stored under the tabl_entry key for
   KEY[0..KEY_LEN - 1].  */
datum
get_tabl_entry (char *key, size_t key_len)
{
  datum key_datum, value_datum;

  key_datum.dsize = key_len + 1;
  key_datum.dptr  = (char *) xmalloc (key_datum.dsize
				      * sizeof (*key_datum.dptr));

  key_datum.dptr[0] = GDBM_TABL_ENTRY;
  memcpy (key_datum.dptr + 1, key, key_len);

  value_datum = gdbm_fetch (g, key_datum);
  free (key_datum.dptr);
  return value_datum;
}


/* Treating datum D as a struct tabl_entry and KEY[0 .. KEY_LEN - 1]
   as the key it is stored under, write the appropriate things to
   KEY_FILE, ANNOTATION_FILE, and SEQUENCE_FILE.  Make no assumptions
   about the validity of D; after all, we have no idea who created
   this database.  */
void
write_tabl_entry (datum d,
		  char *key, size_t key_len,
		  FILE *key_file, char *key_filename,
		  FILE *annotation_file, char *annotation_filename,
		  FILE *sequence_file, char *sequence_filename)
{
  /* If d.dptr is zero, this probably means have an index entry but no
     database entry.  Just print a blank annotation and sequence --- the
     user will figure out what's going on.  */
  if (!d.dptr)
    {
      fwrite (key, key_len, sizeof (*key), key_file);
      putc ('\n', key_file);

      fputs ("\n\f", annotation_file);

      putc ('\n', sequence_file);
    }
  else
    {
      /* What's the maximum reasonable length of the data in the struct
	 tabl_entry?  */
      size_t max_length = d.dsize - sizeof (struct tabl_entry);
      struct tabl_entry *e = (struct tabl_entry *) d.dptr;

      if (e->ann_len > max_length)
	e->ann_len = max_length;
      if (e->seq_len > max_length - e->ann_len)
	e->seq_len = max_length - e->ann_len;

      fwrite (key, key_len, sizeof (*key), key_file);
      putc ('\n', key_file);

      fwrite (e->data, e->ann_len, sizeof (*e->data), annotation_file);
      putc ('\f', annotation_file);
	      
      fwrite (e->data + e->ann_len, e->seq_len, sizeof (*e->data),
	      sequence_file);
      putc ('\n', sequence_file);
    }

  check_file (key_file, key_filename,
	      "writing GDBM keys");
  check_file (annotation_file, annotation_filename, 
	      "writing GDBM annotations");
  check_file (sequence_file, sequence_filename,
	      "writing GDBM sequences");
}



/* Retrieving all entries from the database.  */

void
get (char *keys_filename,
     char *annotations_filename,
     char *sequences_filename)
{
  FILE *key_file        = careful_open (keys_filename,        "w", stdout);
  FILE *annotation_file = careful_open (annotations_filename, "w", stdout);
  FILE *sequence_file   = careful_open (sequences_filename,   "w", stdout);

  char *key_start = index_string.text;
  char *index_end = index_string.text + index_string.len;
  
  /* Retrieve the entries in the order they appear in the index.  */
  while (key_start < index_end)
    {
      char *key_end;

      for (key_end = key_start;
	   key_end < index_end && *key_end != '\n'; 
	   key_end++)
	;

      if (key_start < key_end)
	{
	  /* Retrieve the value.  */
	  datum gdbm_value
	    = get_tabl_entry (key_start, key_end - key_start);

	  write_tabl_entry (gdbm_value,
			    key_start, key_end - key_start,
			    key_file, keys_filename,
			    annotation_file, annotations_filename,
			    sequence_file, sequences_filename);
	  if (gdbm_value.dptr)
	    free (gdbm_value.dptr);
	}

      key_start = key_end + 1;
    }

  careful_close (key_file, keys_filename);
  careful_close (annotation_file, annotations_filename);
  careful_close (sequence_file, sequences_filename);
}


/* Retrieving selected entries from the database.  */

void
choose (char *choose_filename,
	char *keys_filename,
	char *annotations_filename,
	char *sequences_filename)
{
  FILE *choose_file     = careful_open (choose_filename,      "r", stdin);
  FILE *key_file        = careful_open (keys_filename,        "w", stdout);
  FILE *annotation_file = careful_open (annotations_filename, "w", stdout);
  FILE *sequence_file   = careful_open (sequences_filename,   "w", stdout);

  lenstring key;

  while (read_delimited_lenstring (&key, "\n", choose_file) != EOF)
    {
      datum value = get_tabl_entry (key.text, key.len);

      write_tabl_entry (value,
			key.text, key.len,
			key_file, keys_filename,
			annotation_file, annotations_filename,
			sequence_file, sequences_filename);

      if (value.dptr)
	free (value.dptr);
      free (key.text);
    }

  careful_close (choose_file, choose_filename);
  careful_close (key_file, keys_filename);
  careful_close (annotation_file, annotations_filename);
  careful_close (sequence_file, sequences_filename);
}


/* List all the entries in the database.  */

void
list (char *keys_filename)
{
  FILE *keys_file = careful_open (keys_filename, "w", stdout);

  write_index_to_stream (keys_file);

  careful_close (keys_file, keys_filename);
}


/* Verifying that the index lists all the relevant entries in the
   database.  */

/* Traverse the database's hash table, comparing entries against those
   in our hash table, built from the index.  Print error messages
   about keys which don't appear in both places.  */
void
verify_index ()
{
  int repairs_needed = 0;
  
  make_index_hash ();

  /* For each key KEY recorded in the database's hash table, add it to
     the input ring.  This isn't really what the input ring is
     supposed to be for, but it's convenient.  */
  {
    datum key = gdbm_firstkey (g);

    while (key.dptr)
      {
	/* If this is a tabl-entry key, add it to the input ring.  */
	if (key.dsize >= 1
	    && key.dptr[0] == GDBM_TABL_ENTRY)
	  {
	    lenstring k;

	    k.text = key.dptr + 1;
	    k.len  = key.dsize - 1;

	    index_input_key (&k);
	  }

	/* Find the next key.  */
	{
	  datum next_key;

	  next_key = gdbm_nextkey (g, key);
	  free (key.dptr);
	  key = next_key;
	}
      }
  }

  /* Make sure all struct key's have their names set correctly.  */
  map_hash_table (index_hash, hashmap_lengths_and_names, 0);

  /* Walk the input ring, looking for keys which are not also in the
     extant ring.  Such keys are in the database's hash table, but not
     in our index.  */
  {
    struct key *k;

    for (k = index_ring.input_next; k != &index_ring; k = k->input_next)
      if (RING_IS_SINGLETON (k, extant))
	{
	  fprintf (stderr, "%s: key `", progname);
	  write_lenstring (&k->name, stderr);
	  fprintf (stderr, "' is indexed, but not in database\n");

	  repairs_needed = 1;
	}
  }

  /* Walk the extant ring, looking for keys which are not also in the
     input ring.  Such keys are in the index, but not in the
     database's hash table.  */
  {
    struct key *k;

    for (k = index_ring.extant_next; k != &index_ring; k = k->extant_next)
      if (RING_IS_SINGLETON (k, input))
	{
	  fprintf (stderr, "%s: key `", progname);
	  write_lenstring (&k->name, stderr);
	  fprintf (stderr, "' is in the database, but not indexed.\n");

	  repairs_needed = 1;
	}
  }

  /* If any repairs are needed, then that probably indicates a bug in
     this program.  Including code to work around bugs is no
     substitute for fixing the bugs.  However, the biologist would
     like to get some work done, even if there are bugs in the code;
     if we can recover some of her sequences, that would be nice.  */
  if (repairs_needed)
    {
      struct key *k;

      fprintf (stderr, "%s: rebuilding index...\n", progname);

      /* Delete any extant keys which are not also input keys.  */
      for (k = index_ring.extant_next; k != &index_ring; k = k->extant_next)
	if (RING_IS_SINGLETON (k, input))
	  {
	    /* Remove it from the extant ring.  */
	    RING_REMOVE (k, extant);

	    /* Take the key out of the hash table too.  The hash table
	       module doesn't really export a function do to this, but
	       we can fake it by simply setting the key's name's
	       cell_ptr to zero.  */
	    *LOOKUP_KEY (index_hash, k->name.text, k->name.len) = 0;

	    /* Since the hash table entry is still really there, we
	       shouldn't free k's name.  */
	    free (k);
	  }

      /* Re-create the index, and write it out.  */
      rebuild_index_string ();
      write_index_to_gdbm ();

      fprintf (stderr, "%s: index rebuilt\n", progname);
    }
}



/* Usage messages, dying, general utilities  */

void
give_usage ()
{
  static const char usage[] = "\
GDBM access program for the Ale gene editor.\n\
Usage: %s GDBM-FILE\n\
    [--put] [-p]\n\
    [--get] [-g]\n\
    [--choose[=CHOOSE-FILE]] [-c[=CHOOSE-FILE]]\n\
    [--list] [-l]\n\
    [--verify-index]\n\
\n\
    [-k KEY-FILE]        [--keys KEY-FILE]\n\
    [-a ANNOTATION-FILE] [--annotations ANNOTATION-FILE]\n\
    [-s SEQUENCE-FILE]   [--sequences SEQUENCE-FILE]\n\
\n\
    [--help] [-h]\n\
    [--version] [-v]\n\
";

  fprintf (stderr, usage, progname);
  exit (1);
}

void
give_help ()
{
  static const char help[] = "\
GDBM access program for the Ale gene editor.\n\
\n\
Usage: %s GDBM-FILE\n\
    [--put] [-p]\n\
    [--get] [-g]\n\
    [--choose[=CHOOSE-FILE]] [-c[=CHOOSE-FILE]]\n\
    [--list] [-l]\n\
    [--verify-index]\n\
\n\
    [-k KEY-FILE]        [--keys KEY-FILE]\n\
    [-a ANNOTATION-FILE] [--annotations ANNOTATION-FILE]\n\
    [-s SEQUENCE-FILE]   [--sequences SEQUENCE-FILE]\n\
\n\
    [--help] [-h]\n\
    [--version] [-v]\n\
\n\
  -p, --put\n\
        Add sequences to GDBM-FILE.  Read keys, annotations, and\n\
        sequences from KEY-FILE, ANNOTATION-FILE, and SEQUENCE-FILE.\n\
  -g, --get\n\
        Get all the sequences from GDBM-FILE.  Write keys,\n\
        annotations, and sequences to KEY-FILE, ANNOTATION-FILE, and\n\
        SEQUENCE-FILE.\n\
  -c, --choose[=CHOOSE-FILE]\n\
        Get selected sequences from GDBM-FILE.  Read keys from\n\
        CHOOSE-FILE, and write the information filed under that key to\n\
        KEY-FILE, ANNOTATION-FILE, and SEQUENCE-FILE.  (KEY-FILE will\n\
        usually be the same as CHOOSE-FILE.)  If CHOOSE-FILE is\n\
        omitted, read keys from stdin.\n\
  -l, --list\n\
        List the keys of all the sequences present in GDBM-FILE to\n\
        KEY-FILE.  Since the program maintains an index of all the\n\
        keys present, this ought to be faster than using the --get-all\n\
        option and sending the annotations and sequences to /dev/null.\n\
  --verify-index\n\
        Compare the index of sequences used by `--list' against those\n\
        actually present in GDBM-FILE.  If we find any mismatches,\n\
        list the keys of the sequences involved, and update the index\n\
        to reflect those found in GDBM-FILE.\n\
\n\
Keys in KEY-FILE and CHOOSE-FILE are terminated by newlines.\n\
Annotations in ANNOTATION-FILE are terminated by form feeds.\n\
Sequences in SEQUENCE-FILE are terminated by newlines.\n\
\n\
CHOOSE-FILE, KEY-FILE, ANNOTATION-FILE, and SEQUENCE-FILE all default\n\
to the standard input or output, as appropriate.\n\
\n\
If more than one of those is coming from or going to the standard\n\
input, the key comes first, then the annotation, and lastly the\n\
sequence.\n\
\n\
  -h, --help\n\
        Display this message.\n\
  -v, --version\n\
        Display the version number of this program.\n\
";

  printf (help, progname);
  exit (1);
}

void
give_version ()
{
  fputs ("Ale readers version " VERSION "\n", stderr);
  fputs ("RCS revision: $Id: gdbm.c,v 1.3 1995/04/30 20:49:39 jimb Exp $\n", stderr);

  exit (1);
}

void
print_and_die (const char *message)
{
  fprintf (stderr, "%s: %s\n", progname, message);

  if (g) gdbm_close (g);

  exit (2);
}

void
gdbm_print_and_die ()
{
  fprintf (stderr, "%s: %s", progname, gdbm_strerror (gdbm_errno));

  switch (gdbm_errno)
    {
    case GDBM_FILE_OPEN_ERROR:
    case GDBM_FILE_WRITE_ERROR:
    case GDBM_FILE_SEEK_ERROR:
    case GDBM_FILE_READ_ERROR:
      fputs (": ", stderr);
      perror (gdbm_filename);
      break;
    default:
      putc ('\n', stderr);
      break;
    }

  if (g) gdbm_close (g);

  exit (2);
}
