;; gene-ann-mode.el (part of ale).
;; Copyright (C) 1993, 1994  Karl Fogel <kfogel@darwin.life.uiuc.edu>
;;
;; This software is copyrighted under the terms of the GNU General
;; Public License (see terms below).  The copyright covers this file,
;; specifically, and the program of which it is a part.
;; 
;;      This program is free software; you can redistribute it and/or modify
;;      it under the terms of the GNU General Public License as published by
;;      the Free Software Foundation; either version 1, or (at your option)
;;      any later version.
;; 
;;      This program is distributed in the hope that it will be useful,
;;      but WITHOUT ANY WARRANTY; without even the implied warranty of
;;      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;      GNU General Public License for more details.
;; 
;;      You should have received a copy of the GNU General Public License
;;      along with this program; if not, write to the Free Software
;;      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

(require 'cl)

(require 'gene-mode-tools)

(defvar gene-ann-frame nil
  "The frame in which annotations are displayed.
This is an internal variable; users should keep their grubby little
paws^H^H^H^H
I mean, er, users should have no need to set this variable.")


(defun gene-ann-show-frame ()
  "Show the Annotation Frame.  De-iconify and raise it."
  (interactive)
  (setq gene-ann-frame
        (gene-frame-show gene-ann-buffer
                         `((unsplittable     . t)
                           (menu-bar-lines   . 1)
                           (foreground-color . ,gene-foreground-color)
                           (background-color . ,gene-background-color)
                           (cursor-color     . ,gene-cursor-color)
                           (visibility       . t))))
  ;; Obessive preparations.
  (raise-frame gene-ann-frame)
  (sit-for 0) ; force Emacs to acknowledge frame's visiblity.
  (set-buffer gene-ann-buffer)
  (set-window-dedicated-p
   (frame-selected-window gene-ann-frame) t)
  (let ((o-frame (selected-frame)))
    (select-frame gene-ann-frame)
    (gene-menu-bar-make-hide-button)
    (toggle-scroll-bar -1)
    (select-frame o-frame))
  ;; paranoia: update it just in case it's inconsistent.
  (gene-ann-update-annotation))


(defun gene-ann-hide-frame ()
  "Hide the annotation frame."
  (interactive)
  (if (frame-live-p gene-ann-frame)
      (make-frame-invisible gene-ann-frame)))


(defun gene-ann-select-sequence-window ()
  "Switch back to the sequence window from the annotation window.
If annotations are in their own frame, this does nothing that you can
see."
  (interactive)
  (if (eq (current-buffer) gene-ann-buffer)
      (select-window (gene-seq-buffer-window))))


(defvar gene-ann-last-source nil
  "The last value of the `gene-source' text property, so we know
whether the annotation has changed.")


(defun gene-ann-update-annotation-hook-function ()
  "Update the annotation frame according to point motion.
This function belongs in post-command-hook."
  (and (eq (current-buffer) gene-seq-buffer)
       (frame-live-p gene-ann-frame)
       (frame-visible-p gene-ann-frame)
       (gene-ann-update-annotation)
       ))


(defun gene-ann-update-annotation ()
  "Update the annotations according to point."
  (save-excursion
    (set-buffer gene-seq-buffer)
    (let ((source (gene-get-here-source)))
      (and

       ;; Make sure there *is* a source here!
       source

       ;; No source change, so don't update anything
       (not (eq gene-ann-last-source source))
       
       ;; Else source change, so do expensive update
       (let* ((ann-pair (gene-source-annotation source))
              (start (car ann-pair))
              (end (cadr ann-pair)))
	 (setq gene-ann-last-source source)
	 (set-buffer gene-ann-buffer)

	 ;; The end marker doesn't float correctly when you insert
	 ;; text at the end of the annotation; it stays before the new
	 ;; text, as markers will.  Thus, we'll ignore the end marker
	 ;; here, and use the form feeds to find the end of the
	 ;; annotation.
	 (widen)
	 (goto-char start)
	 (narrow-to-region (point) (gene-end-of-page-point))

	 (if (frame-live-p gene-ann-frame)
             (set-window-point
              (frame-selected-window gene-ann-frame)
              (point-min))))))))

(add-hook 'post-command-hook 'gene-ann-update-annotation-hook-function)

;;; Functions which operate on annotations.

;; Charlie wanted to classify sequences by annotation searches.
;; Here's a way...
(defun gene-ann-match-string (re)
  "Return a list of sequences whose annotations contain a match for RE."
  (save-excursion
    (set-buffer gene-ann-buffer)
    (save-restriction
      (widen)
      ;; Now here's an occasion when one really could use 1 instead of
      ;; point-min... but somehow I just can't bring myself to do it.
      (goto-char (point-min))
      (save-match-data
        (let (matches)
          (while (re-search-forward re nil t)
            (let ((sym (gene-get-here-sym)))
              (or
               ;; Grin.  We don't have to ask memq, because if the
               ;; sym is not the car of matches, then it must not be
               ;; in matches at all.
               ;; UnGrin.  Memq on the usual list is actually probably
               ;; faster than asking (eq sym (car matches)) because of
               ;; the extra function call overhead in this language.
               ;; Sigh.
               ;; You take the High Road -- I'll take Emacs.
               (eq sym (car matches))
               (setq matches (cons sym matches)))))
          matches)))))


(defun gene-ann-annotations-containing-string (str)
  "Return a list of sequences whose annotations match STR."
  (interactive "sAnnotations matching string: ")
  (gene-ann-match-string (regexp-quote str)))


(defun gene-ann-search-and-build-group (re &optional sgroup)
  "Make a group of all sequences whose annotations contain a match for RE.
Optional second arg SGROUP is a group to use, or a group name.
If it's a name, the group will be created if it doesn't already
exist."
  (interactive "sSearch annotations: ")
  (let ((sym-list (gene-ann-annotations-containing-string re))
        (str
         (concat "ANN: " (substring re 0 (min (length re) 7)))))
    (if sym-list
        (progn
          (gene-sgroup-make-group-from-sequences sym-list str)
          (gene-sgroup-point-to-group (gene-sgroup-get-group-by-name str))
          (message "Matched %d sequences." (length sym-list)))
      (message "No matches for \"%s\"." re))))



(defvar gene-ann-mode-map (make-sparse-keymap))

(defvar gene-ann-mode-hook nil
  "*Hook run when gene-ann-mode is called.")


(defun gene-ann-mark-modified ()
  "Set the visible modification flag in the alignment buffer."
  (gene-save-excursion-in
      gene-seq-buffer
    (gene-mark-modified)))


(defun gene-ann-mode ()
  "Major mode for annotating gene sequences.
\\{gene-ann-mode-map}"
  (kill-all-local-variables)
  (use-local-map gene-ann-mode-map)

  ;; Setting the page delimiter to this, instead of "^\C-l", lets us
  ;; process null annotations, which some formats (like Phylip)
  ;; generate.
  (make-local-variable 'page-delimiter)
  (setq page-delimiter "\C-l")

  (make-local-variable 'first-change-hook)
  (setq first-change-hook 'gene-ann-mark-modified)

  (setq mode-line-format "Annotations")

  (setq major-mode 'gene-ann-mode)
  (setq mode-name "Gene Sequence Annotation")
  (run-hooks 'gene-ann-mode-hook 'gene-ann-mode-hooks))

(provide 'gene-ann-mode)


