;; gene-clipboard.el (part of ale).
;; Copyright (C) 1994  Karl Fogel <kfogel@darwin.life.uiuc.edu>
;;
;; This software is copyrighted under the terms of the GNU General
;; Public License (see terms below).  The copyright covers this file,
;; specifically, and the program of which it is a part.
;; 
;;      This program is free software; you can redistribute it and/or modify
;;      it under the terms of the GNU General Public License as published by
;;      the Free Software Foundation; either version 1, or (at your option)
;;      any later version.
;; 
;;      This program is distributed in the hope that it will be useful,
;;      but WITHOUT ANY WARRANTY; without even the implied warranty of
;;      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;      GNU General Public License for more details.
;; 
;;      You should have received a copy of the GNU General Public License
;;      along with this program; if not, write to the Free Software
;;      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
;;
;; Search for instances of "todo:" to find notes about future plans.

(require 'gene-mode-tools)

;;; Cutting, Copying, and Pasting of sequences to/from a "clipboard".
;;
;; ;; ;; ;; ;; ;; ;; ;; ;; ;; ;; ;; ;; ;; ;; ;; ;; ;; ;; ;; ;; ;; ;; ;; ;;
;;
;; A "clipboard" is group of three buffers: IDs, Anns, and Seqs.
;; Usually the buffers are grouped together in a list, and accessed
;; with the functions `gene-cb-get-{id,ann,seq}-buffer', which take
;; such a clipboard list as argument.
;;
;; New data gets tacked onto the end of a clipboard.
;;
;; Sequences get added to gene-cb-sequence-clipboard, and groups go
;; into gene-cb-sgroup-clipboard.  Sequences are separated by
;; newlines, but the clipboard uses text properties to delimit both
;; sequences and groups of sequences.
;;
;; The ID, annotation, and Seq-data of each sequence in a clipboard
;; have the property `gene-cb-sequence-delimiter' on their first
;; characters.  The property's value is the organism symbol.
;;
;; The first ID, first annotation, and first seq-data of each group in
;; a clipboard have the property `gene-cb-sgroup-delimiter' on their
;; first characters.  The property's value is the sgroup itselff.
;;
;; The clipboard doesn't know about the distinction between Cut and
;; Copy.  All it knows is that the sequence was added.  If the
;; sequence happened to get deleted afterwards, then it was a Cut;
;; else it was a Copy.
;;
;; To Paste into the alignment, we use the three clipboard buffers as
;; the temp-buffers for `gene-io-place-sequences'.
;; `gene-cb-narrow-to-object' is called  to narrow each of the buffers
;; so that only the desired object is visible, and then
;; `gene-io-place-sequences' is called with the three buffers as
;; arguments.  That's all there is to it.
;;
;; Pasting does not remove an object from the clipboard; you can use
;; the gene-cb-remove-* functions to remove objects.
;;
;; Undo makes use of clipboard functions; sequence deletion is
;; recorded for undo in the gene-undo-sequence-clipboard, accessed
;; with the same gene-cb-* functions as other clipboards.


;;; The central data structures:

;; We use different clipboards for single sequences and sequence
;; groups, just because it's easier that way.

(defconst gene-cb-sequence-clipboard
  (list
   (get-buffer-create "*Gene Sequence ID Clipboard*")
   (get-buffer-create "*Gene Sequence Ann Clipboard*")
   (get-buffer-create "*Gene Sequence Seq Clipboard*")
   )
  "The single-sequence clipboard buffers.")


(defconst gene-cb-sgroup-clipboard
  (list
   (get-buffer-create "*Gene Sgroup ID Clipboard*")
   (get-buffer-create "*Gene Sgroup Ann Clipboard*")
   (get-buffer-create "*Gene Sgroup Seq Clipboard*")
   )
  "The sequence group clipboard buffers.")


(defun gene-cb-get-id-buffer (clipboard)
  "Return the ID buffer given CLIPBOARD.
Such lists are of the form \(ID-BUF ANN-BUF SEQ-BUF\).
See `gene-cb-sequence-buffers', for example."
  (nth 0 clipboard))


(defun gene-cb-get-ann-buffer (clipboard)
  "Return the annotation buffer given CLIPBOARD.
Clipboards are lists are of the form \(ID-BUF ANN-BUF SEQ-BUF\).
See `gene-cb-sequence-buffers', for example."
  (nth 1 clipboard))


(defun gene-cb-get-seq-buffer (clipboard)
  "Return the sequence buffer given CLIPBOARD.
Such lists are of the form \(ID-BUF ANN-BUF SEQ-BUF\).
See `gene-cb-sequence-buffers', for example."
  (nth 2 clipboard))


;; No need for undo in the clipboard:
(mapcar 'buffer-disable-undo gene-cb-sequence-clipboard)
(mapcar 'buffer-disable-undo gene-cb-sgroup-clipboard)


;;; For debugging only
(defun gene-cb-show-clipboard (clipboard)
  "Show the buffers of CLIPBOARD.
Your frame should be pretty tall if you're going to do this."
  (interactive)
  (split-window-vertically)
  (split-window-vertically)
  (other-window 1)
  (switch-to-buffer (gene-cb-get-id-buffer clipboard))
  (other-window 1)
  (split-window-vertically)
  (switch-to-buffer (gene-cb-get-ann-buffer clipboard))
  (other-window 1)
  (switch-to-buffer (gene-cb-get-seq-buffer clipboard))
  (other-window 1)
  )



;;; Interfaces to the clipboards:

(defun gene-cb-sequence-clipboard-empty-p ()
  "Return non-nil iff the sequence clipboard is empty."
  (save-excursion
    (set-buffer
     (gene-cb-get-seq-buffer gene-cb-sequence-clipboard))
    (= (buffer-size) 0)))


(defun gene-cb-sgroup-clipboard-empty-p ()
  "Return non-nil iff the sgroup clipboard is empty."
  (save-excursion
    (set-buffer
     (gene-cb-get-seq-buffer gene-cb-sgroup-clipboard))
    (= (buffer-size) 0)))


(defun gene-clipboard-empty-p ()
  "Return non-nil iff the clipboard is empty."
  (and
   (gene-cb-sequence-clipboard-empty-p)
   (gene-cb-sgroup-clipboard-empty-p)))


(defun gene-cb-sequence-in-clipboard-p (sym clipboard)
  "Return non-nil iff SYM is already in CLIPBOARD.
No buffer of CLIPBOARD will be widened."
  (save-excursion
    (set-buffer (gene-cb-get-seq-buffer clipboard))
    (text-property-any
     (point-min) (point-max) 'gene-cb-sequence-delimiter sym)))


(defun gene-cb-sequence-in-seq-clipboard-p (sym &optional clipboard)
  "Return non-nil iff SYM is already in the sequence clipboard."
  (save-excursion
    (set-buffer (gene-cb-get-seq-buffer gene-cb-sequence-clipboard))
    (widen) ; play it safe
    (text-property-any
     (point-min) (point-max) 'gene-cb-sequence-delimiter sym)))


(defun gene-cb-sgroup-in-sgroup-clipboard-p (sgroup)
  "Return non-nil iff SGROUP is already in the sgroup clipboard."
  (save-excursion
    (set-buffer (gene-cb-get-seq-buffer gene-cb-sgroup-clipboard))
    (widen) ; play it saphe
    (text-property-any
     (point-min) (point-max) 'gene-cb-sgroup-delimiter sgroup)))


(defun gene-cb-make-sequence-menu ()
  "Return a menu showing the contents of the sequence clipboard.
The menu looks like:
\(\(NAME1 . ORGANISM-SYM1\) \(NAME2 . ORGANISM-SYM2\) ...\)"
  (save-excursion
    (set-buffer (gene-cb-get-id-buffer gene-cb-sequence-clipboard))
    (widen)
    (let ((lst ())
          (start (point-min)))
      (while (setq start
                   (text-property-not-all
                    start (point-max) 'gene-cb-sequence-delimiter nil))
        (goto-char start)
        (let* ((sym  (get-text-property (point) 'gene-cb-sequence-delimiter))
               (name (buffer-substring-no-properties (point) (progn (end-of-line) (point))))
               (cell (cons name sym)))
          (setq lst (cons cell lst)))
        (setq start (point)))
      (nreverse lst))))


(defun gene-cb-make-sgroup-menu ()
  "Return a menu showing the contents of the sgroup clipboard.
The menu looks like:
\(\(NAME1 . GROUP1\) \(NAME2 . GROUP2\) ...\)"
  (save-excursion
    (set-buffer (gene-cb-get-id-buffer gene-cb-sgroup-clipboard))
    (widen)
    (let ((lst ())
          (start (point-min)))
      (while (setq start
                   (text-property-not-all
                    start (point-max) 'gene-cb-sgroup-delimiter nil))
        (goto-char start)
        (let* ((group (get-text-property (point) 'gene-cb-sgroup-delimiter))
               (name  (concat "(Group)  " (gene-sgroup-get-name group)))
               (cell  (cons name group)))
          (setq lst (cons cell lst)))
        (setq start (1+ (point))))
      (nreverse lst))))



(defun gene-cb-add-sequence-data-do-id 
  (start end clipboard &optional sgroup-marker)
  "Insert the short-id from START to END into the id buffer of CLIPBOARD.
START and END are locations in `gene-id-buffer', marking the
start and end \(including newline\) of the short id in question.
Optional third argument SGROUP-MARKER is a sequence group used to
mark this as the start of a group in the clipboard buffer.
This function does not affect any restrictions in CLIPBOARD."
  (set-buffer (gene-cb-get-id-buffer clipboard))
  (goto-char (point-max))
  (let ((opoint (point)))
    (insert-buffer-substring gene-id-buffer start end)
    (goto-char opoint))
  ;; The start of clipboard data is marked with a delimitation property.
  (put-text-property (point) (1+ (point)) 'gene-cb-sequence-delimiter sym)
  (if sgroup-marker
      (put-text-property
       (point) (1+ (point)) 'gene-cb-sgroup-delimiter sgroup-marker)))


(defun gene-cb-add-sequence-data-do-ann
  (start end clipboard &optional sgroup-marker)
  "Insert annotation data from START to END into the seq buffer of CLIPBOARD.
START and END are locations in `gene-ann-buffer', marking the
start and end \(including newline\) of the annotation in question.
Optional third argument SGROUP-MARKER is a sequence group used to
mark this as the start of a group in the clipboard buffer.
This function does not affect any restrictions in CLIPBOARD."
  (set-buffer gene-ann-buffer)
  (save-restriction
    ;; Necessary to avoid args-out-of-range error when we do
    ;; `insert-buffer-substring':
    (widen)
    
    ;; Okay, on with the good stuff.
    (set-buffer (gene-cb-get-ann-buffer clipboard))
    (goto-char (point-max))
    (let ((opoint (point)))
      (insert-buffer-substring gene-ann-buffer start end)
      (goto-char opoint))
    ;; The start of clipboard data is marked with a delimitation property.
    (put-text-property (point) (1+ (point)) 'gene-cb-sequence-delimiter sym)
    (if sgroup-marker
        (put-text-property
         (point) (1+ (point)) 'gene-cb-sgroup-delimiter sgroup-marker))))


(defun gene-cb-add-sequence-data-do-seq
  (start end clipboard &optional sgroup-marker)
  "Insert sequence data from START to END into the seq buffer of CLIPBOARD.
START and END are locations in `gene-seq-buffer', marking the
start and end \(including newline\) of the sequence in question.
Optional third argument SGROUP-MARKER is a sequence group used to
mark this as the start of a group in the clipboard buffer.
This function does not affect any restrictions in CLIPBOARD."
  (set-buffer (gene-cb-get-seq-buffer clipboard))
  (goto-char (point-max))
  (let ((opoint (point)))
    (insert-buffer-substring gene-seq-buffer start end)
    (goto-char opoint))
  ;; The start of clipboard data is marked with a delimitation property.
  (put-text-property (point) (1+ (point)) 'gene-cb-sequence-delimiter sym)
  ;; Don't keep regions of selection-ness on the sequence; they're
  ;; not useful, since it might be pasted back anywhere at all.
  (gene-sel-remove-selection-fragments-to-eol)
  (if sgroup-marker
      (put-text-property
       (point) (1+ (point)) 'gene-cb-sgroup-delimiter sgroup-marker)))


(defun gene-cb-add-sequence-data
  (sym clipboard &optional sgroup-marker)
  "Internal clipboard function: add sequence SYM to CLIPBOARD.
A clipboard is a list of clipboard buffers; for example, see the value
of `gene-cb-sequence-clipboard'."
  (let* ((id-posn   (gene-organism-id-posn sym))
         (id-start  (nth 0 id-posn))
         (id-end    (1+ (nth 1 id-posn)))
         (seq-posn  (gene-organism-seq-posn sym))
         (seq-start (nth 0 seq-posn))
         (seq-end   (1+ (nth 1 seq-posn))) ; grab newline too
         (ann-posn  (gene-source-annotation
                     (gene-organism-default-source sym)))
         (ann-start (nth 0 ann-posn))
         (ann-end   (nth 1 ann-posn)))
    (gene-cb-add-sequence-data-do-id
     id-start id-end clipboard sgroup-marker)
    (gene-cb-add-sequence-data-do-ann
     ann-start ann-end clipboard sgroup-marker)
    (gene-cb-add-sequence-data-do-seq
     seq-start seq-end clipboard sgroup-marker)
    ))


(defun gene-cb-add-sequence (sym)
  "Add sequence SYM to the seq clipboard, or update it if already present."
  (if (gene-cb-sequence-in-seq-clipboard-p sym)
      (gene-cb-remove-sequence sym))
  (gene-cb-add-sequence-data sym gene-cb-sequence-clipboard))


(defun gene-cb-add-sgroup (group)
  "Add GROUP (a sequence group or its name) to the sgroup clipboard.
If GROUP is already there, then just update it."
  (setq group (gene-sgroup-canonicalize-group group))
  (if (gene-cb-sgroup-in-sgroup-clipboard-p group)
      (gene-cb-remove-sgroup group))
  (let* ((grp-marker group)
         (sequences (gene-sgroup-sequences group))
         (len       (length sequences))
         (count     0))
    (mapcar
     (lambda (sym)
       (gene-cb-add-sequence-data
        sym gene-cb-sgroup-clipboard grp-marker)
       (gene-%-message "Putting group in clipboard... " count len)
       (setq count (1+ count))
       ;; Only the first call should set the group-marker...
       (setq grp-marker nil))
     sequences)))


(defun gene-cb-widen (clipboard)
  "Remove restrictions in the buffers of CLIPBOARD."
  (save-excursion
    (mapcar (lambda (buf) (set-buffer buf) (widen)) clipboard)))


(defun gene-cb-narrow-to-object (property value clipboard)
  "Narrow buffers of CLIPBOARD (third arg) so only a certain object is visible.
The object is marked by property PROPERTY having value VALUE.
It is usually a good idea, if only to avoid debugging confusion, to
widen CLIPBOARD with `gene-cb-widen' when done."
  (save-excursion
    (mapcar
     (lambda (buf)
       (set-buffer buf)
       (widen) ; remove any previous restriction
       (let* ((start (or
                      (text-property-any
                       (point-min) (point-max) property value)
                      (point-min)))
              (end   (or
                      (text-property-not-all
                       (1+ start) (point-max) property nil)
                      (point-max))))
         (narrow-to-region start end)))
     clipboard)))


(defun gene-cb-remove-object (property value clipboard)
  "Remove an sequence or group from a clipboard.
Specifically, remove the object delimited by property PROPERTY having
value VALUE, in the buffers of CLIPBOARD.
If PROPERTY and/or VALUE are nil, then removes everything in the clipboard."
  (save-excursion
    (if (and property value)
        (gene-cb-narrow-to-object property value clipboard)
      (gene-cb-widen clipboard))
    (mapcar
     (lambda (buf)
       (set-buffer buf)
       (delete-region (point-min) (point-max)))
     clipboard)
    (gene-cb-widen clipboard)))


(defun gene-cb-remove-sequence (sym)
  "Remove sequence SYM from the sequence clipboard."
  (gene-cb-remove-object
   'gene-cb-sequence-delimiter sym gene-cb-sequence-clipboard))


(defun gene-cb-remove-sgroup (group)
  "Remove GROUP (a group or its name) from the clipboard."
  (gene-cb-remove-object
   'gene-cb-sgroup-delimiter group gene-cb-sgroup-clipboard))


(defun gene-cb-insert-then-remove-sequence (sym clipboard &optional line)
  "Behave like `gene-cb-insert-sequence-internal', then remove
sequence SYM from CLIPBOARD."
  (gene-cb-insert-sequence-internal sym clipboard line)
  (gene-cb-remove-object 'gene-cb-sequence-delimiter sym clipboard))


(defun gene-cb-insert-sequence-internal (sym clipboard &optional line)
  "Insert sequence SYM from the clipboard into the alignment.
This does not take care of updating the lock-status buffer, it merely
sets the `gene-lock-status-update-needed' flag to t."
  (gene-cb-narrow-to-object
   'gene-cb-sequence-delimiter sym clipboard)
  (set-buffer gene-seq-buffer)
  (if line (gene-goto-line line))
  (gene-io-place-sequences
   (apply 'vector clipboard) nil nil t)
  (setq gene-lock-status-update-needed t)
  ;; Play it safe
  (gene-cb-widen gene-cb-sgroup-clipboard))


(defun gene-cb-insert-sequence (sym &optional line)
  "Insert sequence SYM from the clipboard to line LINE the alignment."
  (gene-cb-insert-sequence-internal
   sym gene-cb-sequence-clipboard line)
  ;; update lock status since passed NO-LOCK-UPDATE flag above
  (gene-status-buffer-update 1))


(defun gene-cb-insert-sgroup (group &optional line)
  "Insert GROUP from the clipboard to line LINE the alignment."
  (setq group (gene-sgroup-canonicalize-group group))
  (gene-cb-narrow-to-object
   'gene-cb-sgroup-delimiter group gene-cb-sgroup-clipboard)
  (set-buffer gene-seq-buffer)
  (if line (gene-goto-line line))

  (gene-io-place-sequences
   (apply 'vector gene-cb-sgroup-clipboard) nil nil t)

  ;; Play it safe
  (gene-cb-widen gene-cb-sgroup-clipboard)

  ;; update lock status since passed T flag above
  (gene-status-buffer-update 1)
  )



;; Friendly wrappers:

(defun gene-cb-copy-sequence (sym)
  "Copy sequence SYM to the clipboard."
  (let ((name (gene-organism-name sym)))
    (gene-message "Copying %s..." name)
    (gene-cb-add-sequence sym)
    (gene-sequences-to-x-selection-buffer sym)
    (gene-message "Copying %s... done." name)))


(defun gene-cb-copy-sgroup (group)
  "Copy sequence group GROUP to the clipboard."
  (let ((name (gene-sgroup-get-name group)))
    (gene-message "Copying group %s..." name)
    (gene-cb-add-sgroup group)
    (gene-sgroup-to-x-selection-buffer group)
    (gene-message "Copying group %s... done." name)))


(defun gene-cb-cut-sequence (sym)
  "Cut sequence SYM to the clipboard."
  (let ((name (gene-organism-name sym)))
    (gene-message "Cutting %s..." name)
    (gene-cb-add-sequence sym)
    (gene-seq-delete-sequence sym)
    (gene-sequences-to-x-selection-buffer sym)
    (gene-handle-vertical-motion t)
    (gene-message "Cutting %s... done." name)))


(defun gene-cb-cut-sgroup (group)
  "Cut sequence group GROUP to the clipboard."
  (let ((name (gene-sgroup-get-name group)))
    (gene-message "Cutting group %s..." name)
    (gene-cb-add-sgroup group)
    (let* ((sequences (gene-sgroup-sequences group))
           (len       (length sequences))
           (count     0))
      (mapcar
       (lambda (seq)
         (gene-seq-delete-sequence seq)
         (gene-%-message "Removing group from alignment... " count len)
         (setq count (1+ count)))
       sequences))
    (gene-sgroup-to-x-selection-buffer group)
    (gene-handle-vertical-motion t)
    (gene-message "Cutting group %s... done." name)))


(defun gene-cb-paste-sequence (sym &optional line)
  "Paste sequence SYM from the clipboard to line LINE in the alignment.
LINE defaults to the current line."
  (gene-cb-insert-sequence sym line)
  (gene-handle-vertical-motion t)
  )


(defun gene-cb-paste-sgroup (group &optional line)
  "Paste  GROUP from the clipboard to line LINE in the alignment.
LINE defaults to the current line."
  (gene-cb-insert-sgroup group line)
  (gene-handle-vertical-motion t)
  )



;;; Putting sequence data into the X selection buffer.
(defun gene-sequences-to-x-selection-buffer (&rest sequences)
  "Put SEQUENCES into the primary X selection, in fasta format."
  (gene-save-excursion-in
      (get-buffer-create "*Gene X Selection Pasteboard*")
    (mapcar
     (lambda (sym)
       (insert ">" (gene-organism-name sym) "\n")
       (let* ((posn  (gene-organism-seq-posn sym))
              (start (car posn))
              (end   (nth 1 posn)))
         (insert-buffer-substring gene-seq-buffer start end))
       (insert "\n"))
     sequences)
    (x-set-selection
     ;; todo: replace `buffer-substring-no-properties' with `cons' when the Emacs
     ;; bug is fixed...
     nil (buffer-substring-no-properties (point-min-marker) (point-max-marker)))
    (erase-buffer)))


(defun gene-sgroup-to-x-selection-buffer (sgroup)
  "Put SGROUP into the primary X selection buffer, in fasta format."
  (apply 'gene-sequences-to-x-selection-buffer
         (gene-sgroup-sequences sgroup)))



(provide 'gene-clipboard)
