;;;; gene-consensus.el --- calculating consensus (mask) sequences.
;; Copyright (C) 1994 RDP, Indiana University Foundation
;;
;; This software is copyrighted under the terms of the GNU General
;; Public License (see terms below).  The copyright covers this file,
;; specifically, and the program of which it is a part.
;; 
;;      This program is free software; you can redistribute it and/or modify
;;      it under the terms of the GNU General Public License as published by
;;      the Free Software Foundation; either version 1, or (at your option)
;;      any later version.
;; 
;;      This program is distributed in the hope that it will be useful,
;;      but WITHOUT ANY WARRANTY; without even the implied warranty of
;;      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;      GNU General Public License for more details.
;; 
;;      You should have received a copy of the GNU General Public License
;;      along with this program; if not, write to the Free Software
;;      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

(require 'gene-mode-tools)


(defun gene-consensus-generate-position-list (sequences)
  "Create a list of vectors from SEQUENCES \(a list of orgsyms\):

\(\[ORGSYM1 START END\] \[ORGSYM2 START END\] \[ORGSYM3 START END\] ... \)

where each START and END are the boundaries of the sequence data for
their respective organisms."
  (mapcar
   (lambda (sym)
     (let* ((posn (gene-organism-seq-posn sym))
            (b    (nth 0 posn))
            (e    (nth 1 posn)))
       (vector sym b e)))
   sequences))


(defun gene-consensus-disambiguate-vector (vec)
  "There is some question as to how to implement this.  Therefore, for
the time being, we will simply return VEC untouched."
  (or
   vec
   (let ((ch 0)
         (veclen (length count-vec)))
     (while (< ch veclen)
       (let ((count      (aref count-vec ch)))
         (or
          (zerop count)
          (let* ((str        (gene-search-disambiguate-base ch))
                 (strlen     (length str))
                 (resol-idx 0)))
          (while (< resol-idx strlen)
            (let ((resol-ch (aref str resol-idx)))
              (or
               (gene-upper-case-p resol-ch)
               (gene-search-ambiguous-character-p resol-ch)
               (progn
                 (message "%d  %c  -->  %c" count resol-ch ch)
                 (sit-for 1)
                 (aset count-vec
                       (upcase resol-ch)
                       (+ (aref count-vec (upcase resol-ch)) count)))))
            (setq resol-idx (1+ resol-idx)))))
       (setq ch (1+ ch))))))


(defun gene-consensus-filter-column
  (count-vector percentage total &optional mask-p disambiguate-p)
  "Discover if an element COUNT-VECTOR surpasses PERCENTAGE of TOTAL. 
Optional fourth argument MASK-P means return the index \(i.e.: character\)
of the element with the highest count, but if no counts exceed
PERCENTAGE of TOTAL, then return ?-.
Else if MASK-P is nil, just return ?1 if any element exceeds
PERCENTAGE of TOTAL, else return ?0.
Optional fifth argument DISAMBIGUATE-P means resolve ambiguity codes
to all possible residues before counting \(but resolve them to
upper-case only\).

Explanation:
What we're measuring here is column residue percentages, and COUNT-VECTOR
is length 256 and of the form:

\[NUM NUM NUM ...\]

where each NUMber represents the number of times that character was
seen in this column.  Naturally, most NUMbers will probably be zero."
  (if disambiguate-p
      ;; If we do this, TOTAL will need updating too!
      (gene-consensus-disambiguate-vector count-vec))

  (let ((needed-count
         (float (/
                 (float (* (float percentage) (float total)))
                 (float 100)))))
    (if mask-p
        (catch 'hit
          (mapcar
           (lambda (count)
             (or
              (zerop count)
              (if (>= count needed-count)
                  (throw 'hit ?1))))
           count-vector)
          ?0)
      ;; Else we're generating a true consensus sequence, not just a 0/1
      ;; mask, so we'll need to keep track of some extra information.
      (let ((idx 0)
            (ch  ?-))
        (mapcar
         (lambda (count)
           (or
            (zerop count)
            (if (>= count needed-count)
                (setq needed-count count ch idx)))
           (setq idx (1+ idx)))
         count-vector)
        ch
        ))))


(defun gene-consensus-generate-sequence-or-mask
  (name percentage gaps-p sequences &optional mask-p)
  "Generate a consensus sequence.
It's name will be NAME, threshold PERCENTAGE, GAPS-P indicates whether
indels should be counted or not, and it will be generated from
SEQUENCES \(a list of organism symbols\).
Optional fifth arg MASK-P means generate a mask instead of a consensus
sequence."

  (let* ((positions (gene-consensus-generate-position-list sequences))
         (maxlen    (let ((len 0))
                      (mapcar
                       (lambda (vec)
                         (setq len (max len (- (elt vec 2) (elt vec 1)))))
                       positions)
                      len))
         (mask-str  (make-string maxlen ?0))
         (count-vec (make-vector 256 0))
         (cur-total 0)
         (i         0)
         (j         0))
    (gene-save-excursion-in gene-seq-buffer
      (while (< i maxlen)
        (mapcar
         (lambda (pos-vec)
           (let ((start (elt pos-vec 1))
                 (end   (elt pos-vec 2)))
             (unless (>= start end)
               
               (let ((ch (upcase (char-after start))))
                 (if gaps-p
                     (progn
                       (setq cur-total (1+ cur-total))
                       (aset count-vec ch (1+ (aref count-vec ch))))
                   ;; Else we're only counting bases.
                   (if (gene-base-p ch)
                       (progn
                         (setq cur-total (1+ cur-total))
                         (aset count-vec ch (1+ (aref count-vec ch))))))))
             ;; Nudge position forward one char when done.
             (aset pos-vec 1 (1+ start))))
         positions)
        
        ;; Put a "1" in the mask here if we have a consensus.
        (aset mask-str i (gene-consensus-filter-column
                          count-vec percentage cur-total mask-p))
        
        ;; Zero the count vector
        (setq j 0)
        (while (< j 256)
          (aset count-vec j 0)
          (setq j (1+ j)))
        
        ;; Zero the current total
        (setq cur-total 0)
        
        ;; Increment the column counter & keep user happy.
        (gene-message "Counting column %d of %d..." i maxlen)
        (setq i (1+ i)))

      (gene-message "Counting column %d of %d... done." i maxlen)
      
      ;; Finally, open and return the new mask sequence.
      (gene-open-sequence name mask-str "MASK SEQUENCE"))))


(defun gene-consensus-sequence-or-mask-from-group
  (sgroup percentage gaps-p name mask-p)
  "Generate a mask from SGROUP, with consensus PERCENTAGE.
If optional third argument GAPS-P is `yes', then count gaps too; else if
`no', or any other non-nil value, then don't.
Fourth arg NAME is a name for the mask sequence; nil means \"mask\" if
MASK-P \(see below\), else \"consens\".
The mask sequence will become a member of SGROUP.
Optional arg MASK-P means generate a mask instead of a consensus sequence."
  (let ((mask-sym
         ;; Generate the mask...
         (gene-consensus-generate-sequence-or-mask
          (or name (if mask-p "mask" "consens"))
          percentage
          gaps-p
          (gene-sgroup-sequences sgroup)
          mask-p)))
    ;; ... and add it to the group.
    (gene-sgroup-add-seq-to-group sgroup mask-sym)))


;;; consensus mask dialog box

(defvar gene-consensus-dialog-box-buffer nil
  "Non-nil is a cached buffer, the consensus dialog box.")


(defvar gene-consensus-this-sgroup nil
  "Non-nil is sequence group for which a consensus mask will be generated.")


(defun gene-consensus-dialog-box (sgroup)
  "Internal function: pop up dialog box for generating a consensus
sequence for SGROUP.
Optional arg MASK-P means generate a mask instead of a consensus
sequence."
  (setq gene-consensus-this-sgroup sgroup)
  (gene-frame-show
   (or
    gene-consensus-dialog-box-buffer
    (setq gene-consensus-dialog-box-buffer
          (gene-consensus-make-dialog-box-buffer)))
   '((foreground-color . "white")
     (background-color . "black")
     (cursor-color     . "white")
     (height           . 13)
     (width            . 40)
     )
   ))


(defun gene-consensus-unparse-options-and-execute (opt-str)
  "Turn OPT-STR to arguments for `gene-consensus-sequence-or-mask-from-group'.
OPT-STR was generated by gene-dialog-compile-options and probably
looks something like

       \"70 nil t\"

that is, a percentage, followed by count-indels-p, followed by mask-p."
  (let* ((percentage-cell (read-from-string opt-str))
         (percentage      (car percentage-cell))
         (next-idx        (cdr percentage-cell))
         (indels-cell     (read-from-string opt-str next-idx))
         (count-indels-p  (car indels-cell))
         (next-idx        (cdr indels-cell))
         (mask-cell       (read-from-string opt-str next-idx))
         (mask-p          (car mask-cell))
         )
    (gene-consensus-sequence-or-mask-from-group
     gene-consensus-this-sgroup percentage count-indels-p nil mask-p)))


(defun gene-consensus-make-dialog-box-buffer ()
  "Generate a dialog box buffer for consensus sequences.
This buffer is generally cached, so this function is normally only
called once per session."
  (let ((buf (get-buffer-create "*Consensus Mask*")))
    (save-excursion
      (set-buffer buf)
      (gene-dialog-box-mode 'gene-consensus-unparse-options-and-execute)
      (let ((inhibit-read-only t))
        (erase-buffer)) ; total paranoia
      (gene-dialog-insert-text-field "Percentage:" "")
      (insert "\n\n")
      (gene-dialog-insert-boolean-field "Count indels?" "t" "nil" 'no)
      (insert "\n\n")
      (gene-dialog-insert-radio-buttons "Type:" 
                                        "SEQUENCE"
                                        '("MASK" . "t")
                                        '("SEQUENCE" . "nil"))
      (insert "\n\n")
      (gene-dialog-insert-send-button "<Generate Mask>" "")
      (goto-char (point-min))
    buf)))



(provide 'gene-consensus)
