;; gene-id-mode.el (part of ale).
;; Copyright (C) 1993, 1994  Karl Fogel <kfogel@darwin.life.uiuc.edu>
;;
;; This software is copyrighted under the terms of the GNU General
;; Public License (see terms below).  The copyright covers this file,
;; specifically, and the program of which it is a part.
;; 
;;      This program is free software; you can redistribute it and/or modify
;;      it under the terms of the GNU General Public License as published by
;;      the Free Software Foundation; either version 1, or (at your option)
;;      any later version.
;; 
;;      This program is distributed in the hope that it will be useful,
;;      but WITHOUT ANY WARRANTY; without even the implied warranty of
;;      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;      GNU General Public License for more details.
;; 
;;      You should have received a copy of the GNU General Public License
;;      along with this program; if not, write to the Free Software
;;      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

;;; Code:
(require 'gene-mode-tools)

;; Never bind this:
;; (setq the-mouths-of-the-kine-who-tread-the-grain 1)

;; id mode funcs:

(defvar gene-id-mode-map (make-sparse-keymap))
(suppress-keymap gene-id-mode-map)

(define-key gene-id-mode-map [vertical-scroll-bar down-mouse-1]
  'gene-seq-drag-vertical-scroll-bar)
(define-key gene-id-mode-map [vertical-scroll-bar down-mouse-2]
  'gene-seq-drag-vertical-scroll-bar)
(define-key gene-id-mode-map [vertical-scroll-bar down-mouse-3]
  'gene-seq-drag-vertical-scroll-bar)


;; I wonder if this is working.  Hmm.  Well, at least it doesn't seem
;; to be hurting anything.
;; Don't let outside Emacs bindings show through:
;; (nconc
;;  gene-id-mode-map
;;  (list
;;   (cons t
;;         (lambda ()
;;           (interactive)
;;           (gene-error
;;        "%s has been unbound here.  Please check the number and dial again."
;;            (this-command-keys)
;;            )))))



;;; Searching on ID names.
(defun gene-id-match-string (re)
  "Return a list of sequences whose short IDs contain a match for RE."
  (save-excursion
    (set-buffer gene-id-buffer)
    (goto-char (point-min))
    (save-match-data
      (let (matches)
        (while (re-search-forward re nil t)
          (let ((sym (gene-get-here-sym)))
            (or
             (eq sym (car matches))
             (setq matches (cons sym matches)))))
        matches))))


(defun gene-id-ids-containing-string (str)
  "Return a list of organisms whose IDs contain the substring STR."
  (interactive "sIDs matching string: ")
  (gene-id-match-string (regexp-quote str)))


(defun gene-id-search-and-build-group (re &optional sgroup)
  "Make a group of all sequences whose short IDs contain a match for RE.
Optional second arg SGROUP is a group to use, or a group name.
If it's a name, the group will be created if it doesn't already
exist."
  (interactive "sSearch IDs: ")
  (let ((sym-list (gene-id-ids-containing-string re))
        (str
         (concat "ID: " (substring re 0 (min (length re) 7)))))
    (if sym-list
        (progn
          (gene-sgroup-make-group-from-sequences sym-list str)
          (gene-sgroup-point-to-group (gene-sgroup-get-group-by-name str))
          (message "Matched %d sequences." (length sym-list)))
      (message "No matches for \"%s\"." re))))



;;; highlighting stuff ;;

(defvar gene-id-current-line-number 1)
(defvar gene-id-last-line-number 1)
(defvar gene-id-highlight-current-id t
  "*Non-nil means highlight the current ID in the ID window.")
(defvar gene-id-current-id-color "green3"
  "Color for ID highlighting bar which follows point in the alignment buffer.")
(defvar gene-id-orig-face 'gene-id-orig-face
  "Default face of the ID window.")

(defun gene-id-mode-make-faces ()
  "Create the appropriate faces for gene-id-mode.
`gene-made-all-faces' calls this.

Note that it's incorrect to make faces in top-level expressions, since
the construction of a face depends on run-time factors (like the fonts
available)."
  (copy-face 'default 'gene-id-current-id-face)
  (set-face-background
   'gene-id-current-id-face
   (gene-check-color-allowable gene-id-current-id-color "black"))
  (copy-face 'default 'gene-id-orig-face))


;;;; completion stuff
(defun gene-id-generate-id-completion-table ()
  "Build a completion table for ID names.
Generates an alist of form ((id-str . org-sym) ...)."
  ;; note: no real reason to do it in the ID buffer, now that we're
  ;; using symbol plists, hmm.
  (save-excursion
    (set-buffer gene-id-buffer)
    (goto-char (point-min))
    ;; todo: could we make use of gene-seq-do-column-op here?  Hmm...
    (let (table
          (forward 0))
      (while (= forward 0)
        (let* ((sym (gene-get-here-sym))
               (str (gene-organism-name sym)))
          (if sym
              (setq table (cons (cons str sym) table)))
          (setq forward (forward-line 1))))
      ;; Might as well return it in "order", why not:
      (nreverse table))))


;;;; End sequence group stuff.

(defvar gene-id-mode-hook nil
  "*Hook run when gene-id mode is called.")

(defvar gene-id-current-id-overlay nil
  "The overlay that highlights the current ID.")

(defvar gene-id-display-table (make-display-table)
  "The display table for the ID buffer.")
(defvar gene-id-vertical-window-border-char ? 
  "Char used for vertical window borders in the ID buffer.")
(set-display-table-slot gene-id-display-table 'vertical-border
                        gene-id-vertical-window-border-char)

(defun gene-id-mode ()
  "Major mode for the id buffer.

\\{gene-id-mode-map}"
  (kill-all-local-variables)
  (use-local-map gene-id-mode-map)
  (setq truncate-lines t)
  (toggle-read-only 1)
  (transient-mark-mode -1)
  (buffer-disable-undo)
  (setq mark-active nil)

  (goto-char (point-min))
  (setq gene-id-current-id-overlay
        (make-overlay
         (gene-bol-point)
         (gene-eol-point)
         gene-id-buffer))

  (setq buffer-display-table gene-id-display-table)

  (setq mode-line-format "")

  (overlay-put gene-id-current-id-overlay 'face 'gene-id-current-id-face)
  (setq major-mode 'gene-id-mode)
  (setq mode-name "Gene Organism ID")
  (run-hooks 'gene-id-mode-hook 'gene-id-mode-hooks))

(provide 'gene-id-mode)
