;;;; gene-status-mode.el (part of ale).
;; Copyright (C) 1994  Karl Fogel <kfogel@darwin.life.uiuc.edu>
;; Copyright (C) 1994  Indiana University Foundation
;;
;; This software is copyrighted under the terms of the GNU General
;; Public License (see terms below).  The copyright covers this file,
;; specifically, and the program of which it is a part.
;; 
;;      This program is free software; you can redistribute it and/or modify
;;      it under the terms of the GNU General Public License as published by
;;      the Free Software Foundation; either version 1, or (at your option)
;;      any later version.
;; 
;;      This program is distributed in the hope that it will be useful,
;;      but WITHOUT ANY WARRANTY; without even the implied warranty of
;;      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;      GNU General Public License for more details.
;; 
;;      You should have received a copy of the GNU General Public License
;;      along with this program; if not, write to the Free Software
;;      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

(require 'gene-mode-tools)

;; Contains status buffer & locking code.

;;; Status mode stuff
(defvar gene-status-mode-map (make-sparse-keymap)
  "Keymap for the gene status buffer.")

(define-key gene-status-mode-map [mouse-1]
  'gene-status-dispatch-click)
(define-key gene-status-mode-map [drag-mouse-1]
  'gene-absorb-mouse-event)
(define-key gene-status-mode-map [down-mouse-1]
  'gene-absorb-mouse-event)
(define-key gene-status-mode-map [double-down-mouse-1]
  'gene-absorb-mouse-event)
(define-key gene-status-mode-map [triple-down-mouse-1]
  'gene-absorb-mouse-event)

(define-key gene-status-mode-map [mouse-2]
  'gene-absorb-mouse-event)
(define-key gene-status-mode-map [drag-mouse-2]
  'gene-absorb-mouse-event)
(define-key gene-status-mode-map [down-mouse-2]
  'gene-absorb-mouse-event)
(define-key gene-status-mode-map [double-down-mouse-2]
  'gene-absorb-mouse-event)
(define-key gene-status-mode-map [triple-down-mouse-2]
  'gene-absorb-mouse-event)

(define-key gene-status-mode-map [mouse-3]
  'gene-absorb-mouse-event)
(define-key gene-status-mode-map [drag-mouse-3]
  'gene-absorb-mouse-event)
(define-key gene-status-mode-map [down-mouse-3]
  'gene-absorb-mouse-event)
(define-key gene-status-mode-map [double-down-mouse-3]
  'gene-absorb-mouse-event)
(define-key gene-status-mode-map [triple-down-mouse-3]
  'gene-absorb-mouse-event)

(define-key gene-status-mode-map [vertical-scroll-bar down-mouse-1]
  'gene-seq-drag-vertical-scroll-bar)
(define-key gene-status-mode-map [vertical-scroll-bar down-mouse-2]
  'gene-seq-drag-vertical-scroll-bar)
(define-key gene-status-mode-map [vertical-scroll-bar down-mouse-3]
  'gene-seq-drag-vertical-scroll-bar)


;; I wonder if this is working.  Hmm.  Well, at least it doesn't seem
;; to be hurting anything.
;; Don't let outside Emacs bindings show through:
;; (nconc
;;  gene-status-mode-map
;;  (list
;;   (cons t
;;         (lambda ()
;;           (interactive)
;;           (gene-error
;;       "%s has been unbound here.  Please check the number and dial again."
;;            (this-command-keys)
;;            )))))


;; Display table stuff.
(defvar gene-status-display-table (make-display-table)
  "The display table for the Status buffer.")
(defvar gene-status-vertical-window-border-char ? 
  "Char used for vertical window borders in the status buffer.")
(set-display-table-slot gene-status-display-table 'vertical-border
                        gene-status-vertical-window-border-char)

(defvar gene-status-mode-hook nil
  "Hook run when gene-status-mode is called.")

;; todo: should this be in its own file, maybe?  Perhaps later...
;;; Status Window:

;; Right now, this will just handle locking stuff.  It will work in
;; much the same way as the group buffer does, with text properties
;; for commands and data on which to operate.  It needs many of the
;; same commands: update-buffer, with opt arg.  However, it differs in
;; one respect: it does not handle it's own display.  The Window
;; Arranging function(s) below will take care of that.


(defun gene-status-mode ()
  "Major mode for the status buffer.

\\{gene-status-mode-map}"
  (kill-all-local-variables)
  (suppress-keymap gene-status-mode-map t)
  (use-local-map gene-status-mode-map)
  (toggle-read-only 1)

  (make-local-variable 'gene-last-window-start) ; seems to be unused...
  (goto-char (point-min))
  (setq cache-long-line-scans t)
  (setq buffer-display-table gene-status-display-table)
  (setq mark-active nil)
  (buffer-disable-undo)
  (transient-mark-mode -1)
  (setq mode-line-format "Locks")
  (setq major-mode 'gene-status-mode)
  (setq mode-name "Gene Status")
  (run-hooks 'gene-status-mode-hook 'gene-status-mode-hooks))


(defvar gene-lock-status-update-needed nil
  "Non-nil if the lock-status buffer is out of date with the alignment.
This is set to nil by gene-status-buffer-update.
Some of the undo-clipboard paste functions set it to t; it is then
examined by gene-undo and the status-buffer is update if necessary.")


(defun gene-status-buffer-update (&optional sym)
  "Update the status buffer.
If optional argument SYM is a symbol, then it is the orgsym of the only
sequence that needs updating.
If SYM is a number, then do a full, guaranteed correct update.
If SYM is nil or omitted, then do a \"scanning\" update, which does
not check the number of sequences in the buffer, but only the state
strings."
  (save-excursion
    ;; This better be called from one of the gene-buffers.
    (set-buffer gene-status-buffer)
    (let (buffer-read-only)
      
      (cond
       ((and sym (symbolp sym))
        ;; Do a fast, single-seq update on the specified sequence.
        (let ((spot (text-property-any
                     (point-min) (point-max) 'gene-ident sym)))
          (if spot
              ;; This sequence already has a line in the status buffer.
              ;; todo: an even faster method would be to use
              ;; gene-pave-with-string, and so avoid gap-motion and
              ;; text-property shuffling.
              (progn
                (goto-char spot)
                (delete-region (gene-bol-point) (gene-eol-point)))
            ;; Else this sequence does not have a line in the status
            ;; buffer, so we'll make one.
            (let ((line (gene-seq-line-number sym)))
              (goto-char (point-min))
              (forward-line (1- line))
              (insert "\n")
              (forward-line -1)))
          (gene-status-insert-buttons sym)))
        
       ((numberp sym)
        ;; Do a full update of everyone.
        (let ((syms (gene-make-sym-list)))
          (delete-region (point-min) (point-max))
          (mapcar
           (lambda (sym)
             (gene-status-insert-buttons sym)
             (insert "\n"))
           syms)))

       (t
        ;; Only update those who need it
        (save-excursion
          (goto-char (point-min))
          (while (progn
                   (let ((sym (gene-get-here-sym)))
                     (or
                      (not sym)
                      (looking-at (regexp-quote
                                   (gene-locking-state-string
                                    (gene-locking-get-sequence-state
                                     (gene-get-here-sym)))))
                      (progn
                        (delete-region (gene-bol-point) (gene-eol-point))
                        (gene-status-insert-buttons sym))))
                   (= 0 (forward-line 1)))))))
      ;; Record that the buffer is up-to-date.
      (setq gene-lock-status-update-needed nil)
      )))


(defun gene-make-sym-list ()
  "Return an ordered list of the orgsyms of these gene-buffers.
The order is from first to last (as they appear in ID and seq
buffers)."
  (gene-save-excursion-in gene-seq-buffer
    ;; We'll use the seq buffer, so that the cache gets refilled if
    ;; it's missing anything.
    (goto-char (point-min))
    (let (syms sym)
      (while (progn
	       (if (setq sym (gene-get-here-sym nil nil t))
                   (setq syms (cons sym syms)))
               (zerop (forward-line 1))))
      (nreverse syms))))


(defun gene-status-insert-buttons (sym)
  "Insert the correct status buttons for sequence SYM (an orgsym).
This function assumes that point is at the correct place in the status
buffer for the insertion.  It does not insert the newline."
  ;; The names on the user-visible buttons are not necessarily the
  ;; same as the terminology we use in the code.  We need to be able
  ;; to change our minds about the presentation without having to
  ;; rewrite code.  So, internally, the states are "locked",
  ;; "gapshift", "gapinout", and "locked".  Externally, they are
  ;; whatever words we choose, mere whims on the shifting wind of
  ;; passing fashion.
  (let* (buffer-read-only
         (state       (gene-locking-get-sequence-state sym))
         (face        (aref gene-locking-text-face-table state))
         (mouse-face  (aref gene-locking-mouse-face-table state))
         (start-pt    (point))
         (spacer      " "))
    
    ;; Locking button
    (insert (gene-locking-state-string state))
    (set-text-properties start-pt (point)
			 (list 'func 'gene-locking-cycle-to-next-state
			       'gene-ident sym
			       'face face
			       'mouse-face mouse-face))
    (insert spacer)
    (setq start-pt (point))

    ;; Other buttons come later...

    ))

;; These will use the window-arranging routines:
(defun gene-display-status-window ()
  "Display the status buffer in a window."
  (interactive)
  (let (frame)
    (and (not (gene-status-buffer-window))
	 (setq frame (gene-seq-buffer-frame))
	 (progn
	   (select-frame frame)
	   (apply 'gene-arrange-windows
		  gene-status-buffer
		  gene-id-buffer
		  (if gene-window-split-p
		      (list gene-seq-buffer gene-seq-buffer)
		    (list gene-seq-buffer)))
	   (gene-do-window-sync t)))))
    

(defun gene-hide-status-window ()
  "Hide the status window."
  (interactive)
  (let (frame)
    (and (gene-status-buffer-window)
	 (setq frame (gene-seq-buffer-frame))
	 (progn
	   (select-frame frame)
	   (apply 'gene-arrange-windows
		  gene-id-buffer
		  (if gene-window-split-p
		      (list gene-seq-buffer gene-seq-buffer)
		    (list gene-seq-buffer)))))))


(defun gene-toggle-status-window (&optional force)
  "Toggle displaying of the status window.
If it is displayed, it will be hidden, and versa-vice.
If optional arg FORCE is non-nil, then FORCE positive means display
it, else hide it."
  (interactive)
  (if force ; take the docstring literally -- it might not be a number!
      (or (and (numberp force)
               (> force 0)
               (prog1 t (gene-display-status-window)))
          (gene-hide-status-window))
    ;; Else no arg, so do the right thing:
    (if (get-buffer-window gene-status-buffer 0)
        (gene-hide-status-window)
      (gene-display-status-window))))


(defun gene-status-dispatch-click (event)
  "Handle a mouse click in the Status Buffer."
  (interactive "e")
  (let* ((pt (posn-point (event-start event)))
         (command (get-text-property pt 'func gene-status-buffer))
         ;; We don't call gene-get-here-sym because this has to happen
         ;; in the status buffer, at some random spot, not in the
         ;; current buffer at point.
         (seq (gene-get-here-sym pt gene-status-buffer)))
    (if command
        (if seq
            (funcall command seq)
          (funcall command)))))



;;;; Sequence locking
;;; The Basics:
;;
;; There are four possible states of locking for a sequence:
;; 0. "Locked": No changes allowed at all.  Read-only.
;; 1. "GapShift": gaps shift around, but total sequence length stays same.
;; 2. "GapInOut": gaps can be inserted or deleted.
;; 3. "Unlocked": anything can be inserted or deleted.
;;
;; Note that each state is a "superset" of the preceding state: any
;; operation permissible in one state is permissible in all "higher"
;; states, any operation prohibited in one state is prohibited in all
;; lower states.
;;
;; State 2 is the default.  The lock states are stored in the org sym.
;; Operations must check the lock-state before doing their thing, if
;; they might be illegal.
;;
;; A word about terminology: to "lock" a sequence doesn't necessarily
;; mean to put it into state 0, it just means to put it into one of
;; the lock states.  You "lock to state", which state could be
;; `unlocked' as easily as anything else.
;;
;; Because I'm using the word "lock" in that way, state number 1 is
;; called "locked".  However, we could change this if we wanted to, no
;; problem, if "locked" is clearer for users.

;; The locking states themselves.  Ooooh!
(defconst gene-locking-locked-state   0)
(defconst gene-locking-gapshift-state 1)
(defconst gene-locking-gapinout-state 2)
(defconst gene-locking-unlocked-state 3)


;;; Colors & Faces for the status buffer:

(defun gene-locking-make-face-pair (state)
  "Set up the faces for the locking state named STATE, a string.
This sets up two faces named `gene-locking-STATE-face' and
`gene-locking-STATE-mouse-face' which are copies of the default face,
sets their foregound colors to the value of the symbol named
`gene-locking-STATE-face-color',
and makes the latter a bold version of the former."
  (let ((face (intern (format "gene-locking-%s-face" state)))
	(mouse-face (intern (format "gene-locking-%s-mouse-face" state)))
	(color (symbol-value
		(intern (format "gene-locking-%s-face-color" state)))))
    (copy-face 'default face)
    (set-face-foreground face (gene-check-color-allowable color))
    (copy-face face mouse-face)
    (make-face-bold mouse-face nil t)))


(defun gene-locking-make-faces ()
  "Create the appropriate faces for displaying gene locking info.
`gene-made-all-faces' calls this.

Note that it's incorrect to make faces in top-level expressions, since
the construction of a face depends on run-time factors (like the fonts
available)."

  (defvar gene-locking-locked-face-color  "coral")
  (defvar gene-locking-gapshift-face-color "wheat")
  (defvar gene-locking-gapinout-face-color gene-foreground-color)
  (defvar gene-locking-unlocked-face-color "green")

  (gene-locking-make-face-pair "locked")
  (gene-locking-make-face-pair "gapshift")
  (gene-locking-make-face-pair "gapinout")

  ;; Faces for unlocked state:
  (copy-face 'default 'gene-locking-unlocked-face)
  (set-face-background
   'gene-locking-unlocked-face
   (gene-check-color-allowable gene-locking-unlocked-face-color))
  (set-face-foreground
   'gene-locking-unlocked-face
   (gene-check-color-allowable "black"))
  (copy-face 'gene-locking-unlocked-face
	     'gene-locking-unlocked-mouse-face)
  (make-face-bold 'gene-locking-unlocked-mouse-face nil t))
  
(defconst gene-locking-text-face-table
  '[gene-locking-locked-face
    gene-locking-gapshift-face 
    gene-locking-gapinout-face 
    gene-locking-unlocked-face]
  "Table of text faces for the lock states, indexable by state.")

(defconst gene-locking-mouse-face-table
  '[gene-locking-locked-mouse-face
    gene-locking-gapshift-mouse-face 
    gene-locking-gapinout-mouse-face 
    gene-locking-unlocked-mouse-face]
  "Table of mouse faces for the lock states, indexable by state.")


;;; Functions:
;; We probably don't want this to notice sgroups, because it's mainly
;; used by the selection locking code and
;; gene-locking-cycle-to-next-state.
;; The former ought to ignore sgroups, since they're orthoganal to
;; selections, and the latter already notices sgroups.


(defun gene-locking-lock-sequence-to-state (sym state &optional batch)
  "Lock sequence SYM to lock state STATE.
SYM is an organism symbol; nil means the one at point.
STATE is one of 1, 2, 3, or 4.
1 is locked, 2 is gapshift, 3 is gapinout, 4 is unlocked.

If optional third arg BATCH is non-nil, don't update the status
buffer.  This is for use in lisp that sets up initial lock properties."
  (or sym (setq sym (gene-get-here-sym)))
  (put sym 'gene-lock-state state)
  ;; In the future, we may not want to update the status buffer unless
  ;; it's actually displayed.  However, if we do that, then we must
  ;; remember to always update before it is displayed!
  (or batch (gene-status-buffer-update sym)))


(defun gene-locking-next-state-from-this-one (state)
  "Return the next locking state after STATE (an integer)."
  (mod (1+ state) 4))


(defun gene-locking-get-sequence-state (sym)
  "Get sequence SYM's lock state.
SYM is an organism symbol.
The state will be one of 0, 1, 2, or 3.
0 is locked, 1 is gapshift, 2 is gapinout, 3 is unlocked."
  (or 
   (get sym 'gene-lock-state)
   (if (gene-alignment-empty-p)
       (gene-error
        "Open a file, or choose \"Create New Sequence\" from the Edit menu.")
     (gene-error "No locking information available... can't find sequence!"))))


(defun gene-locking-state-string (state)
  "Return a string describing locking state STATE."
  (cond
   ((= state gene-locking-locked-state)
    "-Locked-")
   ((= state gene-locking-gapshift-state)
    "GapShift")
   ((= state gene-locking-gapinout-state)
    "GapInOut")
   ((= state gene-locking-unlocked-state)
    "Unlocked")
   (t
    "Strange!")))


(defun gene-locking-cycle-to-next-state (sym-or-group)
  "Cycle to the \"next\" locking state for sequence SYM-OR-GROUP.
If SYM-OR-GROUP a sequence sym, then lock it to the next logical
locking state and display a message to that effect.  If it's a
sequence sym, but is a member of the selected group, then set every
sequence in the group to SYM-OR-GROUP's next locking state.
If SYM-OR-GROUP is a sequence group, then cycle everyone in the group
according to the first sequence in the group."
  (let* ((sym (if (symbolp sym-or-group)
                  sym-or-group
                ;; Else it's a group, so get first sym in it.
                (car (gene-sgroup-sequences sym-or-group))))
         (group (if (symbolp sym-or-group)
                    (gene-sgroup-selected-group-at-seq sym-or-group)
                  sym-or-group))
         (next-state
          (gene-locking-next-state-from-this-one
           (gene-locking-get-sequence-state sym))))
    
    (if (not sym)
        ;; If we couldn't find a sym anywhere, then this is an empty
        ;; group, and cycling its lock status would be an exercise in
        ;; futility.
        nil
      
      ;; Else we have a sym, so do it to it (or them)!
      (if group
          ;; If there's a group here, lock everyone the same way:
          (progn
            (mapcar
             (lambda (sym)
               (gene-locking-lock-sequence-to-state sym next-state))
             (gene-sgroup-sequences group))
            (message "Locked group \"%s\" to \"%s\"."
                     (gene-sgroup-get-name group)
                     (gene-locking-state-string
                      next-state)))
        ;; Else do it to just this sequence
        (gene-locking-lock-sequence-to-state sym next-state)
        (message "Locked %s to \"%s\"."
                 (gene-organism-name sym)
                 (gene-locking-state-string
                  next-state))))))


(defsubst gene-locking-sequence-state-p (sym state &optional err)
  "Return non-nil iff sequence SYM's locking state permits all
operations allowed under STATE.
SYM defaults to the organism symbol at point.
STATE is an integer 0, 1, 2, or 3. 
Optional third arg ERR means signal an error instead of returning nil."
  (or sym (setq sym (gene-get-here-sym)))
  (or
   (>= (gene-locking-get-sequence-state sym) state)
   (if err (gene-error "%s is locked." (gene-organism-name sym)))))


;; Group checking functions:

(defun gene-locking-sgroup-check-state (sgroup state &optional col)
  "Check that all of SGROUP permits gapshift operations.
Signal an error if it doesn't permit them.
Optional second arg COL means don't check those sequences which are
too short to contain COL.
Don't rely on this function for a return value; it works by signalling
an error or failing to signal an error."
  (if col
      (mapcar
       (lambda (sym)
         (let*  ((posn   (get sym 'gene-seq-posn))
                 (start  (car posn))
                 (end    (car (cdr posn))))
           (or
            (< (- end start) col)
            (gene-locking-sequence-state-p sym state t))))
       (gene-sgroup-sequences sgroup))

    ;; Else no COL specified, so we can do a more efficient mapping:
    (mapcar
     (lambda (sym)
       (gene-locking-sequence-state-p sym state t))
     (gene-sgroup-sequences sgroup))))


;; Selection checking functions:


(defun gene-locking-selection-check-state (selection state &optional no-verify)
  "Check that all of SELECTION permits operations allowable in lock STATE.
Signal an error if it doesn't permit them.
Optional third arg NO-VERIFY means trust that the selection's segment
list is up-to-date."
  (let ((seglist (gene-sel-get-seg-list selection no-verify)))
    (save-excursion
      (mapcar
       (lambda (seg)
         (goto-char (car seg))
         (gene-locking-sequence-state-p (gene-get-here-sym) state t))
       seglist))))



(provide 'gene-locking)
