;;;; gene-message.el --- Displaying messages nicely
;; Copyright (C) 1994 Indiana University Foundation
;;
;; This software is copyrighted under the terms of the GNU General
;; Public License (see terms below).  The copyright covers this file,
;; specifically, and the program of which it is a part.
;; 
;;      This program is free software; you can redistribute it and/or modify
;;      it under the terms of the GNU General Public License as published by
;;      the Free Software Foundation; either version 1, or (at your option)
;;      any later version.
;; 
;;      This program is distributed in the hope that it will be useful,
;;      but WITHOUT ANY WARRANTY; without even the implied warranty of
;;      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;      GNU General Public License for more details.
;; 
;;      You should have received a copy of the GNU General Public License
;;      along with this program; if not, write to the Free Software
;;      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.



(defvar gene-message-prefix ""
  "The prefix to put before all messages displayed with gene-message.")

(defmacro gene-with-message-prefix (prefix &rest forms)
  "Add PREFIX to the ambient set of message prefixes while evaluating FORMS.
Messages displayed with `gene-message' will start with PREFIX,
preceded by the prefixes established by surrounding
`gene-with-message-prefix' forms."
  (append (list 'let
                (list (list 'gene-message-prefix
                            (list 'concat 'gene-message-prefix prefix))))
          forms))
(put 'gene-with-message-prefix 'lisp-indent-function 1)

(defun gene-message (text &rest parameters)
  "Display TEXT (formatted with PARAMETERS), with prefixes.
Establish prefixes using gene-with-message-prefix."
  (apply 'message (concat gene-message-prefix text) parameters))


;;; Percentage message abstraction code.

;; This code was lovingly handcrafted by our award-winning artists to
;; avoid allocation, and is considered by many experts as likely to go
;; up in value.  We have limited the edition to thirty firing-days, so
;; be sure to place your order as soon as possible.


(defconst gene-%-percentage-string
  (let ((str (make-string 4 ? )))
    (aset str 2 ?%)
    (aset str 3 ?%)
    str)
  "Permanent string used by the `gene-%-*' functions.")


(defsubst gene-%-calculate-percentage (num denom)
  "Return a string of the form \"x%\", where x is NUM/DENOM as a percentage."
  (let* ((percentage (/ (* num 100) denom))
         (ones       (mod percentage 10))
         (ones-char  (aref (number-to-string ones) 0))
         (tens       (/ (- percentage ones) 10))
         (tens-char  (aref (number-to-string tens) 0)))
    (if (= tens 0)
        (progn
          (aset gene-%-percentage-string 0 ones-char)
          (aset gene-%-percentage-string 0 ? ))
      (aset gene-%-percentage-string 0 tens-char)
      (aset gene-%-percentage-string 1 ones-char))
    gene-%-percentage-string))


(defvar gene-%-last-raw-string ""
  "Permanent string used by the \"gene-%-*\" functions.
It holds, unmodified, the last format string passed by the user.")


(defvar gene-%-output-string ""
  "Permanent string used by the \"gene-%-*\" functions.
It is the result of concatenating a percentage to
`gene-%-last-raw-string', except that where possible the appending
will be done with the more efficient `aset'.")


(defvar gene-%-output-string-start-idx 0
  "Don't ask.")


(defun gene-%-build-message-string (string num denom)
  "Convert STRING into a percentage message based on NUM and DENOM.
For example, if string is

       \"Copying sequences... \"

and NUM and DENOM are 3 and 10, respectively, then the string

       \"Copying sequences... 30%\"

will be returned."
  (let ((percentage-string (gene-%-calculate-percentage num denom)))
    (if (string-equal string gene-%-last-raw-string)
        (let ((count 0))
          (while (< count 4)
            (aset gene-%-output-string
                  (+ gene-%-output-string-start-idx count)
                  (aref percentage-string count))
            (setq count (1+ count)))
          gene-%-output-string)
      ;; Else we'll have to build a new string
      (setq
       gene-%-last-raw-string          string
       gene-%-output-string-start-idx  (length gene-%-last-raw-string)
       gene-%-output-string            (concat string percentage-string))
      )))


(defun gene-%-message (string num denom)
  "Run string through `gene-%-build-message-string' and then message it."
  (message (gene-%-build-message-string string num denom)))




(provide 'gene-message)
