;;;; gene-mode-tools.el (part of ale).
;; Copyright (C) 1993, 1994  Karl Fogel <kfogel@darwin.life.uiuc.edu>
;; Copyright (C)       1994  Indiana University Foundation
;; Copyright (C)       1994  Sixth Congregational Church of Rodney
;;
;; This software is copyrighted under the terms of the GNU General
;; Public License (see terms below).  The copyright covers this file,
;; specifically, and the program of which it is a part.
;; 
;;      This program is free software; you can redistribute it and/or modify
;;      it under the terms of the GNU General Public License as published by
;;      the Free Software Foundation; either version 1, or (at your option)
;;      any later version.
;; 
;;      This program is distributed in the hope that it will be useful,
;;      but WITHOUT ANY WARRANTY; without even the implied warranty of
;;      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;      GNU General Public License for more details.
;; 
;;      You should have received a copy of the GNU General Public License
;;      along with this program; if not, write to the Free Software
;;      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

(require 'cl)

(require 'gene-message)


;; Needed both for checking buffer-size-exceeded AND in the selection
;; code.  It is intentionally one less than the highest Emacs Lisp
;; integer.
(defconst gene-really-high-number (let ((val 1))
                                    (while (> val 0)
                                      (setq val (lsh val 1)))
                                    (setq
                                     val (1+ val)
                                     val (abs val)
                                     val (1- val)))
  "*A really high number -- don't leave home without one.")


;;; Color checking:
(defvar gene-allowable-colors nil
  "A cache for `gene-check-color-allowable'.")

(make-face 'gene-check-color-dummy-face)

(defun gene-check-color-allowable (color &optional resort)
  "Return COLOR iff COLOR exists, else return a default.
The default is the value of optional argument RESORT, or if that is
nil, then the value of `gene-foreground-color'.
RESORT had better be a color guaranteed to exist and be allocatable."
  (if (member color gene-allowable-colors)
      ;; It's cached, so we know it's safe:
      color
    ;; Else it's not cached, so test it and cache it if legal
    (if (x-color-defined-p color)
        ;; Check that it is allocatable too
        (condition-case
            therr
            (progn
              (set-face-foreground 'gene-check-color-dummy-face color)
              ;; We won't get to this point if there was an error...
              (setq gene-allowable-colors ; (cache it)
                    (cons color gene-allowable-colors))
              color)
          (error ; ... we'll get here on error instead.
           (gene-err-accumulate-error "%S" therr)
           ;; So return a safe default:
           (or resort gene-foreground-color)))
      
      ;; Else accum error and return a default color
      (gene-err-accumulate-error "Color not defined: %s" color) 
      (or resort gene-foreground-color))))



;; Used by all the temporary-frame features:
(defun gene-menu-bar-make-hide-button (&optional hider)
  "Put a \"HIDE\" button on the menu bar of the current buffer.
This is for frames that should be click-dismissable.
The Hide button will invoke `delete-frame', if optional argument
HIDER is nil; else hider is a function of no arguments which will take
care of hiding the frame.

Make sure you're in the right buffer when you call this."
  (local-set-key
   [menu-bar gene-close-frame]
   (cons "<HIDE>" (or hider 'delete-frame))))



;;; Bug reporting stuff

(defconst gene-bug-report-address
  "ale-bug@cyclic.com"
  "The targets of your complaints.")


(defun gene-after-send-bug-report-actions ()
  "Function to run when a Ale bug report is send.
This function belongs in `mail-send-actions'."
  ;; do nothing right now.
  )


(defun gene-report-bug ()
  "Pop up a buffer in a new frame, for reporting bugs in the gene editor."
  (interactive)
  (set-buffer (get-buffer-create "*mail bug report*"))
  (mail-mode)
  (setq mail-send-actions
        (cons (list 'gene-after-send-bug-report-actions) mail-send-actions))

  ;; The mail-send-and-exit command marks the buffer as unmodified
  ;; once it's been sent.  This means that if the buffer is
  ;; unmodified, it probably has an old message in it, which we
  ;; shouldn't mess up.
  (if (not (buffer-modified-p))
      (progn
	(erase-buffer)
	(insert "To: " gene-bug-report-address "\n"
		"Subject: \n"
		mail-header-separator "\n"
		"\n"
		"Ale version: " gene-version "\n"
		"Emacs version: " (emacs-version) "\n"
		"\n"
		"Please describe the bug in as much detail as you can.
We need to be able to reproduce it in order to fix it, so it's better
to err on the side of too much detail than too little.  Include even
seemingly irrelevant facts, like whether there was a group selected at
the time, which windows you had up, approximately how long you'd been
running the editor, and ANYTHING else you can think of.
When you're done, press C-c C-c to send the message.
Thanks!\n\n")
	(goto-char (point-min))
	(re-search-forward "^Subject:.*" nil t)))
  (gene-frame-show (current-buffer)
		   '((name             . "Ale Bug Report")
                     (foreground-color . "white")
                     (background-color . "black")
                     (mouse-color      . "white")
                     (cursor-color     . "white")
                     (dedicated        . t)
		     (minibuffer       . t)))
  (gene-menu-bar-make-hide-button))


(setq completion-ignore-case t)

(setq display-buffer-function
      (function (lambda (buf not-this-window)
                  ;; We just ignore the second arg.
                  ;; See `display-buffer' for details.
                  (let ((largest-w (get-largest-window)))
                    (set-window-buffer largest-w buf)
                    largest-w))))


(defun gene-assoc-pred (predicate lst)
  "Call PREDICATE with elements of LST until one returns non-nil.
That element is returned, and PREDICATE is not called with the
remaining elements."
  (catch 'gene-assoc-pred-mitt
    (mapcar
     (lambda (elt)
       (if (funcall predicate elt)
           (throw 'gene-assoc-pred-mitt elt)))
     lst)
    nil))


;;;; Some killers that don't affect the kill-ring:
(defun gene-delete-whole-line (&optional arg)
  "Like `kill-line' with `kill-whole-line' bound to non-nil, but
doesn't affect the kill-ring or the X-selection paste buffer."
  (delete-region
   (point)
   (progn
     (if arg
         (forward-line (prefix-numeric-value arg))
       (if (eobp)
           (signal 'end-of-buffer nil))
       (if (or (looking-at "[ \t]*$") (bolp))
           (forward-line 1)
         (end-of-line)))
     (point))))


(defun gene-delete-word (arg)
  "Like `kill-word' but doesn't affect the kill-ring or the
X-selection paste buffer."
  (interactive "p")
  (delete-region (point) (progn (forward-word arg) (point))))



;;;; gene-save-excursion-in

(defmacro gene-save-excursion-in (buffer &rest forms)
  "Save current buffer and BUFFER's point while evaluating FORMS in BUFFER.
A form like
  (gene-save-excursion-in BUFFER FORMS ...)
is equivalent to
  (save-excursion
    (set-buffer BUFFER)
    (save-excursion
      FORMS ...))
but terser."
  (list 'save-excursion
        (list 'set-buffer buffer)
        (cons 'save-excursion
              forms)))
(put 'gene-save-excursion-in 'lisp-indent-function 1)


(defmacro gene-save-and-restore-selected-group (&rest forms)
  "Unselect the selected group, if any, while executing FORMS.
Returns the value of the last FORM."

  (list 'let '(gene-sgroup-follow-selected-group
               (gene-sgroup-saved-selected-group
                gene-sgroup-selected-group))
        (list 'gene-sgroup-hide-group 'gene-sgroup-selected-group)
        (list 'prog1
              (cons 'progn forms)
              (list 'gene-sgroup-select-group
                    'gene-sgroup-saved-selected-group))))


(defmacro gene-dont-modify-buffer (&rest forms)
  "Evaluate FORMS, preserving the modification state of the buffer and
suppressing `first-change-hook', `before-change-function[s],' and
`after-change-function[s]'."
  (list 'let '(first-change-hook
               before-change-function
               before-change-functions
               after-change-function
               after-change-functions
               (gene-dont-modify-buffer-saved-mod-p
                (buffer-modified-p)))
        (list 'prog1
              (cons 'progn forms)
              '(set-buffer-modified-p
                gene-dont-modify-buffer-saved-mod-p))))



;; Misc.

(defun gene-alignment-empty-p ()
  "Return non-nil iff the alignment contains no sequences."
  (gene-save-excursion-in
      gene-id-buffer
    (goto-char (point-min))
    (or
     (= 0 (buffer-size))
     (not (re-search-forward "[a-zA-Z1-9.+^#$()&@%=]" nil t)))))


(defun gene-absorb-mouse-event ()
  "Ignore a mouse event."
  (interactive)
  ;; Gulp, yum yum!
  )


(defun gene-mode-tools-make-faces ()
  "Create the appropriate faces for gene-mode-tools.
`gene-make-all-faces' calls this.

Note that it's incorrect to make faces in top-level expressions, since
the construction of a face depends on run-time factors (like the fonts
available)."
  (copy-face 'default 'gene-mouse-face)
  ;; better against black backgrounds
  (make-face-bold 'gene-mouse-face nil t))


(defvar gene-mouse-scroll-delay 0.25
  "*The pause between scroll steps caused by mouse drags, in seconds.
If you drag the mouse beyond the edge of a window, Emacs scrolls the
window to bring the text beyond that edge into view, with a delay of
this many seconds between scroll steps.  Scrolling stops when you move
the mouse back into the window, or release the button.
This variable's value may be non-integral.
Setting this to zero causes Emacs to scroll as fast as it can.")


(defun gene-mouse-scroll-subr (jump &optional overlay start)
  "Scroll the selected window JUMP lines at a time, until new input arrives.
If OVERLAY is an overlay, let it stretch from START to the far edge of
the newly visible text.
Upon exit, point is at the far edge of the newly visible text."
  (while (progn
	   (goto-char (window-start))
	   (if (not (zerop (vertical-motion jump)))
	       (progn
		 (set-window-start (selected-window) (point))
		 (if (natnump jump)
		     (progn
		       (goto-char (window-end (selected-window)))
		       ;; window-end doesn't reflect the window's new
		       ;; start position until the next redisplay.  Hurrah.
		       (vertical-motion (1- jump)))
		   (goto-char (window-start (selected-window))))
		 (if overlay
		     (move-overlay overlay start (point)))
		 (if (not (eobp))
		     (sit-for mouse-scroll-delay))))))
  (point))


(defsubst gene-upper-case-p (ch)
  "Return non-nil iff CH is an upper-case alphabetic character."
  (and (>= ch ?A) (<= ch ?Z)))


;; Four defsubsts in three acts.

(progn
 (defsubst gene-bol-point ()
   "Return the point value of the beginning of the current line."
   (save-excursion (beginning-of-line) (point)))


 (defsubst gene-eol-point ()
   "Return the point value of the end of the current line."
   (save-excursion (end-of-line) (point))))


(defsubst gene-seq-first-displayed-line-p ()
  "Return non-nil if on first *displayed* line of the alignment buffer."
  (<= (gene-bol-point) (window-start)))


(defsubst gene-seq-last-displayed-line-p ()
  "Return non-nil if on last *displayed* line of the alignment buffer."
  (>= (1+ (gene-eol-point)) (window-end)))


;;; Some other defsubsts


;; Bases and Gaps, the Untold Story:
;; Instead of having real per-sequence alphabets, there will be one
;; alphabet for the whole alignment right now.  It has only two
;; classes of characters, "bases" and "gaps".
;; The gaps are: - . ~ and SPACE.
;; The bases are everything else.
;;
;; In the future, it may be necessary to add a third class, "illegal",
;; which means deal with it if you find it in your data, but don't let
;; people insert them.  Maybe two new classes, illegal gaps and
;; illegal bases.  But for now, the only likely illegals are spaces,
;; and we're just treating them as gaps.

(defconst gene-gap-regexp "[-~|. ]"
  "Regexp that matches gaps.

Programmers:
If you're going to modify this, you need to also modify
`gene-base-regexp' and the function `gene-gap-p'.")


(defconst gene-base-regexp "[^-~|. ]"
  "Regexp that matches residues.

Programmers:
If you're going to modify this, you need to also modify
`gene-gap-regexp' and the function `gene-gap-p'.")


(defconst gene-skip-gaps-regexp
  (substring gene-gap-regexp
             1
             (1- (length gene-gap-regexp)))
  "Regexp for calls to skip-chars that need to skip gaps.")


(defconst gene-skip-bases-regexp
  (substring gene-base-regexp
             1
             (1- (length gene-base-regexp)))
  "Regexp for calls to skip-chars that need to skip residues.")


;; To avoid the allocation of char-to-string, we just hardcode the
;; bases here.  Try to always keep this code near the base and gap
;; regexps, so that if they ever change, this changes too.
(defsubst gene-gap-p (ch)
  "Return non-nil iff CH is a gap."
  (member ch '(?- ?~ ?| ?. ? )))


(defsubst gene-base-p (ch)
  "Return non-nil iff CH is a residue."
  (not
   (or
    (gene-gap-p ch)
    (= ?\n ch)
    (= ?\r ch))))


(defsubst gene-seq-base-after-p (&optional spot)
  "Returns non-nil if the character after point is a residue.
If optional arg SPOT is non-nil, then use that as point."
  (or spot (setq spot (point)))
  (let ((ch (char-after spot)))
    (gene-base-p ch)))


(defconst gene-complement-table
  '(
    (?T . ?A)
    (?U . ?A)
    (?A . ?U)
    (?C . ?G)
    (?G . ?C)

    (?R . ?Y)
    (?W . ?S)
    (?M . ?K)
    (?H . ?D)
    (?V . ?B)
    (?Y . ?R)
    (?S . ?W)
    (?K . ?M)
    (?D . ?H)
    (?B . ?V)


    (?t . ?a)
    (?u . ?a)
    (?a . ?u)
    (?c . ?g)
    (?g . ?c)

    (?r . ?y)
    (?w . ?s)
    (?m . ?k)
    (?h . ?d)
    (?v . ?b)
    (?y . ?r)
    (?s . ?w)
    (?k . ?m)
    (?d . ?h)
    (?b . ?v)
    )
  "Assoc table of residues and their complements.")


(defsubst gene-complement-char (ch)
  "Return the complement of RESIDUE.
That is, if RESIDUE is A, return U; if G return C, etc.
If RESIDUE does not have a complement, or is not a residue, then it is
returned unmodified.
The residue's case is preserved."
  (or
   (cdr (assoc ch gene-complement-table))
   ch))



;; if below doesn't work, simply replace it with this:
;; (defalias 'gene-goto-column 'move-to-column)
;; however, as move-to-column turns out to be quite expensive (see
;; indent.c), we will use this inline instead:
(defsubst gene-goto-column (col)
  "Go to column COL on the current line, or end-of-line if that's earlier.
Returns non-nil iff it was able to go all the way to COL.
Unlike gene-goto-colum-interactive, this function assumes columns
start at 0."
  (let ((eol (gene-eol-point)))
    (beginning-of-line)
    (let ((actual-col (min col (- eol (point)))))
      (prog1
	  (= col actual-col)
	(forward-char actual-col)))))


(defsubst gene-beginning-of-page-point ()
  "Return the position of the beginning of the current page.
That's the position of first character after the form feed, dear."
  (save-excursion
    (if (re-search-backward page-delimiter nil t)
        (match-end 0)
      (point-min))))


(defsubst gene-end-of-page-point ()
  "Return the position of the end of the current page.
That's the position of the next form feed, dear."
  (save-excursion 
    (if (re-search-forward page-delimiter nil t)
        (match-beginning 0)
      (point-max))))


;;;; A Vector of Translation Tables, and Helper Functions
(defconst gene-char-table-vector (make-vector 256 0)
  "A vector of 256 strings, each of which holds 256 copies of that char.
They are used in calls to translate-region, and it's more efficient to
have them here than to build them on the fly each time.")


;; Put in the strings
(let ((count 0))
  (while (< count 256)
    (aset gene-char-table-vector count (make-string 256 count))
    (setq count (1+ count))))


(defsubst gene-translate-region-to-char (start end char)
  "Translate the region from START to END to CHAR.
Returns the number of characters changed."
  (translate-region start end (aref gene-char-table-vector char)))


(defsubst gene-translate-point-to-char (pt char)
  "Overwrite the char at PT with CHAR.
Returns 1 \(the number of characters changed\)."
  (gene-translate-region-to-char pt (1+ pt) char))


(defun gene-pave-with-string (str &optional loc)
  "Overwrite the text in the current buffer with STR, starting from LOC.
\(LOC defaults to point.\)
If doing so would require the buffer to be extended, then signal an
error and change nothing.

Do not move point.

This function is for use by certain gene routenes, particularly in
selection undo.  It does not handle text properties at all -- the text
properties of STR (if any) are ignored, and the text properties of the
buffer are not affected by the paving at all, either."
  (or loc (setq loc (point)))
  (let ((strlen (length str))
        (idx 0))
    (if (> (+ loc strlen) (point-max))
        (gene-error
         "String length %d too long to pave from point %d."
         strlen loc)
      (while (< idx strlen)
        (gene-translate-point-to-char (+ loc idx) (aref str idx))
        (setq idx (1+ idx))))))
        

(defconst gene-flipcase-translation-table (make-string 256 0)
  "Translation table for toggling case.")

;; Init properly:
(let ((count 0))
  (while (< count 256)
    ;; I think the modern way to do this might be with a char-table /
    ;; case-table?  But this code works, and we currently have no need
    ;; to translate outside the ASCII subset (hence the folding to 255
    ;; for the aset value -- otherwise can get an args-out-of-range
    ;; error because, e.g., (upcase 255) ==> 376).
    (if (= (upcase count) count)
        (if (= (downcase count) count)
            ;; It's non-alphabetic, so include it unmodified:
            (aset gene-flipcase-translation-table count count)
          ;; Else its lower case is different from its upper case:
          (aset gene-flipcase-translation-table count
                (min (downcase count) 255)))
      ;; Else its upper case is different from its lower case:
      (aset gene-flipcase-translation-table count
            (min (upcase count) 255)))
    (setq count (1+ count))))


(defun gene-flipcase-region (start end)
  "Toggle the the case of all characters from START to END."
  (translate-region start end gene-flipcase-translation-table))


;;;; gene-shift-region and friends

(defun gene-shift-region-leave-properties (start end count char)
  "Shift the region from START to END forward COUNT chars
\(backward if COUNT is negative\), and replace vacated space with
copies of CHAR.  Affect no text properties.
Ignore widespread public outcry.
If START and END are markers, they will be adjusted appropriately.

If COUNT is positive, only the text from START to END + COUNT will be
affected; if negative, only the text from STAR - COUNT to END.

It is an error to try to shift past the beginning or the end of the
accessible range of the buffer."
  (if (> start end)
      (setq start (prog1 end (setq end start))))
  (let ((limit 0) ; initialized below
        (idx1  0) ; initialized below
        (idx2  0) ; initialized below
        )
    (cond
     ((= count 0) nil)
     ((> count 0)
      (setq idx1 (1- end) idx2 (+ end (1- count)) limit start)
      (while (>= idx1 limit)
        (gene-translate-point-to-char idx2 (char-after idx1))
        (setq idx1 (1- idx1) idx2 (1- idx2)))
      (gene-translate-region-to-char start (+ start count) char))
     (t ; count < 0, shifting to the left
      (setq idx1 start idx2 (+ start count) limit end)
      (while (< idx1 limit)
        (gene-translate-point-to-char idx2 (char-after idx1))
        (setq idx1 (1+ idx1) idx2 (1+ idx2)))
      (gene-translate-region-to-char end (+ end count) char))
     ))
  ;; shift the markers too, if that's what they are.
  (if (markerp start) (set-marker start (+ start count)))
  (if (markerp end) (set-marker end (+ end count))))


(defun gene-shift-region (start end count char)
  "Shift the region from START to END forward COUNT chars
\(backward if COUNT is negative\), and replace vacated space with
copies of CHAR.
If START and END are markers, they will be adjusted appropriately.

If COUNT is positive, only the text from START to END + COUNT will be
affected; if negative, only the text from STAR - COUNT to END.

It is an error to try to shift past the beginning or the end of the
accessible range of the buffer."
  (if (> start end)
      (setq start (prog1 end (setq end start))))
  (let ((ttable (aref gene-char-table-vector char)))
    (cond
     ((= count 0) nil)
     ((> count 0)
      (transpose-regions start end end (+ end count) t)
      (translate-region start (+ start count) ttable))
     (t ; count < 0, shifting to the left
      (transpose-regions start end start (+ start count) t)
      (translate-region end (+ end count) ttable))))
  ;; shift the markers too
  (if (markerp start) (set-marker start (+ start count)))
  (if (markerp end) (set-marker end (+ end count))))


(defvar gene-seq-buffer-window nil
  "The selected sequence buffer window.
This variable is used when the aligment window is split.
See also functions `gene-seq-buffer-window' and
`gene-window-record-seq-window'.")


(defun gene-seq-buffer-window (&optional frame)
  "Return the window corresponding to the sequence buffer, nil if none.
Optional argument FRAME behaves as it does in `get-buffer-window'.
See also variable `gene-seq-buffer-window'."
  (or gene-seq-buffer-window (get-buffer-window gene-seq-buffer frame)))


(defun gene-seq-buffer-frame ()
  "Return the frame displaying the sequence buffer, nil if none."
  (let ((window (gene-seq-buffer-window 0)))
    (if window (window-frame window))))

(defun gene-id-buffer-window (&optional frame)
  "Return the window corresponding to the id buffer, nil if none.
Optional argument FRAME behaves as it does in `get-buffer-window'."
  (get-buffer-window gene-id-buffer frame))


(defun gene-ann-buffer-window (&optional frame)
  "Return the window corresponding to the annotation buffer, nil if none.
Optional argument FRAME behaves as it does in `get-buffer-window'."
  (get-buffer-window gene-ann-buffer frame))


(defun gene-status-buffer-window ()
  "Return the window corresponding to the status buffer, nil if none."
  (get-buffer-window gene-status-buffer 0))


;; This can be fast now that Jim has the newline-cache running.
;; Note that the first column is column 0.
(defsubst gene-current-column ()
  "Return the number of the current column (columns start with 0)."
  (- (point) (gene-bol-point)))


;; return the line number on which is point, grammatically:
(defsubst gene-current-line ()
  "Return current line number (lines start with 1)."
  (let ((cnt
         (save-restriction
           (save-excursion
             (narrow-to-region (point-min) (point))
             (goto-char (point-min))
             (- (buffer-size) (forward-line (buffer-size)))))))
    (if (bolp) (1+ cnt) cnt)))


(defun gene-seq-line-number (sym)
  "Return the line number of sequence SYM.
If no such sequence, just return 1, not nil or error."
  (save-excursion
    (set-buffer gene-id-buffer)
    (let ((spot
           (text-property-any (point-min) (point-max) 'gene-ident sym)))
      (if spot
          (progn
            (goto-char spot)
            (gene-current-line))
        1))))


(defsubst gene-goto-line (line)
  "Goto the start of line LINE only if it exists, return point or nil.
See also documentation for function  `gene-current-line'."
  (let ((orig-point (point))
        (cur-line (gene-current-line)))
    (if (= 0 (forward-line (- line cur-line)))
        (point)
      (goto-char orig-point)
      nil)))


(defun gene-this-col-nearby-seq (&optional col)
  "Go to nearest different sequence containing COLumn, or punt if none.
In that sequence, go to column COL.
To punt means to get out of the current sequence but go somewhere
unpredictable (probably eol on the next or previous sequence).
COL defaults to the current column."
  (interactive)
  (or col (setq col (gene-current-column)))

  ;; save-excursion won't do, since we might be in the seq buffer, and
  ;; we're trying to side-effect point!
  (let ((obuffer (current-buffer)))
    (set-buffer gene-seq-buffer)
    (goto-char
     (or
      
      ;; Can we preserve column?  That would be nice...
      (catch 'where
        (let ((count 1)
              (stop-down 0)  ; boundary condition
              (stop-up 0))   ; boundary condition
          (while (or (= 0 stop-up) (= 0 stop-down))
            (if (= 0 stop-down) ; while room still left below
                (save-excursion
                  (setq stop-down (forward-line count))
                  (gene-goto-column col) ; try to go to the column
                  (if (= col (gene-current-column))
                      (throw 'where (point)))))
            (if (= 0 stop-up)   ; while room still left above
                (save-excursion
                  (setq stop-up (forward-line (- count)))
                  (gene-goto-column col) ; try to go to the column
                  (if (= col (gene-current-column))
                      (throw 'where (point)))))
            (setq count (1+ count)))))
      
      ;; Else go to eol on next line if possible:
      (if (= 0 (gene-seq-next-line 1))
          (point)
        ;; In desparation, try previous line:
        (if (= 0 (gene-seq-previous-line 1))
            (point)
          ;; Else we are *really* losing, and this is the only sequence,
          ;; so stay right where we are and pout:
          (point)))))
    
    (setq this-command 'gene-seq-line-motion)
    (gene-handle-vertical-motion)
    (set-buffer obuffer)))


;; todo: why is this better than mouse-set-point?
(defun gene-switch-to-click (event)
  "Go to the site of an EVENT."
  (let ((posn (event-start event)))
    (set-buffer (window-buffer (posn-window posn)))
    (goto-char (posn-point posn))))


;;;; Organism symbols ("l'Obarray und amigos")

;; The Plan:
;; In the alignment buffer, on each character of each sequence, put a
;; text-property named 'gene-ident, which records the a
;; symbol interned in gene-obarray.  The symbol's name will be the
;; same as the organism's short-ID, usually, with perhaps a numeric
;; suffix a la buffers<n>.
;;
;; The plist of the symbol will contain, among other things:
;; 1. gene-id-posn: (start end lineno) of organism name in the ID buffer.
;; 2. gene-seq-posn: (start end lineno) of sequence in the alignment buffer.
;; 3. gene-org-id: an id string -- the name of the organism
;; 4. Any other goodies we can think of... an annotation pointer?
;;
;; Then we can use gene-id-posn to update the highlighting of the id,
;; for example.
;;
;; Timing will be crucial in setting up this obarray and storing the
;; information in the symbols' plists.  Will probably have to discuss
;; the integration of it into the IO system with Jim.  For now, have a
;; function that assumes ID and seq buffers are set up correctly and
;; just fills the plists with the relevant information.
;;
;; The obarray can be used in completion -- jumping to the sequence of
;; an given ID, for example.
;;
;; In fact, there is no reason not to store every ID name in the
;; database in the obarray.  Those not actually in Emacs will simply
;; have null values on the relevant plists.

(defun gene-make-obarray ()
  "Return an obarray suitable for use as `gene-obarray'."
  (make-vector 4001 0))

(defvar gene-obarray (gene-make-obarray)
  "This is the obarray used for looking up information about a sequence.")

;; I am an internist:
(defsubst gene-intern (name)
  "Create and return the symbol named NAME (a string) in `gene-obarray'.
If that symbol already exists, just return it."
  (intern name gene-obarray))


(defsubst gene-intern-soft (name)
  "Return the symbol named NAME (a string) in `gene-obarray'.
If that symbol does not exist, return nil."
  (intern-soft name gene-obarray))


(defun gene-get-default-source (sym)
  "Get SYM's `gene-default-source' property, a gene-source struct."
  (get sym 'gene-default-source))


(defun gene-get-here-sym (&optional pos object no-fiddle)
  "Return the organism symbol at point.
If the text-properties on the line are missing, this function will fix
them.
Optional first arg POS used instead of point, if non-nil.
Optional second arg OBJECT is like the OBJECT arg of
`get-text-property'.
Optional third argument NO-FIDDLE means don't try and make up for
missing properties, just return nil."
  (or
   ;; Simplest case -- it's here, and we just get it:
   (get-text-property (or pos (setq pos (point))) 'gene-ident object)

   (if no-fiddle
       nil

     ;; Else fiddle:
     ;; Deal with the possibility that we're just past last char in a
     ;; sequence (but not an empty one):
     (or
      (and
       (bufferp (or object (setq object (current-buffer))))
       (> pos (point-min))
       (get-text-property (1- pos) 'gene-ident object))
      
      (if (and (eq object gene-seq-buffer) (bolp) (eolp))
          (let ((line (gene-current-line)))
            (gene-save-excursion-in gene-id-buffer
              (gene-goto-line line)
              (get-text-property (point) 'gene-ident)))
        ;; Else this sequence does have some seq data, so we'll have to fix
        ;; up the properties.
        (save-excursion
          (goto-char (or pos (setq pos (point))))
          (if (and (bufferp object)
                   (not (eq object gene-id-buffer)))
              ;; It *is* a buffer, so we should fix up the properties
              ;; there, and return the orgsym.
              (save-excursion
                (set-buffer object)
                (let ((line (gene-current-line)))
                  (set-buffer gene-id-buffer)
                  (gene-goto-line line)
                  (let* ((sym (prog1 (get-text-property (point) 'gene-ident)
                                (set-buffer object)))
                         buffer-read-only
                         (bol-marker (copy-marker (gene-bol-point)))
                         (eol-marker (copy-marker (gene-eol-point))))
                    (put-text-property bol-marker eol-marker
                                       'gene-ident sym)
                    (put-text-property
                     bol-marker eol-marker
                     'gene-source (gene-get-default-source sym))
                    (put sym 'gene-seq-posn (list bol-marker eol-marker))
                    sym))))))))))


(defun gene-get-here-property (prop &optional pos object)
  "Get a property from the organism symbol at point.
Optional second arg POS used instead of point, if non-nil.
Optional third arg OBJECT is like the OBJECT arg of
`get-text-property'."
  (get (gene-get-here-sym pos object) prop))


(defun gene-put-here-property (prop val)
  "Put a property on the organism symbol at point."
  (put (gene-get-here-sym) prop val))


(defun gene-intern-possible-duplicate-id (str)
  "Check if a symbol whose name is STR has already been interned in
the gene-obarray.  If not, intern str there.  Otherwise, make a new
str by adding numeric suffixes until safe, then intern that.
Returns the newly interned symbol (see `gene-intern')."
  (let ((first-sym (gene-intern-soft str)))
    (if first-sym
        ;; Someone with the exact same ID name has already be
        ;; interned, and is being used, so use another name for the
        ;; symbol.
        ;; Start by using the suffix "<1>" (no suffix would mean 0,
        ;; after all), and increment the number until it's safe to
        ;; intern.
        (let ((suffix-num 1))
          (while (gene-intern-soft
                  (concat str "<" (number-to-string suffix-num) ">"))
            (setq suffix-num (1+ suffix-num)))
          (gene-intern (concat str "<" (number-to-string suffix-num) ">")))
      ;; Else it's brand new, so just intern it:
      (gene-intern str))))


(defun gene-list-sequences ()
  "Return a list of all active organism symbols."
  ;; We do this in the ID buffer because lines are short there.
  ;; We do not do this in the gene-obarray because there can be
  ;; leftover symbols there.
  ;; Oh, how I wish for `unintern'.  I wonder why it's so
  ;; unfashionable -- RMS didn't want it, and texts seem to indicate
  ;; that it's rarely implemented.  Why not have shrinkable hash
  ;; tables?  I just don't get it.  -Karl
  (save-excursion
    (set-buffer gene-id-buffer)
    (goto-char (point-min))
    (let (table
          (forward 0))
      (while (= forward 0)
        (let ((sym (gene-get-here-sym)))
          (if sym (setq table (cons sym table)))
          (setq forward (forward-line 1))))
      ;; Might as well return it in "order", why not:
      table)))


(defvar gene-id-buffer nil
  "The ID buffer.")

(defvar gene-seq-buffer nil
  "The sequence buffer.")

(defvar gene-ann-buffer nil
  "The annotation buffer.")

(defvar gene-status-buffer nil
  "The status buffer.")

;;; It's useful to have one function which places all the required
;;; properties on an organism symbol --- this way, if we change the
;;; set of properties we put on organism symbols, we can easily find
;;; all the places in the code which need to provide the new
;;; information, just by searching for the callers of this function.

(defun gene-init-organism-sym
  (name id-region sequence-region &optional source in-sym)
  "Initialize and return the organism symbol for a new organism named NAME.

ID-REGION should be a list (START END), giving the region of the ID
buffer that contains this organism's name.  (ID-REGION does not
contain the newline.)  Set the symbol's `gene-id-posn' appropriately.
If we know the number of the line containing the START--END region,
ID-REGION may also be of the form (START END LINE-NUMBER).  This can
save some time in situations where we're initializing lots of organism
symbols, and we already know their line numbers.

SEQUENCE-REGION is similar, for the organism's sequence.  Set the
symbol's `gene-seq-posn' property appropriately.

Set the symbol's `gene-org-id' property to NAME.
Sets its `gene-lock-state' property to 3 (the default: GapInOut).

Optional fourth arg SOURCE is a gene-source for this sym to start out with.
Cut/Copy/Paste code needs this sometimes.

Optional fifth arg IN-SYM is the organism symbol itself.  If present,
then this function's job is just to store information in its plist."
  (save-excursion
    (let ((sym (or in-sym (gene-intern-possible-duplicate-id name))))
      
      (let (id-line seq-line)
        
        (set-buffer gene-id-buffer)
	(setq id-line (or (elt id-region 2)
			  (save-excursion
			    (goto-char (elt id-region 0))
			    (gene-current-line))))
	(put sym 'gene-id-posn (list
				(copy-marker (elt id-region 0))
				(copy-marker (elt id-region 1))))
    
        (set-buffer gene-seq-buffer)
	(setq seq-line (or (elt sequence-region 2)
			   (save-excursion
			     (goto-char (elt sequence-region 0))
			     (gene-current-line))))
	(put sym 'gene-seq-posn (list
				 (copy-marker (elt sequence-region 0))
				 (copy-marker (elt sequence-region 1))))

        ;; Just a sanity check...
        (if (not (= id-line seq-line))
            (error "gene-init-organism-sym: id not lined up with sequence"))

        (put sym 'gene-org-short-name name)
        (gene-locking-lock-sequence-to-state
         sym gene-locking-gapinout-state t)
        (if source
            (put sym 'gene-default-source source))
        sym))))


(defun gene-sym-accum-sequence (sym start end)
  "Given an organism symbol SYM, add to it the sequence region START..END.
This unions START..END into SYM's `gene-seq-posn' property."
  (let ((old (get sym 'gene-seq-posn)))
    (move-marker (elt old 0) (min (elt old 0) start) gene-seq-buffer)
    (move-marker (elt old 1) (max (elt old 1) end)   gene-seq-buffer)))


(defsubst gene-organism-id-posn (organism)
  "Return the position of the name of ORGANISM in the ID buffer.
ORGANISM is an orgsym.
The return value is of the form (START END MUMBLE ...), where START
and END are markers specifying the buffer region."
  (get organism 'gene-id-posn))


(defsubst gene-organism-seq-posn (organism)
  "Return the position of the sequence of ORGANISM in the alignment buffer.
ORGANISM is an orgsym.
The return value is of the form (START END MUMBLE ...), where START
and END are markers specifying the buffer region."
  (get organism 'gene-seq-posn))


(defun gene-organism-seq-data (organism)
  "Return organism's sequence data as a string."
  (let* ((posn  (gene-organism-seq-posn organism))
         (start (nth 0 posn))
         (end   (nth 1 posn)))
    (save-excursion
      (set-buffer gene-seq-buffer)
      (buffer-substring-no-properties start end))))


(defun gene-organism-seq-data-unpadded (organism)
  "Return organism's sequence data as a string, without indels at either end."
  (let* ((posn  (gene-organism-seq-posn organism))
         (start (nth 0 posn))
         (end   (nth 1 posn)))
    (gene-save-excursion-in gene-seq-buffer
      (goto-char start)
      (skip-chars-forward gene-skip-gaps-regexp end)
      (setq start (point))
      (if (= start end)
          ""
        (goto-char end)
        (skip-chars-backward gene-skip-gaps-regexp start)
        (buffer-substring-no-properties start (point))))))


(defsubst gene-organism-name (&optional organism)
  "Return the name of the organism denoted by ORGANISM, an orgsym.
Nil or omitted ORGANISM means use organism at point."
  (get (or organism (gene-get-here-sym)) 'gene-org-short-name))


(defsubst gene-organism-default-source (&optional organism)
  "Return the default source of ORGANISM, an orgsym.
Nil or omitted ORGANISM means use organism at point."
  (get (or organism (gene-get-here-sym)) 'gene-default-source))


(defun gene-organism-in-alignment-p (organism)
  "Return non-nil iff ORGANISM (an orgsym) is in the alignment buffers."
  (gene-save-excursion-in
      gene-seq-buffer
    (text-property-any (point-min) (point-max) 'gene-ident organism)))


;;;; Sources --- where did these bases come from?

;;; Once we've introduced a bunch of text into the sequence buffer, we
;;; need to record the file it came from (so we can save it back
;;; there), and a pointer to its annotation page.  The `source' data
;;; structure records this information, and is placed on the
;;; `gene-source' property of the sequence text.

(defstruct gene-source

  ;; The file associated with this sequence; ordinary saves write the
  ;; sequence to this file, and `save-as' commands set this.
  filename

  ;; The format of FILENAME.
  file-format

  ;; A list (START END) pointing to the region of the annotation data
  ;; for this sequence.  START and END should both be markers.
  annotation)
  
(defsubst gene-source-at-point ()
  "Return the source structure for the text at point."
  (get-text-property (point) 'gene-source))
  

(defsubst gene-source-end ()
  "Return the end of the region provided by the source at point."
  (or (text-property-not-all (point) (point-max)
                             'gene-source (gene-source-at-point))
      (point-max)))


(defsubst gene-get-here-source (&optional name)
  "Get a `gene-source' for the sequence at point, or make one up if
absolutely necessary.
If optional arg NAME is passed, a new source will be constructed if
necessary, as if for a sequence named NAME." 
  (or
   (gene-source-at-point)
   (get (gene-get-here-sym) 'gene-default-source)
   (progn
     (let ((src (make-gene-source
                 :filename gene-io-current-source-file
                 :file-format 'genbank
                 :annotation (gene-io-make-and-add-annotation
                              (or name "NEW.SEQ")
                              'genbank))))
       (put (gene-get-here-sym) 'gene-default-source src)
       (put-text-property (gene-bol-point) (gene-eol-point) 'gene-source src)
       src))))


(defvar gene-mouse-sensitive-sequences nil
  "Non-nil means that sequences will be bold under the mouse.
Default is nil.")


;;;; Buffer-list operations, restricted to those buffers in gene mode

;;; These functions are supposed to contain all the code which constructs
;;; and destructures the names of gene-mode buffers.  Other code should just
;;; use these functions, so if we want to change the conventions, we only
;;; have to change this code.

(defun gene-split-buffer-name (name)
  "Separate NAME, the name of a gene-mode buffer, into its parts.
If NAME is of the form \"FOO ID's\", return (\"FOO\" \" ID's\").
If NAME is of the form \"FOO Sequences\", return (\"FOO\" \" Sequences\").
If NAME is of the form \"FOO Annotations\", return (\"FOO\" \" Annotations\").
If NAME has none of the above forms, return (\"NAME\" \"\")."
  (if (or (string-match " ID's\\'" name)
          (string-match " Sequences\\'" name)
          (string-match " Annotations\\'" name))
      (list (substring name 0 (match-beginning 0))
            (substring name (match-beginning 0)))
    (list name "")))
   
(defun gene-make-buffer-names (root)
  "Return a quadruplet of gene-mode buffer names, using ROOT as a root name.
The return value is a vector like:
   [\"ROOT ID's\" \"ROOT Annotations\" \"ROOT Sequences\" \"ROOT Status\"]"
  (vector (concat root " ID's")
          (concat root " Annotations")
          (concat root " Sequences")
          (concat root " Status")))


(defconst gene-mode-list '(gene-id-mode
                           gene-seq-mode
                           gene-ann-mode
                           gene-phylo-mode
                           gene-status-mode)
  "List of symbols for the gene editing modes.")

(defun gene-mode-buffers ()
  "Return an alist of the buffers in some gene mode (annotation, id, or seq).
The alist associates names with buffers.  Thus, it may be usefully
passed to completing-read."
  (save-excursion
    (loop for buf in (buffer-list)
          if (memq (progn (set-buffer buf) major-mode) gene-mode-list)
          collect (cons (buffer-name buf) buf))))


(defun gene-other-buffer (buffer)
  "Returns most recently selected gene-mode buffer other than BUFFER.
Buffers not visible in windows are preferred to visible buffers.
If no other gene-mode buffer exists, nil is returned."
  (let ((list (gene-mode-buffers)))
    (let ((tail list)
          chosen)
      (while (consp tail)
        (let ((buf (cdr (car tail))))
          (cond
           ((eq buf buffer))
           ((string-match "\\` " (buffer-name buf)))
           (t
            (if (not (get-buffer-window buf t))
                (setq chosen buf
                      tail '(dummy))
              (or chosen
                  (setq chosen buf))))))
        (setq tail (cdr tail)))
      chosen)))


(defun gene-mode-buffers-and-roots ()
  "Return an alist of the buffers in some gene mode, and their name roots.
Normally, one finds gene-mode buffers in triplets named \"FOO ID's\",
\"FOO Sequences\", and \"FOO annotations.\" In such a case, FOO is the
root name.
The alist associates names with buffers.  Thus, it may be usefully
passed to completing-read."
  (let ((buffer-alist (gene-mode-buffers)))
    (let ((root-table (make-vector 19 0)))
      (dolist (buf-assoc buffer-alist)
        (let* ((name (buffer-name (cdr buf-assoc)))
               (split (gene-split-buffer-name name))
               (root (elt split 0))
               (type (elt split 1)))
          (cond
           ((string= type " ID's")
            (put (intern root root-table) 'id t))
           ((string= type " Sequences")
            (put (intern root root-table) 'seq t)))))
      (let ((roots (loop for sym being the symbols of root-table ;COBOL
                         if (and (get sym 'id) (get sym 'seq))
                         collect (let ((name (symbol-name sym)))
                                   (cons name (get-buffer
                                               (concat name " ID's")))))))
        (append roots buffer-alist)))))


;;;; Exiting and quitting

(defun gene-maybe-offer-save ()
  "Offer to save data if there are unsaved changes."
  (interactive)
  ;; Check both out of paranoia:
  (if (or gene-modified-p (buffer-modified-p gene-seq-buffer))
      (if (gene-y-or-n-p "*** Save data? *** ")
          (gene-save-sequences)))
  ;; Make sure we return non-nil
  ;; (see docstring for var `kill-emacs-query-functions')
  t)

(add-hook 'kill-emacs-query-functions 'gene-maybe-offer-save)

(put 'save-buffers-kill-emacs 'disabled t)


(defun gene-quit-maybe-offer-save ()
  "Quit only after offering to save all sequences."
  (interactive)
  (gene-maybe-offer-save)
  (if (gene-y-or-n-p "Quit editor? ")
      (kill-emacs)
    (ding)
    (message "Quit cancelled.")))



;;;; Closing

(provide 'gene-mode-tools)
