;; gene-search.el (part of ale).
;; Copyright (C) 1993, 1994  Karl Fogel <kfogel@darwin.life.uiuc.edu>
;;
;; This software is copyrighted under the terms of the GNU General
;; Public License (see terms below).  The copyright covers this file,
;; specifically, and the program of which it is a part.
;; 
;;      This program is free software; you can redistribute it and/or modify
;;      it under the terms of the GNU General Public License as published by
;;      the Free Software Foundation; either version 1, or (at your option)
;;      any later version.
;; 
;;      This program is distributed in the hope that it will be useful,
;;      but WITHOUT ANY WARRANTY; without even the implied warranty of
;;      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;      GNU General Public License for more details.
;; 
;;      You should have received a copy of the GNU General Public License
;;      along with this program; if not, write to the Free Software
;;      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
;;
;; Search for instances of "todo:" to find notes about future plans.

;;; searching
;; two possible approaches: build a string that is the current
;; sequence minus the bases, and search in that, using the resulting
;; index in a call to `gene-seq-goto-base', or do the re-search
;; directly on the buffer, gaps and all.

(defconst gene-search-base-disambiguation-table (make-vector 256 nil)
  "Vector mapping characters (residues) onto regular expressions.")


(defvar gene-search-ambiguous-characters nil
  "A list of all characters which are ambiguity codes.
`gene-search-modify-base-disambiguation-table' modifies this list.")


(defun gene-search-ambiguous-character-p (ch)
  "Return non-nil iff CH is an ambiguity code."
  (member ch gene-search-ambiguous-characters))


;; Initially, set all of the chars to themselves:
(let ((i 0))
  (while (< i 256)
    (aset gene-search-base-disambiguation-table
          i
          (let ((up-char (upcase i))
                (down-char (downcase i)))
            (if (= up-char down-char)
                (char-to-string i)
              (concat (downcase (char-to-string i))
                      (upcase (char-to-string i))))))
    (setq i (1+ i))))


;; Then modify those that represent probability vectors:
(defun gene-search-modify-base-disambiguation-table (char string)
  "Set CHAR's entry in the base disambiguation table to STRING.
STRING is a string consisting of only the letters that CHAR should match.
Case conversions will be done automatically, so it's only necessary to
include one case, in both CHAR and STRING."
  (let ((upch (upcase char))
        (dnch (downcase char))
        (upst (upcase string))
        (dnst (downcase string)))
    (setq gene-search-ambiguous-characters
          (cons upch (cons dnch gene-search-ambiguous-characters)))
    (aset gene-search-base-disambiguation-table dnch (concat upst dnst))
    (aset gene-search-base-disambiguation-table upch (concat upst dnst))))


;; These are RNA/DNA specific.  We can make functions to switch among
;; the various alphabets later.
;; Disambiguation happens only once, that's why expanding "n" to
;; "nagcut" doesn't lead to an infinite recursion.
;; Also, because it only happens once, its necessary that both "u" and
;; "t" be included manually in the expansions.
;; Finally, some codes are supersets of other codes, so the former
;; must include the latter in the expansion.  So "n", which includes
;; *everything*, really maps to "nbvdhkmswyragcut".  Whew!
(gene-search-modify-base-disambiguation-table ?u "ut")
(gene-search-modify-base-disambiguation-table ?r "rag")
(gene-search-modify-base-disambiguation-table ?y "ycut")
(gene-search-modify-base-disambiguation-table ?w "waut")
(gene-search-modify-base-disambiguation-table ?s "sgc")
(gene-search-modify-base-disambiguation-table ?m "mac")
(gene-search-modify-base-disambiguation-table ?k "kgut")
(gene-search-modify-base-disambiguation-table ?h "hmywacut")
(gene-search-modify-base-disambiguation-table ?d "drkwagut")
(gene-search-modify-base-disambiguation-table ?v "vrsmagc")
(gene-search-modify-base-disambiguation-table ?b "bsykgcut")
(gene-search-modify-base-disambiguation-table ?n "nbvdhkmswyragcut")


(defun gene-search-disambiguate-base (ch)
  "Return a string of all residues that ambiguity code CH could match.
Put another way, if the returned string were the inside of a this
regular expression:

   \"[STRING]\"

then that regular expression would match any character that CH would
match.  See also `gene-search-disambiguate-base-then-regexpify'."
  (aref gene-search-base-disambiguation-table ch))



(defun gene-search-disambiguate-base-then-regexpify (ch)
  "Return a regexp matching all residues that ambiguity code CH could match."
  (concat "[" (gene-search-disambiguate-base ch) "]"))


(defvar gene-search-last-search-string ""
  "The last string we converted to a gappy-regexp.
This is a cache var, is all.")


(defvar gene-search-last-gappy-regexp ""
  "The last gappy regexp we generated.
This is a cache var, is all.")


(defun gene-search-build-gappy-regexp (str)
  "Build gap-ignoring, ambiguity-resolving regexps from strings of residues.
For example,
      \"AuN\"
becomes 
      \"[aA][-~\. ]*[uU][-~\. ]*[NBVDHKMSWYRAGCUTnbvdhkmswyragcut]\".

No, I'm not joking."

  (if (string-equal str gene-search-last-search-string)
      gene-search-last-gappy-regexp
    
    ;; Else it's a new string, so re-cache everything
    (setq
     gene-search-last-search-string
     str
     gene-search-last-gappy-regexp
     (let* ((here 0)
            (len (1- (length str)))
            ;; we'll build the actual regexp from this at the end:
            regexp-list)
       (while (<= here len)
         (let ((ch (aref str here)))
           (if (gene-gap-p ch)
               
               ;; Niels wants to ignore gaps in the search string.
               ;; However, if we were going to include them literally
               ;; (quoted and protected), this would be the place to do
               ;; it:
               nil
             
             ;; Else we're looking at a base, so disambiguate and
             ;; concatenate with gene-gap-regexp.
             (setq regexp-list
                   (cons (if (= here len)
                             ;; This is the last char, so no need to
                             ;; speculate about the bases that might follow
                             ;; it in the alignment buffer:
                             (gene-search-disambiguate-base-then-regexpify ch)
                           (concat 
                            (gene-search-disambiguate-base-then-regexpify ch)
                            gene-gap-regexp
                            ;; to allow for repeated gaps
                            "*"))
                         regexp-list))))
         
         (setq here (1+ here)))
       
       ;; Now reverse regexp-list and convert it into a string
       (apply 'concat (nreverse regexp-list))))))


;; For search matches:
(defvar gene-search-match-color "wheat3"
  "*Color used for highlighting search matches.")


(defun gene-search-make-faces ()
  "Create the appropriate faces for gene searching.
`gene-made-all-faces' calls this.

Note that it's incorrect to make faces in top-level expressions, since
the construction of a face depends on run-time factors (like the fonts
available)."
  (copy-face 'default 'gene-search-match-face)
  (set-face-background
   'gene-search-match-face
   (gene-check-color-allowable gene-search-match-color)))


(defvar gene-search-overlay-list nil
  "List of overlays used in highlighting searches.")

(defvar gene-search-immediate-repeat-start-point nil
  "Records the last starting position of a search.
This is used in immediate repeats of the same search.")

(defvar gene-search-last-search-record nil
  "This variable records the last search so that it can be redone.")


(defun gene-search-repeat-last-search ()
  "Repeat the last search (omitting any repeat count specifications).
If a search is repeated immediately, then it will start from 1 past
the start of the last match.  That is, in a search-forward, one after
the first character, and in a search-backward, one before the last
character of the previous match."
  (interactive)
  (if (eq last-command 'gene-search-do-search)
      (goto-char gene-search-immediate-repeat-start-point))
  (if gene-search-last-search-record
      (apply 'gene-search-do-search gene-search-last-search-record)
    (gene-error "No previous search on record.")))


;; This is the guts of all of the searching functions.
(defun gene-search-do-search (func str &optional bound noerror count)
  "Search from point using FUNC for regular expression STR.
Set point to the end of the occurrence found and highlight the match.

FUNC should have the conventional return values of Emacs searching
functions: nil if failed and NOERROR \(see below\), else point.
Its return value will be the return value of `gene-search-do-search'.

The end of the match could really be the beginning, if FUNC was a
backwards-searching function.

STR will be converted to a regular expression that ignores indels and
knows about ambiguous residues.

An optional second argument bounds the search; it is a buffer position.
The match found must not extend after that position.
Optional third argument, if t, means if fail just return nil (no error).
  If not nil and not t, move to limit of search and return nil.
Optional fourth argument is repeat count--search for successive occurrences.

See also `gene-search-repeat-last-search', and possibly also functions
`match-beginning', `match-end' and `replace-match'."
  ;; Record the search for possible repeat.  It seems logical not to
  ;; include count as part of this record.
  (set-buffer gene-seq-buffer)
  (let ((compiled-str (gene-search-build-gappy-regexp str)))

    ;; Record for `gene-search-repeat-last-search'.
    (setq gene-search-last-search-record
          (list func str bound noerror nil))
    
    ;; Do the search
    (let ((retval (funcall func compiled-str bound noerror count)))
      (if  retval
          ;; Take care of highlighting here, by using the match-data.
          (let* ((s-beg (match-beginning 0))
                 (s-end (match-end 0)))
            (gene-search-highlight-match s-beg s-end)
            ;; Record in case of an immediate repeat search:
            (setq gene-search-immediate-repeat-start-point
                  (if (< (point) s-end)
                      ;; was a backward search
                      (1- s-end)
                    ;; else was forward
                    (1+ s-beg)))))
      ;; Set this-command for `gene-search-repeat-last-search'.
      (setq this-command 'gene-search-do-search)
      retval)))


(defun gene-search-highlight-match (b e)
  "Highlight the search match from B to E."
  (interactive)
  (let ((ov (make-overlay s-beg s-end)))
    (overlay-put ov 'face 'gene-search-match-face)
    (setq gene-search-overlay-list (cons ov gene-search-overlay-list))
    (gene-undo-register-method
     `(lambda () 
        (setq
         gene-search-overlay-list
         (delq ,ov gene-search-overlay-list
        (delete-overlay ,ov)))))))


(defun gene-search-unhighlight-all-matches ()
  "Unhighlight all search matches by deleting their overlays."
  (interactive)
  (mapcar
   (lambda (ov)
     (delete-overlay ov))
   gene-search-overlay-list)
  (setq gene-search-overlay-list nil))


(defun gene-search-forward-in-sequence (str &optional parg noerror)
  "Search forward in this sequence for STR, ignoring gaps.
Ambiguous characters will be disambiguated, whatever that means.
Point will be left at the end (right side) of the match.
Numeric argument (optional second arg PARG) means to search forward
for the Nth match.
Optional third arg NOERROR means just return nil on failure."
  (interactive "sSearch forward for: \np")
  (gene-search-do-search
   're-search-forward str (gene-eol-point) noerror parg))


(defun gene-search-backward-in-sequence (str &optional parg noerror)
  "Search backward in this sequence for STR, ignoring gaps.
Ambiguous characters will be disambiguated, whatever that means.
Point will be left at the start (left side) of the match.
Numeric argument (optional second arg PARG) means to search forward
for the Nth match.
Optional third arg NOERROR means just return nil on failure."
  (interactive "sSearch backward for: \np")
  (gene-search-do-search
   're-search-backward str (gene-bol-point) noerror parg))


(defun gene-search-forward-in-alignment (str &optional parg noerror)
  "Search forward in the alignment for STR, ignoring gaps.
Ambiguous characters will be disambiguated, whatever that means.
Point will be left at the end (right side) of the match.
Numeric argument (optional second arg PARG) means to search forward
for the Nth match.
Optional third arg NOERROR means just return nil on failure."
  (interactive "sSearch forward for: \np")
  (gene-search-do-search 're-search-forward str nil noerror parg)
  (gene-handle-vertical-motion t))


(defun gene-search-backward-in-alignment (str &optional parg noerror)
  "Search backward in this sequence for STR, ignoring gaps.
Ambiguous characters will be disambiguated, whatever that means.
Point will be left at the start (left side) of the match.
Numeric argument (optional second arg PARG) means to search forward
for the Nth match.
Optional third arg NOERROR means just return nil on failure."
  (interactive "sSearch backward for: \np")
  (gene-search-do-search 're-search-backward str nil noerror parg)
  (gene-handle-vertical-motion t))



;; Matching in selections
(defun gene-search-matches-in-selection (selection &optional search-string)
  "Highlight all matches in SELECTION for SEARCH-STRING.
If SEARCH-STRING is nil or omitted, the user will be prompted for it.
Indels are ignored, as is case, and ambiguity codes will be disambiguated."
  (or search-string (setq search-string (read-string "Search for: ")))
  (save-excursion
    (gene-sel-map-over-selection
     selection
     (lambda (start end)
       (while (< start end)
         (goto-char start) 
         (gene-search-do-search 're-search-forward search-string end t)
         (setq start (1+ start)))))))


;; NOT IMPLEMENTED YET!
(defun gene-search-forward-in-sgroup (str &optional parg)
  "Search forward in current sequence group for BASE-STRING, ignoring gaps.
Point will be left at the end (right side) of the match.
Numeric argument means to search forward for the Nth match"
  (interactive "sSearch forward for: \np")

  ;; First, put ourselves in the current sgroup.  If already in there,
  ;; then just search forward from point, else visibly go to start of
  ;; sgroup, and take it from there.

  (let ((re (gene-search-build-gappy-regexp str))
        (sgroup gene-sgroup-selected-group))
    (if sgroup
        (if (re-search-forward re (gene-eol-point) t parg)
            (let ((match-s (match-beginning 0))
                  (match-e (match-end 0)))
              (progn
                (goto-char (1- match-e))
                (move-overlay gene-search-match-overlay match-s match-e)))
          (if gene-use-bell (ding))
          (gene-message "No match for %s." str))
      (gene-message "No current sequence group."))))


;; NOT IMPLEMENTED YET!
(defun gene-search-backward-in-sgroup (str &optional parg)
  "Search backward in current sequence group for BASE-STRING, ignoring gaps.
Point will be left at the start (left side) of the match.
Numeric argument means to search backward for the Nth match."
  (interactive "sSearch backward for: \np")

  ;; First, put ourselves in the current sgroup.  If already in there,
  ;; then just search backward from point, else visibly go to end of
  ;; sgroup, and take it from there.
  (gene-error "Beep!  Not implemented yet, sorry!")
  (let ((re (gene-search-build-gappy-regexp str))
        (sgroup gene-sgroup-selected-group))
    (if sgroup
        (if (re-search-backward re (gene-bol-point) t parg)
            (let ((match-s (match-beginning 0))
                  (match-e (match-end 0)))
              (progn
                (goto-char match-s)     ; damn o-b-o-e
                (move-overlay gene-search-match-overlay match-s match-e)))
          (if gene-use-bell (ding))
          (gene-message "No match for %s." str))
      (gene-message "No current sequence group."))))


;; NOT IMPLEMENTED YET!
(defun gene-search-forward-in-selection (str &optional parg)
  "Search forward in this selection for BASE-STRING, ignoring gaps.
Point will be left at the end (right side) of the match.
Numeric argument means to search forward for the Nth match"
  (interactive "sSearch forward for: \np")
  (gene-error "Beep!  Not implemented yet, sorry!")
  (let ((re (gene-search-build-gappy-regexp str)))
    (if (re-search-forward re (gene-eol-point) t parg)
        (let ((match-s (match-beginning 0))
              (match-e (match-end 0)))
          (progn
            (goto-char (1- match-e))
            (move-overlay gene-search-match-overlay match-s match-e)))
      (if gene-use-bell (ding))
      (gene-message "No match for %s." str))))


;; NOT IMPLEMENTED YET!
(defun gene-search-backward-in-selection (str &optional parg)
  "Search backward in this selection for BASE-STRING, ignoring gaps.
Point will be left at the start (left side) of the match.
Numeric argument means to search backward for the Nth match."
  (interactive "sSearch backward for: \np")
  (gene-error "Beep!  Not implemented yet, sorry!")
  (let ((re (gene-search-build-gappy-regexp str)))
    (if (re-search-backward re (gene-bol-point) t parg)
        (let ((match-s (match-beginning 0))
              (match-e (match-end 0)))
          (progn
            (goto-char match-s)         ; damn o-b-o-e
            (move-overlay gene-search-match-overlay match-s match-e)))
      (if gene-use-bell (ding))
      (gene-message "No match for %s." str))))



(provide 'gene-search)
