;;;; gene-windows.el (part of ale).
;; Copyright (C) 1994  Karl Fogel <kfogel@darwin.life.uiuc.edu>
;;
;; This software is copyrighted under the terms of the GNU General
;; Public License (see terms below).  The copyright covers this file,
;; specifically, and the program of which it is a part.
;; 
;;      This program is free software; you can redistribute it and/or modify
;;      it under the terms of the GNU General Public License as published by
;;      the Free Software Foundation; either version 1, or (at your option)
;;      any later version.
;; 
;;      This program is distributed in the hope that it will be useful,
;;      but WITHOUT ANY WARRANTY; without even the implied warranty of
;;      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;      GNU General Public License for more details.
;; 
;;      You should have received a copy of the GNU General Public License
;;      along with this program; if not, write to the Free Software
;;      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

(require 'gene-mode-tools)

;;; Window Arranging:

;; The Plan:
;; We'll have a function, gene-arrange-windows, that takes a bunch o'
;; buffers, maybe a frame (defaults to selected frame) and arranges
;; the windows accordingly.  Widths are decided by both request and
;; heuristics.  Things like gene-display-all, whatever, can just call
;; this with the right arguments, and wondrous magic will happen.

(defun gene-arrange-windows (&rest buffers)
  "Arrange BUFFERS in windows of their own on the alignment frame.
Each one of BUFFERS is a buffer object.
If no BUFFERS are given, then just refresh the visible ones to insure
their widths are correct.
They are displayed, from left to right, in the order they are passed.
Those buffers who have \"desired\" widths will receive that width, if
possible, and the others will be assigned width by a heuristic that
divides up whatever width is left over on the frame.
In practice, this means that the ID buffer will get just enough to
show all its ID strings, the status buffer will get whatever it needs
these days, and the sequence window will get the rest, after
accomodating whatever else is needed (perhaps the annotation buffer?)"

  (let ((o-frame    (selected-frame))
        (algn-frame (window-frame (get-buffer-window gene-seq-buffer t))))
    (select-frame algn-frame)
    (unwind-protect
        (if (null buffers)
            ;; Find out who's displayed and refresh them.
            (let* ((first-w (frame-first-window algn-frame))
                   (w first-w)
                   (already-there-bufs (list first-w)))
              (while (not (eq (setq w (next-window w 0 -1)) first-w))
                (setq already-there-bufs (cons w already-there-bufs)))
              (let ((buflist (mapcar
                              (lambda (w)
                                (window-buffer w))
                              (nreverse already-there-bufs))))
                (apply 'gene-arrange-windows
                       (if (> (length buflist) 1)
                           buflist
                         (list gene-id-buffer gene-seq-buffer)))))
          
          ;; Else, we were passed buffers.
          ;; First, make a list of buffers and the appropriate widths.  It's a
          ;; list of vectors, each vector being [BUFFER WIDTH].
          (let* ((bufwidth-list (copy-sequence buffers))
                 (lst bufwidth-list))
            (while buffers
              (let ((vec (setcar lst (make-vector 2 nil))))
                (aset vec 0 (car buffers))
                (aset vec 1 (gene-deduce-buffer-window-width (car buffers))))
              ;; They're the same length, so they'll both bottom out at the
              ;; same time, thus we need only test `buffers':
              (setq buffers (cdr buffers))
              (setq lst (cdr lst)))
            
            ;; Some of the widths didn't get their widths set.  That's because
            ;; `gene-deduce-buffer-window-width' can return nil if there is no
            ;; particular preferred width for that window.  We have to
            ;; calculate the permissible size for those windows on this frame
            ;; now that we know what the other windows need:
            
            (setq lst bufwidth-list) ; re-use lst
            (let ((fwidth (frame-width (selected-frame)))
                  (count 0)
                  (default-width 0)) ; will be set below
              (while lst
                (let ((desired-width (aref (car lst) 1)))
                  (if desired-width
                      (setq fwidth (- fwidth desired-width))))
                (setq lst (cdr lst)))
              
              ;; Count the number of undecided widths:
              (setq lst bufwidth-list) ; re-use lst
              (while lst
                (let ((desired-width (aref (car lst) 1)))
                  (or desired-width
                      (setq count (1+ count))))
                (setq lst (cdr lst)))
              
              ;; Divide remaining frame width among the undecided windows, but
              ;; don't divide by 0!
              (setq default-width (/ fwidth (max count 1)))
              (setq lst bufwidth-list) ; re-use lst
              
              ;; If the frame isn't wide enough to handle everybody, we error
              ;; for now.  Later, we'll see if we can't resize the frame
              ;; before giving up.
              (if (<= default-width 1)
                  (gene-error "Please widen frame and try again."))
              
              ;; Else do it!
              (delete-other-windows)
              (while bufwidth-list
                ;; windowize the first buffer
                (switch-to-buffer (aref (car bufwidth-list) 0))
                (unless (null (cdr bufwidth-list))
                  (split-window-horizontally (or (aref (car bufwidth-list) 1)
                                                 default-width))
                  (other-window 1))
                (setq bufwidth-list (cdr bufwidth-list))))))
      (select-frame o-frame))))


(defun gene-id-calculate-window-width (&optional max-width min-width)
  "Calculate the appropriate width for the ID window.
Just figures out the length of the longest ID and adds 3.
Optional first arg MAX-WIDTH is a value not to be exceeded.
Optional second arg MIN-WIDTH is a value not to be inceeded."
  (gene-save-excursion-in gene-id-buffer
    (goto-char (point-min))
    (let ((width (or min-width 1)))
      (while (progn
               (end-of-line)
               (setq width (max width (gene-current-column)))
               (= (forward-line 1) 0)))
      ;; A catch-and-throw could make this much more efficient, but
      ;; let's face it, who cares?
      (setq width (+ width 3))
      (if max-width
          (min max-width width)
        width))))


(defun gene-status-calculate-window-width ()
  "Return the optimal display width for the status buffer."
  (gene-save-excursion-in
      gene-status-buffer
    ;; Use a line we know is the "right" length.  Actually, almost all
    ;; are, but we might be on the last line of the buffer or
    ;; something.
    (goto-char (point-min))
    ;; Add two to line length:
    (+ (if gene-seq-vertical-scroll-bars 3 2)
       (- (gene-eol-point) (gene-bol-point)))))

(setq window-min-width 4)


(defun gene-deduce-buffer-window-width (buf)
  "Return BUF's preferred window width, or nil if has no preference.
For example, the ID buffer wants to be a little bit longer than its
longest ID.  The status buffer wants to be only long enough to show
all the status buttons."
  ;; Examine buf's mode, return a value based on that.
  (save-excursion
    (set-buffer buf)
    (cond
     ((eq major-mode 'gene-id-mode)
      (max window-min-width (gene-id-calculate-window-width)))
     ((eq major-mode 'gene-status-mode)
      (max window-min-width (gene-status-calculate-window-width)))
     ((eq major-mode 'gene-ann-mode)
      (max window-min-width 80)) ; hah!
     (t
      nil))))


(defun gene-show-only-canonical-windows ()
  "Display the ID and Alignment windows, hiding all others."
  (interactive)
  (gene-arrange-windows gene-id-buffer gene-seq-buffer)
  (setq gene-window-split-p nil))


(defvar gene-window-split-p nil
  "Non-nil means the seq window is divided by a vertical line.")

(defun gene-split-unsplit-window-horizontally (&optional force)
  "Toggle the horizontal splitted-ness of the seq window.
Optional argument FORCE positive -- greater than zero -- means split
it, non-nil and not positive means unsplit it."
  (interactive "P")
  (if force

      ;; Just do whatever FORCE says to do
      (if (and (numberp force) (> force 0))
          (or gene-window-split-p (gene-split-window-horizontally))
        (and gene-window-split-p (gene-unsplit-window-horizontally)))

    ;; Else base the decision on the current splittedness
    (if gene-window-split-p
        (gene-unsplit-window-horizontally)
      (gene-split-window-horizontally))

    (if gene-window-split-p
        (add-hook 'pre-command-hook 'gene-window-record-seq-window)
      (setq gene-seq-buffer-window nil)
      (remove-hook 'pre-command-hook 'gene-window-record-seq-window))

    ))

;; This is necessary because some functions, like
;; `gene-sgroup-point-to-group', mess up the selected window when the
;; alignment is down the middle.  If you're in the right-hand window,
;; it'll return you to the left, because the function
;; gene-seq-buffer-window, which just calls get-buffer-window, finds
;; the leftmost window display the seq buffer.
;;
;; The fix is to record the selected window, if it's the seq buffer's
;; window, in a variable (also named gene-seq-buffer-window), and then
;; make gene-seq-buffer-window first check that variable before it
;; tries to find out by using get-buffer-window.  We have the
;; recording function be called in `pre-command-hook'.

(defun gene-window-record-seq-window ()
  "Record the current window for use by `gene-seq-buffer-window'.
This function belongs in `pre-command-hook'."
  (and
   gene-window-split-p
   (eq (window-buffer (selected-window)) gene-seq-buffer)
   (setq gene-seq-buffer-window (selected-window))))


(defun gene-split-window-horizontally ()
  "Internal function used by `gene-split-unsplit-window-horizontally'.
Please use that function instead of this one."
  (if (gene-status-buffer-window)
      (gene-arrange-windows gene-status-buffer
                            gene-id-buffer
                            gene-seq-buffer
                            gene-seq-buffer)
    (gene-arrange-windows gene-id-buffer
                          gene-seq-buffer
                          gene-seq-buffer))
  (setq gene-window-split-p t))


(defun gene-unsplit-window-horizontally ()
  "Internal function used by `gene-split-unsplit-window-horizontally'.
Please use that function instead of this one."
  (if (gene-status-buffer-window)
      (gene-arrange-windows gene-status-buffer
                            gene-id-buffer
                            gene-seq-buffer)
    (gene-arrange-windows gene-id-buffer
                          gene-seq-buffer))
  (setq gene-window-split-p nil))
    



(provide 'gene-windows)
